/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include "mgl.h"

/*
ɸ७Υ

        +---+
        | 1 | 16
        +---+
         16

+----------------------------------------------------------+
|Esc| 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 0 | - | ^ | BS   |
+---+---+---+---+---+---+---+---+---+---+---+---+---+------+
|Tab | q | w | e | r | t | y | u | i | o | p | @ | [ |Enter|
+----+---+---+---+---+---+---+---+---+---+---+---+----+    |
| Caps| a | s | d | f | g | h | j | k | l | ; | : | ] |    |
+-----+---+---+---+---+---+---+---+---+---+---+---+---+----|
| Shift| z | x | c | v | b | n | m | , | . | / | \ | | | |
+------+---+---+---+---+---+---+---+---+---+---+---+---+---+
| Ctrl |Alt  |Hide |             |Kanji| Kata  | | | |
+------+-----+-----+-------------+-----+-------+---+---+---+
*/


char *kb_layout1="\
-----------------------------------------------------------+\n\
 Esc| 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 | 0 | - | ^ | BS   |\n\
----+---+---+---+---+---+---+---+---+---+---+---+---+------+\n\
 Tab | q | w | e | r | t | y | u | i | o | p | @ | [ |Enter|\n\
-----+---+---+---+---+---+---+---+---+---+---+---+----+    |\n\
  Caps| a | s | d | f | g | h | j | k | l | ; | : | ] |  |\n\
------+---+---+---+---+---+---+---+---+---+---+---+---+----|\n\
  Shift| z | x | c | v | b | n | m | , | . | / | \\ | | || |\n\
-------+---+---+---+---+---+---+---+---+---+---+---+---+---+\n\
  Ctrl |Alt  |Hide |             |Kanji|       | | | |\n\
-------+-----+-----+-------------+-----+-------+---+---+---+\n\
";

char *kb_layout2="\
-----------------------------------------------------------+\n\
 Esc| ! | \" | # | $ | % | & | ' | ( | ) | ~ | = | ~ | BS   |\n\
----+---+---+---+---+---+---+---+---+---+---+---+---+------+\n\
 Tab | Q | W | E | R | T | Y | U | I | O | P | ` | { |Enter|\n\
-----+---+---+---+---+---+---+---+---+---+---+---+----+    |\n\
  Caps| A | S | D | F | G | H | J | K | L | + | * | } |  |\n\
------+---+---+---+---+---+---+---+---+---+---+---+---+----|\n\
  Shift| Z | X | C | V | B | N | M | < | > | ? | _ | | || |\n\
-------+---+---+---+---+---+---+---+---+---+---+---+---+---+\n\
  Ctrl |Alt  |Hide |             |Kanji| Kata  | | | |\n\
-------+-----+-----+-------------+-----+-------+---+---+---+\n\
";

char *kb_layout3="\
-----------------------------------------------------------+\n\
 Esc|                                               | BS   |\n\
----+-----------------------------------------------+------+\n\
 Han | q | w | e | r | t | y | u | i | o | p | @ | [ |Enter|\n\
-----+---+---+---+---+---+---+---+---+---+---+---+----+    |\n\
 Zen  | a | s | d | f | g | h | j | k | l | ; | : | ] |  |\n\
------+---+---+---+---+---+---+---+---+---+---+---+---+----|\n\
 Kata  | z | x | c | v | b | n | m | , | . | / | \\ | | || |\n\
-------+---+---+---+---+---+---+---+---+---+---+---+---+---+\n\
 Dic   |Alt  |Hide |             |Alpha| Kata  | | | |\n\
-------+-----+-----+-------------+-----+-------+---+---+---+\n\
";

struct {
	char *str;
	char *alt;
	int size;
	int attr;
	int val;
} conv_table[] = {
{"Esc"	,NULL	,12,0		,033},
{"BS"	,NULL	,12,0		,'\b'},
{"Tab"	,NULL	,12,0		,'\t'},
{"Caps"	,NULL	,12,0		,MK_V1},
{"Shift",NULL	,12,0		,MK_V2},
{"Ctrl"	,NULL	,12,0		,MK_V3},
{"Alt"	,NULL	,12,0		,MK_V4},
{"Hide"	,""	,12,0		,MK_V5},
{"Kanji",""	,12,0		,MK_V6},
{"Alpha","ѿ"	,12,0		,MK_V6},
{"Kata"	,""	,12,0		,MK_V7},
{"Han","Ⱦ"	,12,0		,MK_V9},
{"Zen",""	,12,0		,MK_V10},
{"Dic","Ͽ"	,12,0		,MK_V11},
{""	,NULL	,12,FA_BOLD	,MK_UP},
{""	,NULL	,12,FA_BOLD	,MK_LEFT},
{""	,NULL	,12,FA_BOLD	,MK_RIGHT},
{""	,NULL	,12,FA_BOLD	,MK_DOWN},
{"Enter","Ent"	,12,0		,'\n'},
{""	, "er",'\n'},
};

int conv_keyname(s) char *s; {
	int i;
	for (i=0; i<sizeof(conv_table)/sizeof(conv_table[0]); i++) {
		if (!strcmp(conv_table[i].str,s)) {
			if (conv_table[i].alt) {
				strcpy(s,conv_table[i].alt);
			}
			set_font(conv_table[i].size,conv_table[i].attr);
			return conv_table[i].val;
		}
	}
	set_font(12,0);
	return s[0];
}

create_keymap(layout,v) char *layout; struct virtual_key *v; {
	int cw = 4;
	int ch = 16;
	char buf[1024];
	char *p,*top,*line_buf[10];
	int line_size[10];
	char key_name[20];
	int line = 0;
	int ln,ls;
	int x,xx,kx,w;
	int i,j,jmax,code;
	struct virtual_key *vk;

printf("create_keymap");
	strcpy(buf,layout);
	p = buf;
	top = buf;
	ls = 0;
	while (*p) {
		if ((*p == '-') || (*p == '+')) {
			ln++;
		}
		if (!ls && (*p == ' ')) ls = p - top;
		if (*p == '\n') {
			*p = 0;
			if (ln < 10) {
				line_buf[line] = top;
				line++;
			} else {
				line_size[line] = ls;
			}
			top = p+1;
			ln = ls = 0;
		}
		p++;
	}
	set_font(12,0);
	for (i=0; i<line; i++) {
		jmax = strlen(line_buf[i]);
		xx = 0;
		p = line_buf[i];
		kx = 0;
		for (x=0; *p; p++,x++) {
			if (*p == '|' && p[1] != '|') {
				if (!kx) {
					key_name[kx++] = ' ';
				}
				key_name[kx] = 0;
				code = conv_keyname(key_name);
				w = ((x - xx)*cw - strlen(key_name)*6)/2;
				vk = create_virtual_key(3+xx*cw,i*ch
						,(x-xx)*cw,ch,code);
				vk_attach(v,vk);
				set_color(COLOR_BLACK);
				draw_string(3+xx*cw+w,i*ch+2,key_name);
				set_color(COLOR_DARKGRAY);
				draw_line(3+x*cw,i*ch,3+x*cw,i*ch+ch);
				xx = x;
				kx = 0;
			}
			else if (*p == '|' && p[1] == '|') {
				key_name[kx++] = *p;
				p++;
			} else if (*p != ' ') {
				key_name[kx++] = *p;
			}
		}
		set_color(COLOR_DARKGRAY);
		if (line_size[i+1])
			draw_line(0,i*ch+ch,3+(line_size[i+1]-1)*cw,i*ch+ch);
		else
			draw_line(0,i*ch+ch,3+(x-1)*cw,i*ch+ch);
	}
}


int main(argc, argv) {
	int c;
	struct virtual_key *v;
	char *layout;
	int lock = 0;
	int kanji = 0;

	SCREEN_WIDTH =  240;
	SCREEN_HEIGHT = 100;
	open_graph();
retry:
	set_color(COLOR_LIGHTGRAY);
	clear_screen();
	if (kanji)
		layout = kb_layout3;
	else if (lock) 
		layout = kb_layout2;
	else
		layout = kb_layout1;
		
	v = create_virtual_key(0,0,240,240,MK_V0);
	create_keymap(layout,v);
	vk_attach(NULL,v);
refresh();
	while ((c = get_key(-1)) != 033) {
		if (c == MK_V1) {
			lock = !lock;
			vk_detach(v,1);
			goto retry;
		}
		if (c == MK_V6) {
			kanji = !kanji;
			vk_detach(v,1);
			goto retry;
		}
	}
	close_graph();
	exit(0);
}
