/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <sys/param.h>
#include <sys/ioctl.h>

int debug_draw_font;

#ifdef __FreeBSD__
#  include <machine/console.h>
#endif
#ifdef __linux__
#  include <linux/kd.h>
#  include <linux/vt.h>
#  define memcpy(dst,src,size) memmove(dst,src,size)
#  ifndef VT_TRUE
#    define VT_TRUE 1
#  endif
#endif

#ifndef MAP_FILE
#  define MAP_FILE 0
#endif

#ifdef X_EMU
#  include <errno.h>
#  include <sys/stat.h>
#  include <X11/Xlib.h>
#  include <X11/Xutil.h>
#  include <X11/Xatom.h>
#  include <X11/keysym.h>
#  ifdef X_ROTATE
#    include <X11/cursorfont.h>
#  endif /* X_ROTATE */
#  ifdef LOCALE
#    ifdef X_LOCALE
#      include <X11/Xlocale.h>
#    else  /* non X_LOCALE */
#      include <locale.h>
#    endif /* X_LOCALE */

#    ifdef SOLARIS /* for SunOS 5.5.1 i86pc */
#      define FONT_8       "-misc-*-medium-r-normal-*-8-*-*-*-*-*-iso8859-*,\
                            -*-*-medium-r-*-*-8-*-*-*-*-*-jisx0208.1983-*"
#      define FONT_12      "-misc-fixed-medium-r-semicondensed--12-*-iso8859*,\
                            -morisawa-gothic medium bbb-medium-r-normal-sans-12-*-jisx0208*"
#      define FONT_16      "-sony-fixed-medium-r-normal--16-*-iso8859*,\
                            -sony-fixed-medium-r-normal--16-*-iso8859*"
#      define FONT_24      "-*-fixed-medium-r-normal--24-*-iso8859*,\
                            -*-fixed-medium-r-normal--24-*-jisx0208*"
#    else  /* non SOLARIS */
#      define FONT_8       "-*-*-medium-r-normal-*-8-*-*-*-*-*-iso8859-*,\
                            -*-*-medium-*-*-*-8-*-*-*-*-*-jisx0208.1983-*"
#      define FONT_12      "-misc-fixed-medium-r-*--12-*-iso8859*,\
                            -*-fixed-medium-r-normal-*-12-*-jisx0208*"
#      define FONT_16      "-*-fixed-medium-r-normal--16-*-iso8859*,\
                            -*-fixed-medium-r-normal--16-*-jisx0208*"
#      define FONT_24      "-*-fixed-medium-r-normal--24-*-iso8859*,\
                            -*-fixed-medium-r-normal--24-*-jisx0208*"
#    endif /* SOLARIS */
#  else  /* non LOCALE */
#    ifdef SOLARIS /* for SunOS 5.5.1 i86pc */
#      define FONT_8A      "-misc-*-medium-r-normal-*-8-*-*-*-*-*-iso8859-*"
#      define FONT_8       "-*-*-medium-*-*-*-8-*-*-*-*-*-jisx0208.1983-*"
#      define FONT_8K      "-*-*-medium-*-*-*-8-*-*-*-*-*-jisx0201*"

#      define FONT_12A     "-misc-fixed-medium-r-semicondensed--12-*-iso8859*"
#      define FONT_12      "-morisawa-gothic medium bbb-medium-r-normal-sans-12-*-jisx0208*"
#      define FONT_12K     "-morisawa-gothic medium bbb-medium-r-normal-sans-12-*-jisx0201*"

#      define FONT_16A     "-sony-fixed-medium-r-normal--16-*-iso8859*"
#      define FONT_16      "-*-fixed-medium-r-normal--16-*-jisx0208*"
#      define FONT_16K     "-*-fixed-medium-r-normal--16-*-jisx0201*"

#      define FONT_24A     "-*-fixed-medium-r-normal--24-*-iso8859*"
#      define FONT_24      "-*-fixed-medium-r-normal--24-*-jisx0208*"
#      define FONT_24K     "-*-fixed-medium-r-normal--24-*-jisx0201*"
#    else  /* non SOLARIS */
#      define FONT_8A      "-*-*-medium-r-normal-*-8-*-*-*-*-*-iso8859-*"
#      define FONT_8       "-*-*-medium-*-*-*-8-*-*-*-*-*-jisx0208.1983-*"
#      define FONT_8K      "-*-*-medium-*-*-*-10-*-*-*-*-*-jisx0201*"

#      define FONT_12A     "-misc-fixed-medium-r-*--12-*-iso8859*"
#      define FONT_12      "-*-fixed-medium-r-normal-*-12-*-jisx0208*"
#      define FONT_12K     "-*-fixed-medium-r-normal-*-12-*-jisx0201*"

#      define FONT_16A     "-*-fixed-medium-r-normal--16-*-iso8859*"
#      define FONT_16      "-*-fixed-medium-r-normal--16-*-jisx0208*"
#      define FONT_16K     "-*-fixed-medium-r-normal--16-*-jisx0201*"

#      define FONT_24A     "-*-fixed-medium-r-normal--24-*-iso8859*"
#      define FONT_24      "-*-fixed-medium-r-normal--24-*-jisx0208*"
#      define FONT_24K     "-*-fixed-medium-r-normal--24-*-jisx0201*"
#    endif /* SOLARIS */
#  endif /* LOCALE */

#  define W_TITLE      "Mobile Gear"
#  define I_TITLE      "Mobile Gear"
#  define R_NAME       "Mobile_Gear"
#  define R_CLASS      "Mobile"

#  ifdef VGA16
#    define WHITE      (getenv("MGL_WHITE")    ?getenv("MGL_WHITE")    :"#b8cca3")
#    define LIGHTGRAY  (getenv("MGL_LIGHTGRAY")?getenv("MGL_LIGHTGRAY"):"#95a685")
#    define DARKGRAY   (getenv("MGL_DARKGRAY") ?getenv("MGL_DARKGRAY") :"#565f4c")
#    define BLACK      (getenv("MGL_BLACK")    ?getenv("MGL_BLACK")    :"#171914")
#  else
#    define WHITE        "#b8cca3"
#    define LIGHTGRAY    "#95a685"
#    define DARKGRAY     "#565f4c"
#    define BLACK        "#171914"
#  endif /* VGA16 */

#  define PIXMAP_W     32
#  define PIXMAP_H     32

static Display        *display;
static Window         main_win;
static XEvent         event;
static XTextProperty  w_title_property;  /* window name as a property */
static XTextProperty  i_title_property;  /* window name as a property */
static Pixmap         pixmap;
static XImage         *ximage;
static XColor         white[2];
static XColor         lgray[2];
static XColor         dgray[2];
static XColor         black[2];
static GC             gc;
static GC             fg_gc;
static GC             bg_gc;
#  ifdef LOCALE
static XFontSet       fontset8;
static XFontSet       fontset12;
static XFontSet       fontset16;
static XFontSet       fontset24;
#  else  /* non LOCALE */
static Font           font8a;
static Font           font12a;
static Font           font16a;
static Font           font24a;
static Font           font8;
static Font           font12;
static Font           font16;
static Font           font24;
static Font           font8k;
static Font           font12k;
static Font           font16k;
static Font           font24k;
static XFontStruct    *fs8a;
static XFontStruct    *fs12a;
static XFontStruct    *fs16a;
static XFontStruct    *fs24a;
static XFontStruct    *fs8;
static XFontStruct    *fs12;
static XFontStruct    *fs16;
static XFontStruct    *fs24;
static XFontStruct    *fs8k;
static XFontStruct    *fs12k;
static XFontStruct    *fs16k;
static XFontStruct    *fs24k;
#  endif /* LOCALE */
#if 0
static char           *SetDisplay;
#endif
static char           *ximage_data;
static char           *w_title;
static char           *i_title;
static char           r_name[] = R_NAME;
static char           r_class[] = R_CLASS;
static int            x_font_alloc;
static int            bitsperpixel;
#  ifdef X_ROTATE
static XImage         *rotated_ximage;
//static Cursor         cursor; suz
#  endif /* X_ROTATE */
#endif /* X_EMU */

#include "mgl.h"

#ifdef X_EMU
#  ifdef X_ROTATE
static int            x_rotated = DIR_NORTH;

#    define CURSOR XC_exchange
#    define GET_EVENT (ButtonPressMask | KeyPressMask | ExposureMask \
		     | ButtonReleaseMask | Button1MotionMask)
#  else  /* X_ROTATE */
#    define GET_EVENT (KeyPressMask | ExposureMask)
#  endif /* X_ROTATE */
#endif /* X_EMU */

#ifndef MGLDIR
#  define MGLDIR "."
#endif
#define FONTPATH	MGLDIR "/krom.fnt"

static struct screen physical_screen,*current_screen,*svscreen;
static volatile int busy = 0;
static volatile int switchreq = 0;
static volatile int bgmode = 0;

int pen_color = 0; /* 3: white 2: lightgray 1: darkgray 0: black */
int pen_mask_data[4] = {0x00000000, 0x55555555, 0xaaaaaaaa, 0xffffffff};
int pen_mask;

int mgl_screen_width = 640;
int mgl_screen_height = 240;

#if !defined(X_EMU) || defined(TTY_KEY)
static struct termios ttysave;
#endif
static int need_close;

static unsigned char *font_base;

#if 1
/* internal function */
int (*im_readp)(char *ibuf, int ibufsize);
#  ifdef IM_STATIC
extern int im_read(char *ibuf, int ibufsize);
#  endif
static void signal_exit(int n);
static void screen_switch(void);
void im_load(char *name);
static void inline _draw_pixel(int x, int y);
static void inline _draw_pixel_wc(int x, int y, int col);
void toggle_im_mode(void);
int get_im_mode(void);
void put_key(int c);
void put_key_xy(int c,int x,int y);
#  ifndef X_EMU
static void switch_request(int sig);
static int getch(void);
#  endif

#  ifdef X_EMU
void draw_mode_power(void);
void draw_ac_line(void);
void draw_battery(int level);
int  get_apm(int *level);
#    ifdef X_ROTATE
void north_refresh(void);
void west_refresh(void);
void south_refresh(void);
void east_refresh(void);
#    endif
static void inline rec_point(int x, int y);
int font_base_create(void);
int xset_font(char *str, int bytes, int code);
static long long millitime(void);
int key_conv(KeySym key);
#  endif
#endif

#if 0
extern char *malloc(),*getenv();
#endif

struct screen *create_memscreen();

static struct screen *s_freelist;
static struct textscreen *t_freelist;

#ifdef IM_DYNAMIC
#if 0
extern void *dlopen(),*dlsym();
#endif
#endif
void *im_handler;

static struct sighandler {
  int sig;
  void (*handler)(int);
} orighandlers[] = {
  {SIGHUP, NULL},
  {SIGINT, NULL},
  {SIGABRT, NULL},
  {SIGTERM, NULL},
  {SIGPIPE, NULL},
  {SIGSEGV, NULL}
};

struct virtual_key *vk_root,*vk_free;
int vk_x,vk_y;
static struct virtual_key *vkp;
#ifdef X_EMU
static struct virtual_key *vk_cur;
static void vk_init();
static void vk_reverse();
#endif
static int vk_search();
int vk_match(struct virtual_key *p,int x,int y);

#ifdef X_EMU
int r_rect = 0;
int r_minx;
int r_maxx;
int r_miny;
int r_maxy;

#ifdef X_APM
#if 0
#define  MODE_SX 578
#define  MODE_SY 1
#define  MODE_EX 610
#define  MODE_EY 8

#define  APM_SX  618
#define  APM_SY  1
#else
#define  MODE_SX 576
#define  MODE_SY 0
#define  MODE_EX 608
#define  MODE_EY 8

#define  APM_SX  616
#define  APM_SY  0
#endif

#define  AC_LINE 0
#define  BATTERY 1

void draw_mode_power(void) {
    int mode_sx;
    int mode_sy;
    int mode_w;
    int mode_h;
    int status;
    int level;
    int c;
    mode_sx = MODE_SX;
    mode_sy = MODE_SY;
    mode_w  = MODE_EX - MODE_SX;
    mode_h  = MODE_EY - MODE_SY;

    c = pen_color;
    /* ⡼ɤ *//* ̤Ȥʤ */
    set_color(COLOR_LIGHTGRAY);
    fill_rect(mode_sx, mode_sy, mode_w, mode_h);

    /* APM ξּ */
    status = get_apm(&level);

    /* Ÿޡ */
    switch(status){
    case AC_LINE:
        draw_ac_line();
        break;
    case BATTERY:
        draw_battery(level);
        break;
    }

    set_color(c);
    return;
}

#define  BATT_SX 0
#define  BATT_SY 0
#define  BATT_W  14
#define  BATT_H  7
#define  BATT_ADD_SX 13
#define  BATT_ADD_SY 2
#define  BATT_ADD_W  3
#define  BATT_ADD_H  3
#define  BATT_DEL_SX 13
#define  BATT_DEL_SY 3
#define  LEVEL_SX 2
#define  LEVEL_SY 2
#define  LEVEL_W  10
#define  LEVEL_H  3

void draw_ac_line(void) {
    int ac[][2] = {
        {11,1},
        {12,1},
        {13,1},
        {10,2},
        {11,2},
        {12,2},
        {13,2},
        {14,2},
        {15,2},
        { 1,3},
        { 2,3},
        { 9,3},
        {10,3},
        {11,3},
        {12,3},
        {13,3},
        { 3,4},
        { 4,4},
        { 7,4},
        { 8,4},
        {10,4},
        {11,4},
        {12,4},
        {13,4},
        {14,4},
        {15,4},
        { 5,5},
        { 6,5},
        {11,5},
        {12,5},
        {13,5},
    };
    int p_max = 31;
    int i;

    /* ꥢ */
    set_color(COLOR_WHITE);
    fill_rect(APM_SX, APM_SY, BATT_ADD_SX + BATT_ADD_W, BATT_H);

    /* AC-Line  */
    set_color(COLOR_BLACK);
    for(i = 0 ; i < p_max ; i++){
        draw_pixel(APM_SX + ac[i][0], APM_SY + ac[i][1]);
    }
    return;
}

void draw_battery(int level) {
    /* ꥢ */
    set_color(COLOR_WHITE);
    fill_rect(APM_SX, APM_SY, BATT_ADD_SX + BATT_ADD_W, BATT_H);

    /* Battery  */
    set_color(COLOR_BLACK);
    draw_rect(APM_SX + BATT_SX, APM_SY + BATT_SY, BATT_W, BATT_H);
    draw_rect(APM_SX + BATT_ADD_SX, APM_SY + BATT_ADD_SY, BATT_ADD_W, BATT_ADD_H);
    set_color(COLOR_WHITE);
    draw_pixel(APM_SX + BATT_DEL_SX, APM_SY + BATT_DEL_SY);

    /* Battery ٥ */
    level = level / 10;
    if(level > 0){
        set_color(COLOR_BLACK);
        fill_rect(APM_SX + LEVEL_SX, APM_SY + LEVEL_SY, level, LEVEL_H);
    }

    return;
}

#if defined (__FreeBSD__)
#include <sys/ioctl.h>
#include <machine/apm_bios.h>

#define APMDEV  "/dev/apm"

int get_apm(int *level) {
    int fd;
    struct apm_info info;

    /* Init */
    *level = 0;

    /* device open */
    fd = open(APMDEV, O_RDONLY); 
    if(fd == -1){
        return(AC_LINE);
    }

    /* get info */
    if(ioctl(fd, APMIO_GETINFO, &info) == -1){
        close(fd);
        return(AC_LINE);
    }
    close(fd);

    /* power line */
    if(info.ai_acline == 1){
        return(AC_LINE);
    }

    /* battery level */
    if(info.ai_batt_stat != 255){
        if(info.ai_batt_stat <= 3){
            switch(info.ai_batt_stat){
            case 0: /* high */
                *level = 100;
                break;
            case 1: /* low */
                *level = 30;
                break;
            case 2: /* critical */
                *level = 10;
                break;
            }
        }
    }

    if(info.ai_batt_life != 255){
        if(info.ai_batt_life <= 100){
            *level = info.ai_batt_life;
            if(*level < 0){
                *level = 0;
            }
        }
    }

    if(*level == 0){
        return(AC_LINE);
    }else{
        return(BATTERY);
    }
}
#elif defined (__linux__)
/* from xbatt-1.2beta3 */
#define APMDEV  "/proc/apm"

int get_apm(int *level) {
    FILE *fp;
    char buffer[64];
    int  battLife;
    char driver_version[64];
    int  apm_bios_info_major;
    int  apm_bios_info_minor;
    int  apm_bios_info_flags;
    int  ac_line_status;
    int  battery_status;
    int  battery_flag;
    int  time_units;
    char units[64];


    /* Init */
    *level = 0;

    /* device open */
    fp = fopen(APMDEV, "r"); 
    if(!fp){
        return(AC_LINE);
    }

    while(fgets(buffer, 64, fp) != NULL){
        /*
         * for linux-1.3.58 or later
         */
        if(sscanf(buffer, "%s %d.%d 0x%x 0x%x 0x%x 0x%x %d%% %d %s\n",
                  driver_version,
                  &apm_bios_info_major,
                  &apm_bios_info_minor,
                  &apm_bios_info_flags,
                  &ac_line_status,
                  &battery_status,
                  &battery_flag,
                  &battLife,
                  &time_units,
                  units) == 10) {
            *level = battLife;
            if(ac_line_status == 0x01){
                *level = 0;
            }
        }
        /*
         * for apm_bios-0.5
         */
        else
        if(sscanf(buffer, "Battery life: %d%%\n", &battLife) == 1){
            *level = battLife;
        }else
        if(strcmp("AC: on line\n", buffer)==0){
            *level = 0;
        }
    }
    fclose(fp);

    if(*level == 0){
        return(AC_LINE);
    }else{
        return(BATTERY);
    }
}
#else
int get_apm(int *level) {
    /* Init */
    *level = 0;

    /*  ac_line */
    return(AC_LINE);
}
#endif
#endif /* X_APM */

#ifdef X_ROTATE
void north_refresh()
#else  /* X_ROTATE */
void refresh()
#endif /* X_ROTATE */
{
#if 0
int i, j, k;
#else
int i, j;
#endif
int endian = 0;
char pixel;
char *p;
#if 0
static char mask = 0x03;
#endif
int x,y,b;

    p = physical_screen.bitmap;

    if (!r_rect) {
	return;
    }
    if (r_rect == 2) {
	r_minx = 0;
	r_miny = 0;
	r_maxx = physical_screen.width-1;
	r_maxy = physical_screen.height-1;
    } else {
    	r_minx = r_minx<<2;
    	r_maxx = (r_maxx<<2) + 3;
    }
    /* Check */
    if(r_minx < 0){
        r_minx = 0;
    }
    if(r_miny < 0){
        r_miny = 0;
    }
    if(r_maxx > physical_screen.width-1){
        r_maxx = physical_screen.width-1;
    }
    if(r_maxy > physical_screen.height-1){
        r_maxy = physical_screen.height-1;
    }
#if 0
printf("refresh %d (%d %d) - (%d - %d)\n",r_rect,r_minx,r_miny,r_maxx,r_maxy);
#endif
    switch(DefaultDepth(display, DefaultScreen(display))){
#ifdef VGA16
    case 4:
#endif /* VGA16 */
    case 8:
{
char *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
		for (x=r_minx; x<=r_maxx; x++,i++) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = &ximage_data[i];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
            	}
        }
}
        break;
    case 15:
    case 16:
{
unsigned short *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
		for (x=r_minx; x<=r_maxx; x++,i++) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = (unsigned short *)&ximage_data[i * 2];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
#if 0
                        if (XBitmapBitOrder(display)) {
unsigned short w;
                            w = *pcolor;
                            *pcolor = *pcolor<<8;
                            w = w>>8;
                            *pcolor = *pcolor|w;
                        }
#endif
            	}
        }
}
        break;
    case 24:
    case 32:
        if(bitsperpixel == 24){
int offset[3];

		if(ImageByteOrder(display)){
			offset[0] = 0;
			offset[1] = 1;
			offset[2] = 2;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
			for (x=r_minx; x<=r_maxx; x++,i++) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
               		 	switch(pixel){
               		 	case 0:
               		     		ximage_data[i * 3 + offset[0]] = white[0].red;
                    			ximage_data[i * 3 + offset[1]] = white[0].green;
                    			ximage_data[i * 3 + offset[2]] = white[0].blue;
                			break;
                		case 1:
                    			ximage_data[i * 3 + offset[0]] = lgray[0].red;
                    			ximage_data[i * 3 + offset[1]] = lgray[0].green;
                    			ximage_data[i * 3 + offset[2]] = lgray[0].blue;
                    			break;
                		case 2:
                    			ximage_data[i * 3 + offset[0]] = dgray[0].red;
                    			ximage_data[i * 3 + offset[1]] = dgray[0].green;
                    			ximage_data[i * 3 + offset[2]] = dgray[0].blue;
                    			break;
                		case 3:
                    			ximage_data[i * 3 + offset[0]] = black[0].red;
                    			ximage_data[i * 3 + offset[1]] = black[0].green;
                    			ximage_data[i * 3 + offset[2]] = black[0].blue;
                    			break;
                		default:
                    			ximage_data[i * 3 + offset[0]] = white[0].red;
                    			ximage_data[i * 3 + offset[1]] = white[0].green;
                    			ximage_data[i * 3 + offset[2]] = white[0].blue;
                    			break;
                		}
            		}
        	}
        }else
        if(bitsperpixel == 32){
int offset[4];

		if(ImageByteOrder(display)){
			offset[0] = 1;
			offset[1] = 2;
			offset[2] = 3;
			offset[3] = 0;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
			offset[3] = 3;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
			for (x=r_minx; x<=r_maxx; x++,i++) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
                		switch(pixel){
                		case 0:
                    			ximage_data[i * 4 + offset[0]] = white[0].red;
                    			ximage_data[i * 4 + offset[1]] = white[0].green;
                    			ximage_data[i * 4 + offset[2]] = white[0].blue;
                    			ximage_data[i * 4 + offset[3]] = 0x00;
                			break;
                		case 1:
                    			ximage_data[i * 4 + offset[0]] = lgray[0].red;
                    			ximage_data[i * 4 + offset[1]] = lgray[0].green;
                    			ximage_data[i * 4 + offset[2]] = lgray[0].blue;
                    			ximage_data[i * 4 + offset[3]] = 0x00;
                    			break;
                		case 2:
                    			ximage_data[i * 4 + offset[0]] = dgray[0].red;
                    			ximage_data[i * 4 + offset[1]] = dgray[0].green;
                    			ximage_data[i * 4 + offset[2]] = dgray[0].blue;
                    			ximage_data[i * 4 + offset[3]] = 0x00;
                    			break;
                		case 3:
                    			ximage_data[i * 4 + offset[0]] = black[0].red;
                    			ximage_data[i * 4 + offset[1]] = black[0].green;
                    			ximage_data[i * 4 + offset[2]] = black[0].blue;
                    			ximage_data[i * 4 + offset[3]] = 0x00;
                    			break;
                		default:
                    			ximage_data[i * 4 + offset[0]] = white[0].red;
                    			ximage_data[i * 4 + offset[1]] = white[0].green;
                    			ximage_data[i * 4 + offset[2]] = white[0].blue;
                    			ximage_data[i * 4 + offset[3]] = 0x00;
                    			break;
                		}
            		}
        	}
        }else{
            fprintf(stderr, "Unsupport BitmapPad\n");
        }
        break;
    default:
        fprintf(stderr, "Unsupport Depth\n");
        break;
    }

    /* XImage¤Τ򥦥ɥž */
    XPutImage(display, main_win, gc, ximage,
              r_minx, r_miny,
              r_minx, r_miny,
              r_maxx - r_minx +1,
              r_maxy - r_miny +1);
    XSync(display, 0);
    r_rect = 0;
}

#ifdef X_ROTATE
void west_refresh()
{
#if 0
int i, j, k;
#else
int i, j;
#endif
int endian = 0;
char pixel;
char *p;
#if 0
static char mask = 0x03;
#endif
int x,y,b;
int d_width = physical_screen.height;
int d_height = physical_screen.width;
int d_minx;
int d_miny;
int d_maxx;
int d_maxy;
int dst;

    p = physical_screen.bitmap;

    if (!r_rect) return;
    if (r_rect == 2) {
	r_minx = 0;
	r_miny = 0;
	r_maxx = physical_screen.width-1;
	r_maxy = physical_screen.height-1;
    } else {
    	r_minx = r_minx<<2;
    	r_maxx = (r_maxx<<2) + 3;
    }
    /* Check */
    if(r_minx < 0){
        r_minx = 0;
    }
    if(r_miny < 0){
        r_miny = 0;
    }
    if(r_maxx > physical_screen.width-1){
        r_maxx = physical_screen.width-1;
    }
    if(r_maxy > physical_screen.height-1){
        r_maxy = physical_screen.height-1;
    }
#if 0
printf("west_refresh %d (%d %d) - (%d - %d)\n",r_rect,r_minx,r_miny,r_maxx,r_maxy);
#endif
    d_minx = r_miny;
    d_miny = physical_screen.width-r_minx-r_maxx-1;
    d_maxx = r_maxy;
    d_maxy = physical_screen.width-r_minx-1;
    /* Check */
    if(d_minx < 0){
        d_minx = 0;
    }
    if(d_miny < 0){
        d_miny = 0;
    }
    if(d_maxx > d_width-1){
        d_maxx = d_width-1;
    }
    if(d_maxy > d_height-1){
        d_maxy = d_height-1;
    }
#if 0
printf("west_refresh (%d %d) - (%d %d)\n",d_minx,d_miny,d_maxx,d_maxy);
#endif

    switch(DefaultDepth(display, DefaultScreen(display))){
#ifdef VGA16
    case 4:
#endif /* VGA16 */
    case 8:
{
char *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = y+(d_width*d_maxy);
		for (x=r_minx; x<=r_maxx; x++,i++,dst-=d_width) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = &ximage_data[dst];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
            	}
        }
}
        break;
    case 15:
    case 16:
{
unsigned short *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = y+(d_width*d_maxy);
		for (x=r_minx; x<=r_maxx; x++,i++,dst-=d_width) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = (unsigned short *)&ximage_data[dst * 2];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
#if 0
                        if (XBitmapBitOrder(display)) {
unsigned short w;
                            w = *pcolor;
                            *pcolor = *pcolor<<8;
                            w = w>>8;
                            *pcolor = *pcolor|w;
                        }
#endif
            	}
        }
}
        break;
    case 24:
    case 32:
        if(bitsperpixel == 24){
int offset[3];

		if(ImageByteOrder(display)){
			offset[0] = 0;
			offset[1] = 1;
			offset[2] = 2;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
                        dst = y+(d_width*d_maxy);
			for (x=r_minx; x<=r_maxx; x++,i++,dst-=d_width) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
               		 	switch(pixel){
               		 	case 0:
               		     		ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                			break;
                		case 1:
                    			ximage_data[dst * 3 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = lgray[0].blue;
                    			break;
                		case 2:
                    			ximage_data[dst * 3 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = dgray[0].blue;
                    			break;
                		case 3:
                    			ximage_data[dst * 3 + offset[0]] = black[0].red;
                    			ximage_data[dst * 3 + offset[1]] = black[0].green;
                    			ximage_data[dst * 3 + offset[2]] = black[0].blue;
                    			break;
                		default:
                    			ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                    			break;
                		}
            		}
        	}
        }else
        if(bitsperpixel == 32){
int offset[4];

		if(ImageByteOrder(display)){
			offset[0] = 1;
			offset[1] = 2;
			offset[2] = 3;
			offset[3] = 0;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
			offset[3] = 3;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
                        dst = y+(d_width*d_maxy);
			for (x=r_minx; x<=r_maxx; x++,i++,dst-=d_width) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
                		switch(pixel){
                		case 0:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                			break;
                		case 1:
                    			ximage_data[dst * 4 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = lgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 2:
                    			ximage_data[dst * 4 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = dgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 3:
                    			ximage_data[dst * 4 + offset[0]] = black[0].red;
                    			ximage_data[dst * 4 + offset[1]] = black[0].green;
                    			ximage_data[dst * 4 + offset[2]] = black[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		default:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		}
            		}
        	}
        }else{
            fprintf(stderr, "Unsupport BitmapPad\n");
        }
        break;
    default:
        fprintf(stderr, "Unsupport Depth\n");
        break;
    }

    /* XImage¤Τ򥦥ɥž */
    XPutImage(display, main_win, gc, rotated_ximage,
              d_minx, d_miny,
              d_minx, d_miny,
              d_maxx - d_minx +1,
              d_maxy - d_miny +1);
    XSync(display, 0);
    r_rect = 0;
    return;
}

void south_refresh()
{
#if 0
int i, j, k;
#else
int i, j;
#endif
int endian = 0;
char pixel;
char *p;
#if 0
static char mask = 0x03;
#endif
int x,y,b;
int d_width = physical_screen.width;
int d_height = physical_screen.height;
int d_minx;
int d_miny;
int d_maxx;
int d_maxy;
int dst;

    p = physical_screen.bitmap;

    if (!r_rect) return;
    if (r_rect == 2) {
	r_minx = 0;
	r_miny = 0;
	r_maxx = physical_screen.width-1;
	r_maxy = physical_screen.height-1;
    } else {
    	r_minx = r_minx<<2;
    	r_maxx = (r_maxx<<2) + 3;
    }
    /* Check */
    if(r_minx < 0){
        r_minx = 0;
    }
    if(r_miny < 0){
        r_miny = 0;
    }
    if(r_maxx > physical_screen.width-1){
        r_maxx = physical_screen.width-1;
    }
    if(r_maxy > physical_screen.height-1){
        r_maxy = physical_screen.height-1;
    }
#if 0
printf("south_refresh %d (%d %d) - (%d - %d)\n",r_rect,r_minx,r_miny,r_maxx,r_maxy);
#endif
    d_minx = physical_screen.width-r_minx-r_maxx-1;
    d_miny = physical_screen.height-r_miny-r_maxy-1;
    d_maxx = physical_screen.width-r_minx-1;
    d_maxy = physical_screen.height-r_miny-1;
    /* Check */
    if(d_minx < 0){
        d_minx = 0;
    }
    if(d_miny < 0){
        d_miny = 0;
    }
    if(d_maxx > d_width-1){
        d_maxx = d_width-1;
    }
    if(d_maxy > d_height-1){
        d_maxy = d_height-1;
    }
#if 0
printf("south_refresh (%d %d) - (%d %d)\n",d_minx,d_miny,d_maxx,d_maxy);
#endif

    switch(DefaultDepth(display, DefaultScreen(display))){
#ifdef VGA16
    case 4:
#endif /* VGA16 */
    case 8:
{
char *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = (d_width * (d_height - (y+1))) + d_maxx;
		for (x=r_minx; x<=r_maxx; x++,i++,dst--) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = &ximage_data[dst];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
            	}
        }
}
        break;
    case 15:
    case 16:
{
unsigned short *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = (d_width * (d_height - (y+1))) + d_maxx;
		for (x=r_minx; x<=r_maxx; x++,i++,dst--) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = (unsigned short *)&ximage_data[dst * 2];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
#if 0
                        if (XBitmapBitOrder(display)) {
unsigned short w;
                            w = *pcolor;
                            *pcolor = *pcolor<<8;
                            w = w>>8;
                            *pcolor = *pcolor|w;
                        }
#endif
            	}
        }
}
        break;
    case 24:
    case 32:
        if(bitsperpixel == 24){
int offset[3];

		if(ImageByteOrder(display)){
			offset[0] = 0;
			offset[1] = 1;
			offset[2] = 2;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
			dst = (d_width * (d_height - (y+1))) + d_maxx;
			for (x=r_minx; x<=r_maxx; x++,i++,dst--) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
               		 	switch(pixel){
               		 	case 0:
               		     		ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                			break;
                		case 1:
                    			ximage_data[dst * 3 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = lgray[0].blue;
                    			break;
                		case 2:
                    			ximage_data[dst * 3 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = dgray[0].blue;
                    			break;
                		case 3:
                    			ximage_data[dst * 3 + offset[0]] = black[0].red;
                    			ximage_data[dst * 3 + offset[1]] = black[0].green;
                    			ximage_data[dst * 3 + offset[2]] = black[0].blue;
                    			break;
                		default:
                    			ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                    			break;
                		}
            		}
        	}
        }else
        if(bitsperpixel == 32){
int offset[4];

		if(ImageByteOrder(display)){
			offset[0] = 1;
			offset[1] = 2;
			offset[2] = 3;
			offset[3] = 0;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
			offset[3] = 3;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
			dst = (d_width * (d_height - (y+1))) + d_maxx;
			for (x=r_minx; x<=r_maxx; x++,i++,dst--) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
                		switch(pixel){
                		case 0:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                			break;
                		case 1:
                    			ximage_data[dst * 4 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = lgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 2:
                    			ximage_data[dst * 4 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = dgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 3:
                    			ximage_data[dst * 4 + offset[0]] = black[0].red;
                    			ximage_data[dst * 4 + offset[1]] = black[0].green;
                    			ximage_data[dst * 4 + offset[2]] = black[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		default:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		}
            		}
        	}
        }else{
            fprintf(stderr, "Unsupport BitmapPad\n");
        }
        break;
    default:
        fprintf(stderr, "Unsupport Depth\n");
        break;
    }

    /* XImage¤Τ򥦥ɥž */
    XPutImage(display, main_win, gc, ximage,
              d_minx, d_miny,
              d_minx, d_miny,
              d_maxx - d_minx +1,
              d_maxy - d_miny +1);
    XSync(display, 0);
    r_rect = 0;
    return;
}

void east_refresh()
{
#if 0
int i, j, k;
#else
int i, j;
#endif
int endian = 0;
char pixel;
char *p;
#if 0
static char mask = 0x03;
#endif
int x,y,b;
int d_width = physical_screen.height;
int d_height = physical_screen.width;
int d_minx;
int d_miny;
int d_maxx;
int d_maxy;
int dst;

    p = physical_screen.bitmap;

    if (!r_rect) return;
    if (r_rect == 2) {
	r_minx = 0;
	r_miny = 0;
	r_maxx = physical_screen.width-1;
	r_maxy = physical_screen.height-1;
    } else {
    	r_minx = r_minx<<2;
    	r_maxx = (r_maxx<<2) + 3;
    }
    /* Check */
    if(r_minx < 0){
        r_minx = 0;
    }
    if(r_miny < 0){
        r_miny = 0;
    }
    if(r_maxx > physical_screen.width-1){
        r_maxx = physical_screen.width-1;
    }
    if(r_maxy > physical_screen.height-1){
        r_maxy = physical_screen.height-1;
    }
#if 0
printf("east_refresh %d (%d %d) - (%d - %d)\n",r_rect,r_minx,r_miny,r_maxx,r_maxy);
#endif
    d_minx = physical_screen.height-r_miny-r_maxy-1;
    d_miny = r_minx;
    d_maxx = physical_screen.height-r_miny-1;
    d_maxy = r_maxx;
    /* Check */
    if(d_minx < 0){
        d_minx = 0;
    }
    if(d_miny < 0){
        d_miny = 0;
    }
    if(d_maxx > d_width-1){
        d_maxx = d_width-1;
    }
    if(d_maxy > d_height-1){
        d_maxy = d_height-1;
    }
#if 0
printf("east_refresh (%d %d) - (%d %d)\n",d_minx,d_miny,d_maxx,d_maxy);
#endif

    switch(DefaultDepth(display, DefaultScreen(display))){
#ifdef VGA16
    case 4:
#endif /* VGA16 */
    case 8:
{
char *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = d_width*d_miny+(d_width-(y+1));
		for (x=r_minx; x<=r_maxx; x++,i++,dst+=d_width) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = &ximage_data[dst];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
            	}
        }
}
        break;
    case 15:
    case 16:
{
unsigned short *pcolor;
	for (y=r_miny; y<=r_maxy; y++) {
		i = y*physical_screen.width + r_minx;
                dst = d_width*d_miny+(d_width-(y+1));
		for (x=r_minx; x<=r_maxx; x++,i++,dst+=d_width) {
			j = i/4;
			b = i%4;
			if (endian) b = 3-b;
			b *= 2;
			pixel = ((p[j]>>b) & 3);
                	pcolor = (unsigned short *)&ximage_data[dst * 2];
                	switch(pixel){
                	case 0:
                        	*pcolor = white[0].pixel;
                		break;
                	case 1:
                    		*pcolor = lgray[0].pixel;
                    		break;
                	case 2:
                    		*pcolor = dgray[0].pixel;
                    		break;
                	case 3:
                    		*pcolor = black[0].pixel;
                    		break;
                	default:
                    		*pcolor = white[0].pixel;
                    		break;
                	}
#if 0
                        if (XBitmapBitOrder(display)) {
unsigned short w;
                            w = *pcolor;
                            *pcolor = *pcolor<<8;
                            w = w>>8;
                            *pcolor = *pcolor|w;
                        }
#endif
            	}
        }
}
        break;
    case 24:
    case 32:
        if(bitsperpixel == 24){
int offset[3];

		if(ImageByteOrder(display)){
			offset[0] = 0;
			offset[1] = 1;
			offset[2] = 2;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
                	dst = d_width*d_miny+(d_width-(y+1));
			for (x=r_minx; x<=r_maxx; x++,i++,dst+=d_width) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
               		 	switch(pixel){
               		 	case 0:
               		     		ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                			break;
                		case 1:
                    			ximage_data[dst * 3 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = lgray[0].blue;
                    			break;
                		case 2:
                    			ximage_data[dst * 3 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 3 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 3 + offset[2]] = dgray[0].blue;
                    			break;
                		case 3:
                    			ximage_data[dst * 3 + offset[0]] = black[0].red;
                    			ximage_data[dst * 3 + offset[1]] = black[0].green;
                    			ximage_data[dst * 3 + offset[2]] = black[0].blue;
                    			break;
                		default:
                    			ximage_data[dst * 3 + offset[0]] = white[0].red;
                    			ximage_data[dst * 3 + offset[1]] = white[0].green;
                    			ximage_data[dst * 3 + offset[2]] = white[0].blue;
                    			break;
                		}
            		}
        	}
        }else
        if(bitsperpixel == 32){
int offset[4];

		if(ImageByteOrder(display)){
			offset[0] = 1;
			offset[1] = 2;
			offset[2] = 3;
			offset[3] = 0;
		}else{
			offset[0] = 2;
			offset[1] = 1;
			offset[2] = 0;
			offset[3] = 3;
		}

		for (y=r_miny; y<=r_maxy; y++) {
			i = y*physical_screen.width + r_minx;
                	dst = d_width*d_miny+(d_width-(y+1));
			for (x=r_minx; x<=r_maxx; x++,i++,dst+=d_width) {
				j = i/4;
				b = i%4;
				if (endian) b = 3-b;
				b *= 2;
				pixel = ((p[j]>>b) & 3);
                		switch(pixel){
                		case 0:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                			break;
                		case 1:
                    			ximage_data[dst * 4 + offset[0]] = lgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = lgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = lgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 2:
                    			ximage_data[dst * 4 + offset[0]] = dgray[0].red;
                    			ximage_data[dst * 4 + offset[1]] = dgray[0].green;
                    			ximage_data[dst * 4 + offset[2]] = dgray[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		case 3:
                    			ximage_data[dst * 4 + offset[0]] = black[0].red;
                    			ximage_data[dst * 4 + offset[1]] = black[0].green;
                    			ximage_data[dst * 4 + offset[2]] = black[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		default:
                    			ximage_data[dst * 4 + offset[0]] = white[0].red;
                    			ximage_data[dst * 4 + offset[1]] = white[0].green;
                    			ximage_data[dst * 4 + offset[2]] = white[0].blue;
                    			ximage_data[dst * 4 + offset[3]] = 0x00;
                    			break;
                		}
            		}
        	}
        }else{
            fprintf(stderr, "Unsupport BitmapPad\n");
        }
        break;
    default:
        fprintf(stderr, "Unsupport Depth\n");
        break;
    }

    /* XImage¤Τ򥦥ɥž */
    XPutImage(display, main_win, gc, rotated_ximage,
              d_minx, d_miny,
              d_minx, d_miny,
              d_maxx - d_minx +1,
              d_maxy - d_miny +1);
    XSync(display, 0);
    r_rect = 0;
    return;
}

void refresh(void) {
    switch(x_rotated){
    case DIR_NORTH:
        north_refresh();
        break;
    case DIR_WEST:
        west_refresh();
        break;
    case DIR_SOUTH:
        south_refresh();
        break;
    case DIR_EAST:
        east_refresh();
        break;
    }
    return;
}
#endif /* X_ROTATE */
#else
void refresh(void) {
    return;
}
#endif /* X_EMU */

static void screen_switch(void) {
#ifndef X_EMU
	busy++;
	if (switchreq & 1) {
	    int sav_col;
	    switchreq &= ~1;
	    if (!bgmode) {
		bitblt(svscreen,0,0,NULL,0,0
				,SCREEN_WIDTH,SCREEN_HEIGHT,0);
		sav_col = pen_color;
		set_color(COLOR_WHITE);
		fill_rect(0,SCREEN_HEIGHT-8,SCREEN_WIDTH,8);
		fill_rect(0,0,SCREEN_WIDTH,8);
		set_color(sav_col);
		push_screen(svscreen);
		ioctl(0, KDSETMODE, KD_TEXT);
		ioctl(0, VT_RELDISP, VT_TRUE);
		bgmode = 1;
	    }
	}
	if (switchreq & 2) {
	    switchreq &= ~2;
	    if (bgmode) {
		ioctl(0, KDSETMODE, KD_GRAPHICS);
		ioctl(0, VT_RELDISP, VT_ACKACQ);
		pop_screen();
		bitblt(NULL,0,0,svscreen,0,0,SCREEN_WIDTH,SCREEN_HEIGHT,0);
		bgmode = 0;
	    }
	}
	busy--;
#endif
	return;
}

#ifndef X_EMU
static void switch_request(int sig) {
    signal(SIGUSR1, switch_request);  /* should use sigaction()? */
  if (!bgmode) {
    switchreq |= 1;
    if (!busy) {
	screen_switch();
    }
  } else {
    switchreq |= 2;
    if (bgmode && busy == 1) {
	screen_switch();
    }
  }
  return;
}
#endif

/* եå̤Ѳǽˤޤ */
int open_graph(void) {
int i;
#ifdef X_EMU
XSizeHints            xsh;
XClassHint            xch;
XWMHints              xwm;
XSetWindowAttributes  xswa;
XGCValues             gcv;
XPixmapFormatValues   *pmf;
Status                status;
#if 0
int                   ret;
#endif
int                   bytes;
#ifdef X_ROTATE
int                   rotated_bytes;
#endif /* X_ROTATE */
int                   num;

    /****** ǥץ쥤Υץ *******/
    display = XOpenDisplay(NULL);
    if(!display){
        return 0;
    }

    /****** EMUɥκ *******/
    xswa.background_pixel    = WhitePixel(display, DefaultScreen(display));
    xswa.border_pixel        = BlackPixel(display, DefaultScreen(display));

    main_win = XCreateWindow(display,
               DefaultRootWindow(display),
               0, 0,
              (unsigned int)SCREEN_WIDTH,
              (unsigned int)SCREEN_HEIGHT,
               1,
               DefaultDepth(display, DefaultScreen(display)),
               InputOutput,
               DefaultVisual(display, DefaultScreen(display)),
               CWBackPixel | CWBorderPixel,
               &xswa);

    /* wm size hints */
    xsh.flags  = (PBaseSize | USSize | PMinSize | PMaxSize);
    xsh.width       = SCREEN_WIDTH;
    xsh.height      = SCREEN_HEIGHT;
    xsh.base_width  = SCREEN_WIDTH;
    xsh.base_height = SCREEN_HEIGHT;
    xsh.min_width   = SCREEN_WIDTH;
    xsh.min_height  = SCREEN_HEIGHT;
    xsh.max_width   = SCREEN_WIDTH;
    xsh.max_height  = SCREEN_HEIGHT;

    xch.res_name = r_name;
    xch.res_class = r_class;

    xwm.flags = InputHint;
    xwm.input = True;

    w_title = (char *)malloc(strlen(W_TITLE) + 1);
    if(!w_title){
        perror("malloc");
        return 0;
    }
    strcpy(w_title, W_TITLE);

    /* create window_title_property */
    if(!XStringListToTextProperty(&w_title, 1, &w_title_property)){
        fprintf(stderr,
                "%s: structure w_title_property can't be allocated.\n",
                w_title);
        free(w_title);
        return 0;
    }

    i_title = (char *)malloc(strlen(I_TITLE) + 1);
    if(!i_title){
        perror("malloc");
        free(w_title);
        return 0;
    }
    strcpy(i_title, I_TITLE);

    /* create icon_title_property */
    if(!XStringListToTextProperty(&i_title, 1, &i_title_property)){
        fprintf(stderr,
                "%s: structure i_title_property can't be allocated.\n",
                i_title);
        free(w_title);
        free(i_title);
        return 0;
    }

    XSetWMProperties(display, main_win, &w_title_property, &i_title_property,
                     NULL, 0, &xsh, &xwm, &xch);

    /*** ƻ뤹륤٥Ȥλؼ ***/
    XSelectInput(display, main_win, GET_EVENT);

    /****** physical_screenκ *******/
    physical_screen.type = 0;
    physical_screen.wbytes = SCREEN_WIDTH * 2 / 8;
    physical_screen.width  = SCREEN_WIDTH;
    physical_screen.height = SCREEN_HEIGHT;
    physical_screen.bitmap = (char *)malloc(SCREEN_WIDTH * SCREEN_HEIGHT / 4);
    if(!(physical_screen.bitmap)){
        perror("malloc");
        free(w_title);
        free(i_title);
        return 0;
    }

    /****** current_screenphysical_screen *******/
    current_screen = &physical_screen;
    svscreen = create_memscreen(SCREEN_WIDTH,SCREEN_HEIGHT,NULL);

    /* 0: white 1: lightgray 2: darkgray 3: black */
    status = XAllocNamedColor(display,
                              DefaultColormap(display, DefaultScreen(display)),
                              WHITE, &white[0], &white[1]);
    if(!status){
        fprintf(stderr, "white color can't created\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }

    status = XAllocNamedColor(display,
                              DefaultColormap(display, DefaultScreen(display)),
                              LIGHTGRAY, &lgray[0], &lgray[1]);
    if(!status){
        fprintf(stderr, "lightgray color can't created\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }

    status = XAllocNamedColor(display,
                              DefaultColormap(display, DefaultScreen(display)),
                              DARKGRAY, &dgray[0], &dgray[1]);
    if(!status){
        fprintf(stderr, "darkgray color can't created\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }

    status = XAllocNamedColor(display,
                              DefaultColormap(display, DefaultScreen(display)),
                              BLACK, &black[0], &black[1]);
    if(!status){
        fprintf(stderr, "black color can't created\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }

    gcv.function= GXcopy;
    gc = XCreateGC(display, main_win, GCFunction, &gcv);

    switch(DefaultDepth(display, DefaultScreen(display))){
    case 15:
        bytes = SCREEN_WIDTH * 16 / 8;
#ifdef X_ROTATE
        rotated_bytes = SCREEN_HEIGHT * 16 / 8;
#endif /* X_ROTATE */
        break;
#ifdef VGA16
    case 4: /* VGA16 */
        bytes = SCREEN_WIDTH;
#ifdef X_ROTATE
        rotated_bytes = SCREEN_HEIGHT;
#endif /* X_ROTATE */
        break;
#endif /* VGA16 */
    case 8:
    case 16:
        bytes = SCREEN_WIDTH *
                DefaultDepth(display, DefaultScreen(display)) / 8;
#ifdef X_ROTATE
        rotated_bytes = SCREEN_HEIGHT *
                        DefaultDepth(display, DefaultScreen(display)) / 8;
#endif /* X_ROTATE */
        break;
    case 24:
    case 32:
        /* number of supported pixmap formats */
        pmf = XListPixmapFormats(display, &num);
        if(!pmf){
            fprintf(stderr, "supported pixmap formats can't get\n");
            free(w_title);
            free(i_title);
            free(physical_screen.bitmap);
            return 0;
        }
        bitsperpixel = 0;
        for(i = 0 ; i < num ; i++){
            if(pmf[i].depth == DefaultDepth(display, DefaultScreen(display))){
                bitsperpixel = pmf[i].bits_per_pixel;
                break;
            }
        }
        XFree ((char *) pmf);
        if(!bitsperpixel){
            fprintf(stderr, "this pixmap formats can't supported\n");
            free(w_title);
            free(i_title);
            free(physical_screen.bitmap);
            return 0;
        }
        bytes = SCREEN_WIDTH * bitsperpixel / 8;
        /* 640 padding no need ? */
#ifdef X_ROTATE
        rotated_bytes = SCREEN_HEIGHT * bitsperpixel / 8;
        /* 240 padding no need ? */
#endif /* X_ROTATE */
        break;
    default:
        fprintf(stderr, "Unsupport Depth\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }
    ximage_data = (char *)malloc(bytes * SCREEN_HEIGHT);
    if(!ximage_data){
        perror("malloc");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        return 0;
    }

    /* XImage¤Τ */
    ximage = XCreateImage(display,
             DefaultVisual(display, DefaultScreen(display)),
             DefaultDepth(display, DefaultScreen(display)),
             ZPixmap, 0, ximage_data,
            (unsigned int)SCREEN_WIDTH,
            (unsigned int)SCREEN_HEIGHT,
             BitmapPad(display),
             bytes);
    if(ximage == 0){
        fprintf(stderr, "XImage can't created.\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        free(ximage_data);
        return 0;
    }
#ifdef X_ROTATE
    /* XImage¤Τ */
    rotated_ximage = XCreateImage(display,
                     DefaultVisual(display, DefaultScreen(display)),
                     DefaultDepth(display, DefaultScreen(display)),
                     ZPixmap, 0, ximage_data,
                    (unsigned int)SCREEN_HEIGHT,
                    (unsigned int)SCREEN_WIDTH,
                     BitmapPad(display),
                     rotated_bytes);
    if(rotated_ximage == 0){
        fprintf(stderr, "XImage can't created.\n");
        free(w_title);
        free(i_title);
        free(physical_screen.bitmap);
        XDestroyImage(ximage);
        XFree(rotated_ximage);
        return 0;
    }
#endif /* X_ROTATE */

    x_font_alloc = 0;
    font_base = (unsigned char *) mmap((caddr_t)0, 0x200000,
				   PROT_READ,
				   MAP_SHARED | MAP_FILE,
				   open(FONTPATH,0), 0);
    if(font_base == (void *)(-1)){
        perror("font_base");
        if(!font_base_create()){
            fprintf(stderr, "font_base create failed\n");
            free(w_title);
            free(i_title);
            free(physical_screen.bitmap);
            XDestroyImage(ximage);
            XFree(rotated_ximage);
            return 0;
        }
    }
#ifdef X_ROTATE
    /**** ޥκ ****/
    //cursor = XCreateFontCursor(display, CURSOR); suz

    /**** ޥ ****/
    //XDefineCursor(display, main_win, cursor); suz
#endif /* X_ROTATE */
    vk_init();

    /**** ɥĥդ *****/
    XMapWindow(display, main_win);
    XSync(display, 0);

#else  /* non X_EMU */
  struct vt_mode smode;

#if 0
  ioctl(0, VT_WAITACTIVE, 0);
  ioctl(0, KDSETMODE, KD_GRAPHICS);
#endif

   physical_screen.type = 0;
   physical_screen.wbytes = 0x100;
   physical_screen.width  = SCREEN_WIDTH;
   physical_screen.height = SCREEN_HEIGHT;
#ifdef __FreeBSD__
   physical_screen.bitmap = (unsigned char *) mmap((caddr_t)0, 0x10000,
				   PROT_READ | PROT_WRITE,
				   MAP_SHARED | MAP_FILE,
				   0,
				   0);
   if (physical_screen.bitmap == (void *)(-1)) {
	perror("mmap");
#if 0
        exit(1);
#endif
	return 0;
   }
#endif
#ifdef __linux__
   {
       int fb = open("/dev/mgfb", O_RDWR);
       if(fb < 0){
         perror("open /dev/mgfb");
#if 0
         exit(1);
#endif
         return 0;
       }
       physical_screen.bitmap = (unsigned char *) mmap((caddr_t)0, 0x10000,
                                                     PROT_READ | PROT_WRITE,
                                                     MAP_SHARED | MAP_FILE,
                                                     fb,
                                                     0);
       if (physical_screen.bitmap == (void *)(-1)) {
         perror("mmap");
#if 0
         exit(1);
#endif
         return 0;
       }
       close(fb);
   }
#endif
#if (SCREEN_WIDTH < 240)
  physical_screen.bitmap += 8 * 0x100;
#endif
  current_screen = &physical_screen;
  svscreen = create_memscreen(SCREEN_WIDTH,SCREEN_HEIGHT,NULL);

#ifdef __FreeBSD__
  font_base = (unsigned char *) mmap((caddr_t)0, 0x200000,
				   PROT_READ,
				   MAP_SHARED | MAP_FILE,
				   0, 0x10000);
   if (font_base == (void *)-1) {
	perror("mmap font");
#if 0
	exit(1);
#endif
	return 0;
   }
#endif
#ifdef __linux__
   {
       int fn = open("/dev/mgfonts", O_RDONLY);
       if(fn < 0){
         perror("open /dev/mgfonts");
#if 0
         exit(1);
#endif
         return 0;
       }
       font_base = (unsigned char *) mmap((caddr_t)0, 0x200000,
                                        PROT_READ,
                                        MAP_SHARED | MAP_FILE,
                                        fn, 0);
       if (font_base == (void *)-1) {
         perror("mmap font");
#if 0
         exit(1);
#endif
         return 0;
       }
       close(fn);
   }
#endif

#if 1
  ioctl(0, VT_WAITACTIVE, 0);
  ioctl(0, KDSETMODE, KD_GRAPHICS);
#endif

  signal(SIGUSR1,switch_request);

#endif /* X_EMU */
  for (i = 0; i < sizeof(orighandlers) / sizeof(struct sighandler); i++) {
    orighandlers[i].handler = signal(orighandlers[i].sig, signal_exit);
  }

  atexit(close_graph);

#ifndef X_EMU
  smode.mode = VT_PROCESS;
  smode.waitv = 0;
  smode.relsig = SIGUSR1;
  smode.acqsig = SIGUSR1;
  smode.frsig  = SIGINT;
  ioctl(0,VT_SETMODE, &smode);
#endif

#if !defined(X_EMU) || defined(TTY_KEY)
{
  struct termios term;
  tcgetattr(0, &ttysave);
  term = ttysave;
  term.c_iflag = IGNPAR | IGNBRK;
  term.c_oflag = ONLCR;
  term.c_cflag = CREAD | CS8;
  term.c_lflag &= ~(ICANON | IEXTEN | ECHO | ISIG);
  term.c_cc[VTIME] = 0;
  term.c_cc[VMIN] = 1;
  tcsetattr(0, TCSANOW, &term);
}
#endif

#ifdef IM_DYNAMIC
{
  char *name;
  name = getenv("MGLIM");
  if (name) im_load(name);
}
#endif
#ifdef IM_STATIC
  im_readp = im_read;
#endif
  need_close = 1;
  return 1;
}

/* եå̤λѤ򽪤ޤ */
void close_graph(void) {
#ifndef X_EMU
  struct vt_mode smode;
#endif
  if (!need_close) return;
  need_close = 0;
#ifdef X_EMU
    /****** 񸻤β *******/
    if(x_font_alloc){
        free(font_base);
    }
    free(physical_screen.bitmap);

    XDestroyImage(ximage);
    XFreeGC(display, gc);

    free(w_title);
    free(i_title);

#ifdef X_ROTATE
#if 0
    /* ximage_data ϶ͭ */
    rotated_ximage->data = NULL;
    XDestroyImage(rotated_ximage);
#else
    XFree(rotated_ximage);
#endif

    /**** ޥβ ****/
    //XUndefineCursor(display, main_win); suz
    //XFreeCursor(display, cursor);
#endif /* X_ROTATE */

    /****** ǥץ쥤Υ *******/
    XCloseDisplay(display);
#else  /* non X_EMU */
  /****** ̤Υꥢ *******/
  set_color(COLOR_WHITE);
  clear_screen();

  ioctl(0, KDDISABIO, 0);
  ioctl(0, KDSETMODE, KD_TEXT);
  smode.mode = VT_AUTO;
  ioctl(0, VT_SETMODE, &smode);
#endif /* X_EMU */
#if !defined(X_EMU) || defined(TTY_KEY)
  tcsetattr(0, TCSANOW, &ttysave);
#endif
}

static void signal_exit(int n) {
  struct sighandler *sigp;

  close_graph();

  for (sigp = orighandlers; 
       sigp < orighandlers + sizeof(orighandlers) / sizeof(struct sighandler);
       sigp++) {
    if (sigp->sig == n && sigp->handler == SIG_IGN) {
	return;
    }
  }
  for (sigp = orighandlers; 
       sigp < orighandlers + sizeof(orighandlers) / sizeof(struct sighandler);
       sigp++) {
    signal(sigp->sig, sigp->handler);
  }
  kill(getpid(), n);		/* this signal is blocked */
  return;
}

void im_load(char *name) {
#ifdef IM_DYNAMIC
#ifdef __FreeBSD__
#define SYMPREFIX "_"
#else
#define SYMPREFIX
#endif
  char buf[256];

  if (im_handler) dlclose(im_handler);
  im_handler = 0;
  im_readp = 0;
  if (name) {
	if (name[0] == '/') {
		buf[0] = 0;
	} else {
		strcpy(buf,MGLDIR);
		strcat(buf,"/");
	}
	strcat(buf,name);
	if (!(strlen(name) > 3 && !strcmp(name+strlen(name)-3,".so"))) {
		strcat(buf,".so");
	}
	im_handler = dlopen(buf,1);
#if 0
      fprintf(stderr, "%d = dlopen(%s)\n", im_handler, buf);
#endif
	if (im_handler) {
		im_readp = dlsym(im_handler,SYMPREFIX "im_read");
	}
  }
#endif
  return;
}

#ifdef X_EMU
static void inline rec_point(int x, int y) {
	char *p;
	int x1,y1;
	p = (char *)current_screen->bitmap
		+ y * current_screen->wbytes + (x>>2);

	if (p < physical_screen.bitmap
	 || p > physical_screen.bitmap
		 + physical_screen.wbytes*physical_screen.height) return;

	y1 = (p - physical_screen.bitmap) / physical_screen.wbytes;
	x1 = (p - physical_screen.bitmap) % physical_screen.wbytes;

	if (!r_rect) {
		r_minx = x1;
		r_maxx = x1;
		r_miny = y1;
		r_maxy = y1;
		r_rect = 1;
	} else {
		if (x1 < r_minx) r_minx = x1;
		if (x1 > r_maxx) r_maxx = x1;
		if (y1 < r_miny) r_miny = y1;
		if (y1 > r_maxy) r_maxy = y1;
	}
	return;
}
#endif

/* ̤ɤĤ֤ޤ */
void clear_screen(void) {
	char *fb;
	int i,d,y;

	busy++;
	fb = current_screen->bitmap;

#ifdef X_EMU
	rec_point(0,0);
	rec_point(current_screen->width-1,current_screen->height-1);
#endif
	for (y=0; y< current_screen->height; y++) {
		if (pen_color == COLOR_REVERSE) {
			int w = current_screen->width/4;
			for (i=0; i< w; i++) {
				fb[i] ^= 0xff;
			}
		} else {
			memset(fb, pen_mask & 0xff, current_screen->width/4);
		}
		d = current_screen->width%4;
		for (i=0; i<d; i++) {
			draw_pixel(current_screen->width-1-i,y);
		}
		fb += current_screen->wbytes;
	}
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

/* 褹Ȥοꤷޤ
        ϡ0: 1:Ť졼 2:뤤졼 3:
        ط뵡ǽϡclear_screen , draw_pixel, draw_rect,fill_rect
        ,draw_string,draw_font  Ǥ
*/

void set_color(int col) {
	pen_color = col;
	pen_mask = pen_mask_data[col & 3];
	return;
}

static void inline _draw_pixel(int x, int y) {
	char *p;
	int mask;
#if 0
      if(debug_draw_font){
          fprintf(stderr, "go _draw_pixel(%d, %d; %d, %d, %d)\r\n",
              x, y, current_screen->width, current_screen->height, current_screen->wbytes);
      }
#endif
	if (x < 0 || x >= current_screen->width
	   || y < 0 || y >= current_screen->height) return;
	p = (char *)current_screen->bitmap
		+ y * current_screen->wbytes + (x>>2);
	mask = 3 << ((x & 3)*2);
	if (pen_color == COLOR_REVERSE)
		*p = (*p ^ mask);
	else
		*p = (*p & ~mask) | (pen_mask & mask);
#if 0
      if(debug_draw_font){
          fprintf(stderr, "end of _draw_pixel()\r\n");
      }
#endif
      return;
}

/*  x,y  */
void draw_pixel(int x, int y) {
	busy++;
#ifdef X_EMU
	rec_point(x,y);
#endif
	_draw_pixel(x,y);
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

static void inline _draw_pixel_wc(int x, int y, int col) {
	char *p;
	int mask;
	if (col < 0) return;
	if (x < 0 || x >= current_screen->width
	   || y < 0 || y >= current_screen->height) return;

	p = (char *)current_screen->bitmap
		+ y * current_screen->wbytes + (x>>2);
	mask = 3 << ((x & 3)*2);
	*p = (*p & ~mask) | (pen_mask_data[col&3] & mask);
	return;
}

/* ǡ x,y  */
void draw_pixel_wc(int x, int y, int col) {
	busy++;
#ifdef X_EMU
	rec_point(x,y);
#endif
	_draw_pixel_wc(x,y,col);
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}


#define ABS(a) (((a)<0) ? -(a) : (a))

/* (x1,y1)  (x2,y2) ޤޤ */
int draw_line(int x1, int y1, int x2, int y2) {
    int dx = x2 - x1;
    int dy = y2 - y1;
    int ax = ABS(dx) << 1;
    int ay = ABS(dy) << 1;
    int sx = (dx >= 0) ? 1 : -1;
    int sy = (dy >= 0) ? 1 : -1;

    int x = x1;
    int y = y1;

    busy++;
#ifdef X_EMU
    rec_point(x1,y1);
    rec_point(x2,y2);
#endif
#if 1 /* HLine */
    if (y1 == y2) {
	char *p;
	int wb;
	int i;

	if (x1 > x2) {int tmp; tmp = x1; x1 = x2; x2 = tmp;}
	if (x1 <= 0) {
		x1 = 0;
		if (x2 <= 0) goto ret;
	}
	if (x2 >= current_screen->width -1) {
		x2 = current_screen->width - 1;
		if (x1 >= current_screen->width -1) goto ret;
	}
	while (x1 % 4 != 0) {
	    _draw_pixel(x1, y1);
	    x1++;
	}
	wb = (x2 + 1 - x1)/4;
	p = (char *)current_screen->bitmap
		+ y1 * current_screen->wbytes + (x1>>2);
	if (pen_color == COLOR_REVERSE) {
		for (i=0; i< wb; i++) {
			*p++ ^= 0xff;
		}
	} else {
		memset(p, pen_mask & 0xff, wb);
	}
	x1 += wb * 4;
	while (x1 <= x2) {
	    _draw_pixel(x1, y1);
	    x1++;
	}
	goto ret;
    }
#endif
    if (ax > ay) {
	int d = ay - (ax >> 1);
	while (x != x2) {
	    _draw_pixel(x, y);

	    if (d > 0 || (d == 0 && sx == 1)) {
		y += sy;
		d -= ax;
	    }
	    x += sx;
	    d += ay;
	}
    } else {
	int d = ax - (ay >> 1);
	while (y != y2) {
	    _draw_pixel(x, y);

	    if (d > 0 || (d == 0 && sy == 1)) {
		x += sx;
		d -= ay;
	    }
	    y += sy;
	    d += ax;
	}
    }
    _draw_pixel(x, y);

ret:
    if (--busy -bgmode == 0 && switchreq) screen_switch();

    return 0;
}

/* x,y  xs,ys ΥȢ񤭤ޤ */
void draw_rect(int x, int y, int xs, int ys) {
	xs -= 1;
	ys -= 1;
	draw_line(x,y,x,y+ys);
	draw_line(x+xs,y,x+xs,y+ys);
	draw_line(x,y,x+xs,y);
	draw_line(x,y+ys,x+xs,y+ys);
	return;
}

/* x,y  xs,ys ΥȢɤĤ֤ޤ */
void fill_rect(int x, int y, int xs, int ys) {
	int i;
	xs -= 1;
	for (i=0; i<ys; i++) {
		draw_line(x,y+i,x+xs,y+i);
	}
	return;
}

struct romfont {
	int height;
	int width;
	int koffset;
	int kwidth;
	int kbytes;
	int aoffset;
	int awidth;
	int abytes;
	int goffset;
	int gwidth;
	int gbytes;
	int fd;	/* external font */
} romfont[6] = {
	{ 12, 12, 0x00000, 2,32,  0x3f000, 1,16 ,  0x40000, 2,32},
	{ 16, 16, 0x44800, 2,32,  0x83800, 1,16 ,  0x84800, 2,32},
	{ 24, 24, 0x89800, 3,80, 0x127000, 2,48 , 0x12a000, 3,80},
	{  8, 16, 0x00000, 0, 0, 0x89000,  1, 8 , 0x000000, 0, 0},
};

struct romfont *pen_font;

static int pen_fattr; /* 1: italic 2: bold*/

/* եȥ°ꤷޤ
        size ϡ12,16,24 Τɤ줫
        attr ϡFA_ITALIC  FA_BOLD ξꤷޤ
*/
void set_font(int size, int attr) {
	switch (size) {
	case 12: pen_font = &romfont[0]; break;
	case 16: pen_font = &romfont[1]; break;
	case 24: pen_font = &romfont[2]; break;
	case  7: pen_font = &romfont[3]; break;
	default: 
		if (size == romfont[4].height) {
			pen_font = &romfont[4]; break;
		} else if ( size == romfont[5].height) {
			pen_font = &romfont[5]; break;
		}
		pen_font = &romfont[0]; break;
	}
	pen_fattr = attr;
	return;
}

int get_font_width(void) {
	return pen_font->width;
}

int get_font_height(void) {
	return pen_font->height;
}

/* x,y 顢str ʸ褷ޤ
        ) x,y ϡʸκκɸǤ
*/
void draw_string(int x, int y, char *str) {
	int code;
	while (*str) {
		if (*str & 0x80) {
			if ((unsigned char)*str == 0x8E){
				code = (*str++ << 8) & 0xffff;
				code |= *str++ & 0xff;
				draw_font(x,y,code);
				x += pen_font->width/2;
			}else{
				code = (*str++ << 8) & 0xffff;
				code |= *str++ & 0xff;
				draw_font(x,y,code);
				x += pen_font->width;
			}
		} else {
			draw_font(x,y,*str++ & 0xff);
			x += pen_font->width/2;
		}
	}
	return;
}

/* x,y 顢code ʸ褷ޤ
        ʸɤϡ2Хʸ (EUC1Х * 256 + EUC2Х)
        1Хʸ (<256) Ǥ
        ) x,y ϡʸκκɸǤ
*/
void draw_font(int x, int y, int code) {
	unsigned char *p;
	int xcode;
	int width;
	int height;
	int i,j,k;
	int d = (pen_fattr & FA_ITALIC)?64:0;
	int xx,yy;
#ifdef X_EMU
	int x0 = x;
#endif
	
	x <<= 8;

#if 0
      fprintf(stderr, "start of draw_font(%d, %d, %x)\r\n", x, y, code);
#endif
	if (code < 0x100) {
		p = font_base + pen_font->aoffset + pen_font->abytes * code;
		height = pen_font->height;
		width = pen_font->awidth;
	} else if (((code >> 8) & 0xff) == 0x8E) {
		xcode = (code & 0xff);
		p = font_base + pen_font->aoffset + pen_font->abytes * xcode;
		height = pen_font->height;
		width = pen_font->awidth;
	} else {
		xcode = (((code >> 8) & 0x7f) - 0x21) * 96 
			+ ((code & 0x7f) - 0x20);
		p = font_base + pen_font->koffset + pen_font->kbytes * xcode;
		height = pen_font->height;
		width = pen_font->kwidth;
	}
	if (width == 0) return;
	busy++;
#ifdef X_EMU
	rec_point(x0,y);
	rec_point(x0+pen_font->width-1,y+height-1);
#endif

	for (i=0; i< height; i++) {
		yy = y+i;
		for (j=0; j < width; j++,p++) {
			for (k=0; k<8; k++) {
				if (*p & (1<<(7-k))) {
					xx = ((x+d*(height-1-i))>>8)+j*8+k;
					_draw_pixel(xx,yy);
					if (pen_fattr & FA_BOLD)
						_draw_pixel(xx+1,yy);
					
				}
			}
		}
	}
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

/* x,y 顢code ʸ褷ޤ(åԥդ)
        ʸɤϡ2Хʸ (EUC1Х * 256 + EUC2Х)
        1Хʸ (<256) Ǥ
        ) x,y ϡʸκκɸǤ
*/
void draw_font_clipping(int x, int y, int code, int x1, int y1, int x2, int y2) {
	char *p;
	int xcode;
	int width;
	int height;
	int i,j,k;
	int d = (pen_fattr & FA_ITALIC)?64:0;
	int xx,yy;
	int x0 = x;
	
	x <<= 8;

	if (code < 0x100) {
		p = font_base + pen_font->aoffset + pen_font->abytes * code;
		height = pen_font->height;
		width = pen_font->awidth;
	} else if (((code >> 8) & 0xff) == 0x8E) {
		xcode = (code & 0xff);
		p = font_base + pen_font->aoffset + pen_font->abytes * xcode;
		height = pen_font->height;
		width = pen_font->awidth;
	} else {
		xcode = (((code >> 8) & 0x7f) - 0x21) * 96 
			+ ((code & 0x7f) - 0x20);
		p = font_base + pen_font->koffset + pen_font->kbytes * xcode;
		height = pen_font->height;
		width = pen_font->kwidth;
	}
	if (width == 0) return;

	busy++;
#ifdef X_EMU
	rec_point(x0,y);
	rec_point(x0+pen_font->width-1,y+height-1);
#endif
	if (x1 <= x0 && y1 <= y
	    && x0 + pen_font->width < x2 && y + height < y2){
	    /* NO CLIPPING */
	    for (i = 0; i < height; i++){
		yy = y + i;
		for (j = 0; j < width; j++, p++){
		    for (k = 0; k < 8; k++){
			if (*p & (1<<(7-k))){
			    xx = ((x+d*(height-1-i))>>8)+j*8+k;
			    _draw_pixel(xx,yy);
			    if (pen_fattr & FA_BOLD)
				_draw_pixel(xx+1,yy);
			}
		    }
		}
	    }	
	}else{
	    /* CLIPPING */
	    for (i = 0; i < height; i++){
		yy = y + i;
		for (j = 0; j < width; j++, p++){
		    if (y1 <= yy && yy < y2){
			for (k = 0; k < 8; k++){
			    if (*p & (1<<(7-k))){
				xx = ((x+d*(height-1-i))>>8)+j*8+k;
				if (x1 <= xx && xx < x2)
				    _draw_pixel(xx,yy);
				if (pen_fattr & FA_BOLD)
				    if (x1 <= xx+1 && xx+1 < x2)
					_draw_pixel(xx+1,yy);
			    }
			}
		    }
		}
	    }	
	}
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

/* x,y 顢code ʸ褷ޤ
        ʸɤϡ2Хʸ (EUC1Х * 256 + EUC2Х)
        1Хʸ (<256) Ǥ
        ) x,y ϡʸκκɸǤ
*/
void draw_font_rotated(int x, int y, int code, int direction) {
	char *p;
	int xcode;
	int width;
	int height;
	int i,j,k;
	int d = (pen_fattr & FA_ITALIC)?64:0;
	int xx,yy;
	int x2,y2;
	int x0 = x;
	
	x <<= 8;

	if (code < 0x100) {
		p = font_base + pen_font->aoffset + pen_font->abytes * code;
		height = pen_font->height;
		width = pen_font->awidth;
	} else if (((code >> 8) & 0xff) == 0x8E) {
		xcode = (code & 0xff);
		p = font_base + pen_font->aoffset + pen_font->abytes * xcode;
		height = pen_font->height;
		width = pen_font->awidth;
	} else {
		xcode = (((code >> 8) & 0x7f) - 0x21) * 96 
			+ ((code & 0x7f) - 0x20);
		p = font_base + pen_font->koffset + pen_font->kbytes * xcode;
		height = pen_font->height;
		width = pen_font->kwidth;
	}
	if (width == 0) return;

	busy++;

	switch (direction){
	    default:
	    case DIR_NORTH:
#ifdef X_EMU
	    rec_point(x0, y);
	    rec_point(x0 + pen_font->width - 1, y + height - 1);
#endif
	    for (i = 0; i < height; i++) {
		yy = y + i;
		for (j = 0; j < width; j++, p++){
		    for (k = 0; k < 8; k++) {
			if (*p & (1<<(7-k))){
			    xx = ((x+d*(height-1-i))>>8)+j*8+k;
			    _draw_pixel(xx,yy);
			    if (pen_fattr & FA_BOLD)
				_draw_pixel(xx+1,yy);
			}
		    }
		}
	    }
	    break;
	    case DIR_SOUTH:
#ifdef X_EMU
	    rec_point(x0, y);
	    rec_point(x0 + pen_font->width - 1, y + height - 1);
#endif
	    for (i = 0; i < height; i++) {
		yy = y + height - i;
		for (j = 0; j < width; j++, p++){
		    for (k = 0; k < 8; k++) {
			if (*p & (1<<(7-k))){
			    xx = ((x+d*(height-1-i))>>8)+j*8+k;
			    xx = x0 + pen_font->width - (xx - x0);
			    _draw_pixel(xx,yy);
			    if (pen_fattr & FA_BOLD)
				_draw_pixel(xx-1,yy);
			}
		    }
		}
	    }
	    break;
	    case DIR_WEST:
#ifdef X_EMU
	    rec_point(x0, y);
	    rec_point(x0 + height - 1, y + pen_font->width - 1);
#endif
	    for (i = 0; i < height; i++) {
		yy = y + i;
		x2 = x0 + (yy - y);
		for (j = 0; j < width; j++, p++){
		    for (k = 0; k < 8; k++) {
			if (*p & (1<<(7-k))){
			    xx = ((x+d*(height-1-i))>>8)+j*8+k;
			    y2 = y + pen_font->width - (xx - x0);
			    _draw_pixel(x2, y2);
			    if (pen_fattr & FA_BOLD)
				_draw_pixel(x2, y2-1);
			}
		    }
		}
	    }
	    break;
	    case DIR_EAST:
#ifdef X_EMU
	    rec_point(x0, y);
	    rec_point(x0 + height - 1, y + pen_font->width - 1);
#endif
	    for (i = 0; i < height; i++) {
		yy = y + i;
		x2 = x0 + height - (yy - y);
		for (j = 0; j < width; j++, p++){
		    for (k = 0; k < 8; k++) {
			if (*p & (1<<(7-k))){
			    xx = ((x+d*(height-1-i))>>8)+j*8+k;
			    y2 = y + (xx - x0);
			    _draw_pixel(x2, y2);
			    if (pen_fattr & FA_BOLD)
				_draw_pixel(x2, y2+1);
			}
		    }
		}
	    }
	    break;
	}
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

/* dst  dx,dy  src  sx,sy  xsize,ysize Υ᡼
        žޤ

        op ˡBLT_TILING ꤹȡsrc ϡ󥰤줿
        ̵¤礭ʥ꡼ȸʤޤ

        op  BLT_MASKING+ ꤹ ꤷƩȤ
        ޤ

        BLT_TILING  BLT_MASKING ꤻdx,sxХȶ
	(ɥå)ˤƱ֤ξ˹®ޤ
	(㡧ĥ롤ư̤ܿβ)

        dst  NULL ꤹȡ˻ѤƤ륹꡼
        ꤵ줿ΤȤޤ

        src  NULL ꤹȡ˻ѤƤ륹꡼
        ꤵ줿ΤȤޤ
*/
void bitblt(struct screen *dst, int dx, int dy,
            struct screen *src, int sx, int sy, int xsize, int ysize, int op) {
	int x,y,col;
	int i,j;
	char *p,*q;

	busy++;
	if (src == NULL) src = current_screen;
	if (dst != NULL) {
		push_screen(dst);
	}
	if (xsize < 0 || ysize < 0) {
		xsize = src->width - sx;
		ysize = src->height - sy;
	}
#ifdef X_EMU
	rec_point(dx,dy);
	rec_point(dx+xsize-1,dy+ysize-1);
#endif
	if (!(sx & 3) && !(dx & 3) && !(xsize & 3)
		&& !(op & (BLT_TILING|BLT_MASKING)) ) {
		if (src == current_screen
		     && (dy > sy || (dy == sy && dx > sx))) {
		  for (i=ysize-1; i>=0; --i) {
	 	    p = (char *)src->bitmap + (sy + i) * src->wbytes + (sx>>2);
	 	    q = (char *)current_screen->bitmap 
			+ (dy + i) * current_screen->wbytes + (dx>>2);
		    memcpy(q,p,xsize>>2);
		  }
		} else {
		  for (i=0; i<ysize; i++) {
	 	    p = (char *)src->bitmap + (sy + i) * src->wbytes + (sx>>2);
	 	    q = (char *)current_screen->bitmap 
			+ (dy + i) * current_screen->wbytes + (dx>>2);
		    memcpy(q,p,xsize>>2);
		  }
		}
		goto ret;
	}

	if ((sx & 3) == (dx & 3) && xsize >= 4
	    && !(op & (BLT_TILING|BLT_MASKING))){
	    unsigned int head_mask[] = { 0, 3<<6, 3<<6|3<<4, 3<<6|3<<4|3<<2 };
	    unsigned int tail_mask[] = { 0, 3<<0, 3<<2|3<<0, 3<<4|3<<2|3<<0 };
	    int headlen = (sx & 3) ? 4 - (sx & 3): 0;		/* in bit */
	    int taillen = (sx + xsize) & 3;   			/* in bit */
	    int bodylen = (xsize - headlen - taillen)>>2;	/* in byte */
	    unsigned int headmask = head_mask[headlen];
	    unsigned int tailmask = tail_mask[taillen];
	    if (src == current_screen && (dy > sy || (dy == sy && dx > sx))){
		/* bottom to top */
		for (i = ysize - 1; i >= 0; --i){
		    /* head to tail */
	 	    p = (char *)src->bitmap + (sy + i) * src->wbytes + ((sx + xsize)>>2);
	 	    q = (char *)current_screen->bitmap + (dy + i) * current_screen->wbytes + ((dx + xsize)>>2);
		    /* tail */
		    if (tailmask){
			*q = (*q & ~tailmask) | (*p & tailmask);
		    }
		    p -= bodylen;
		    q -= bodylen;
		    /* body */
		    memcpy(q, p, bodylen);
		    p--;
		    q--;
		    /* head */
		    if (headmask){
			*q = (*q & ~headmask) | (*p & headmask);
		    }
		}
	    }else{
		/* top to bottom */
		for (i = 0; i < ysize; i++){
		    /* head to tail */
	 	    p = (char *)src->bitmap + (sy + i) * src->wbytes + (sx>>2);
	 	    q = (char *)current_screen->bitmap + (dy + i) * current_screen->wbytes + (dx>>2);
		    /* head */
		    if (headmask){
			*q = (*q & ~headmask) | (*p & headmask);
			q++;
			p++;
		    }
		    /* body */
		    memcpy(q, p, bodylen);
		    p += bodylen;
		    q += bodylen;
		    /* tail */
		    if (tailmask){
			*q = (*q & ~tailmask) | (*p & tailmask);
		    }
		}
	    }
	    goto ret;
	}

	if ((op & BLT_TILING) == 0) {
		if (src->width  - sx < xsize) xsize = src->width  - sx;
		if (src->height - sy < ysize) ysize = src->height - sy;
	}

	if (src == current_screen && (dy > sy || (dy == sy && dx > sx))) {
	    for (i=ysize-1; i>=0; --i) {
		y = (sy + i) % src->height;
		for (j=xsize-1; j>=0; --j) {
			x = (sx + j) % src->width;
			p = (char *)src->bitmap
				+ y * src->wbytes + (x>>2);
			col = (*p >> (x & 3)*2) & 3;
			if (!((op & BLT_MASKING) && (op & BLT_MASKCOL) == col))
				_draw_pixel_wc(dx+j,dy+i,col);
		}
	    }
	} else {
	    for (i=0; i< ysize; i++) {
		y = (sy + i) % src->height;
		for (j=0; j< xsize; j++) {
			x = (sx + j) % src->width;
			p = (char *)src->bitmap
				+ y * src->wbytes + (x>>2);
			col = (*p >> (x & 3)*2) & 3;
			if (!((op & BLT_MASKING) && (op & BLT_MASKCOL) == col))
				_draw_pixel_wc(dx+j,dy+i,col);
		}
	    }
	}
ret:
	if (dst != NULL) {
		pop_screen();
#ifdef X_EMU
/*
        }else{
            refresh();
*/
#endif /* X_EMU */
	}
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}

/* ꥹ꡼ޤ
        xs,ys ǥӥåñ̤Υꤷޤ
        bitmap ϡNULL ΤȤưޤ
        ) ư줿ȤΤ free_screen  bitmap  free ޤ
*/
struct screen *create_memscreen(int xs, int ys, char *bitmap) {
	struct screen *ret;
	int wb;
	int type = 1;

	busy++;
	wb = (xs+3)/4;
	if (!bitmap) {
		bitmap = malloc(ys * wb);
        	if(!bitmap){
			perror("malloc");
			return (struct screen *)NULL;
        	}
		type |= 2;
	}
	memset(bitmap,0,ys * wb);

	if (s_freelist) {
		ret = s_freelist;
		s_freelist = *(struct screen **)s_freelist;
	} else {
		ret = (struct screen *)malloc(sizeof(*ret));
	}
        if(!ret){
		perror("malloc");
		if (type & 2) free(bitmap);
		return (struct screen *)NULL;
        }

	ret->type = type;
	ret->wbytes = wb;
	ret->width =  xs; 
	ret->height = ys;
	ret->bitmap = bitmap;
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return ret;
}


/* org  x,y ΰ֤顢xs,ys Υ
        ʬ꡼ޤ
        ʬ꡼ϡorg Υӥåȥޥåפ򤽤Τޤ޻ѤޤΤǡ
        org  free ѤǤޤ
*/
struct screen *create_subscreen(struct screen *org, int x, int y,
                                int xs, int ys, struct screen *b) {
	struct screen *ret;

	busy++;
	if (org == NULL) org = current_screen;
	if (b) {
		ret = b;
		ret->type = 0;
	} else {
		if (s_freelist) {
			ret = s_freelist;
			s_freelist = *(struct screen **)s_freelist;
		} else {
			ret = (struct screen *)malloc(sizeof(*ret));
		}
        	if(!ret){
			perror("malloc");
			return (struct screen *)NULL;
        	}
		ret->type = 1;
	}
	ret->wbytes = org->wbytes;
	ret->width = xs;
	ret->height = ys;
	ret->bitmap = org->bitmap + y * org->wbytes + x/4;
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return ret;
}

/* create_memscreen create_subscreen Ǻ꡼
        free ޤ
*/
void free_screen(struct screen *s) {
	if (s->type & 2) free(s->bitmap);
	if (s->type & 1) {
		*(struct screen **)s = s_freelist;
		s_freelist = s;
	}
	return;
}

static struct screen *screen_stack[32];
static int screen_stack_pos;

/* 褹륹꡼ꤹȤȤˡΥ꡼򥻡֤ޤ */
void push_screen(struct screen *s) {
	busy++;
	screen_stack[screen_stack_pos++] = current_screen;
	current_screen = s;
	return;
}

/* 褹륹꡼򸵤ᤷޤ */
void pop_screen(void) {
	current_screen = screen_stack[--screen_stack_pos];
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return;
}


int get_pixel(int x, int y) {
	char *p;
	int ret;

	if (x < 0 || x >= current_screen->width
	   || y < 0 || y >= current_screen->height) return 0;
	busy++;
	p = (char *)current_screen->bitmap
		+ y * current_screen->wbytes + (x>>2);
	ret = (*p >> ((x & 3)*2) ) & 3;
	if (--busy -bgmode == 0 && switchreq) screen_switch();
	return ret;
}

struct textscreen *create_textscreen(struct screen *s, int x,
                                     int y, int xs, int ys, int attr) {
	struct textscreen *t;
	if (t_freelist) {
		t = t_freelist;
		t_freelist = *(struct textscreen **)t_freelist;
	} else {
		t = (struct textscreen *)malloc(sizeof(*t));
	}
       	if(!t){
		perror("malloc");
		return (struct textscreen *)NULL;
       	}
	t->attr = attr;
	t->s = s;
	t->x = x;
	t->y = y;
	t->xs = xs;
	t->ys = ys;
	t->xpos = 0;
	t->ypos = 0;
	t->lh = 0;
	t->bgcolor = COLOR_WHITE;
	if (t->attr & TS_SAVE) {
		t->save = create_memscreen(xs,ys,NULL);
		bitblt(t->save,0,0,s,x,y,xs,ys,0);
	} else {
		t->save = NULL;
	}
	return t;
}

void free_textscreen(struct textscreen *t) {
	if (t->save) {
		bitblt(t->s,t->x,t->y,t->save,0,0,t->xs,t->ys,0);
		free_screen(t->save);
	}
	*(struct textscreen **)t = t_freelist;
        t_freelist = t;
	return;
}

static void ts_scroll(struct screen *t, int dy) {
	int c;
	bitblt(NULL,0,0,NULL,0,dy,t->width,t->height-dy,0);
	c = pen_color;
	set_color(COLOR_WHITE);
	fill_rect(0,t->height-dy,t->width,dy);
	set_color(c);
	return;
}


void ts_clear(struct textscreen *t) {
	int c;
	struct screen *ss;

	ss = create_subscreen(t->s,t->x,t->y,t->xs,t->ys,NULL);
	push_screen(ss);
	c = pen_color;
	set_color(t->bgcolor);
	clear_screen();
	if ((t->attr & TS_BORDER ) && (t->attr & TS_BLINE)) {
		int d=2;
		set_color(COLOR_BLACK);
		draw_rect(d,d,t->xs-d*2,t->ys-d*2);
	}
	set_color(c);
	pop_screen();
	return;
}

void ts_put_string(struct textscreen *t, char *str, int op) {
	struct screen *ss;
	int code;
	int w,h;
	int d=0;
	if (t->attr & TS_BORDER) d = 4;
	ss = create_subscreen(t->s,t->x+d,t->y+d,t->xs-d*2,t->ys-d*2,NULL);
	push_screen(ss);
	h = pen_font->height;
	while (*str) {
		if (*str & 0x80) {
			if ((unsigned char)*str == 0x8E){
				code = (*str++ << 8) & 0xffff;
				code |= *str++ & 0xff;
				w = pen_font->height/2;
			}else{
				code = (*str++ << 8) & 0xffff;
				code |= *str++ & 0xff;
				w = pen_font->height;
			}
		} else {
			code = *str++;
			w = pen_font->height/2;
		}
		if (t->xpos + w > ss->width || code == '\n') {
			t->xpos = 0;
			if (h > t->lh) t->lh = h;
			t->ypos += t->lh;
			t->lh = 0;
		}
		if (code >= 0x20) {
			if (t->ypos + h > ss->height) {
				ts_scroll(ss,t->ypos + h - ss->height);
				t->ypos = ss->height - h;
			}
			if (op==0) {
				int c = pen_color;
				int hh = h;
				if (t->lh > hh) hh = t->lh;
				set_color(t->bgcolor);
				fill_rect(t->xpos,t->ypos,w,hh);
				set_color(c);
			}
			draw_font(t->xpos,t->ypos,code);
			t->xpos += w;
			if (h > t->lh) t->lh = h;
		}
	}
	pop_screen();
	return;
}

void ts_goto(struct textscreen *t, int x, int y) {
	t->xpos = x;
	t->ypos = y;
	t->lh = 0;
	return;
}

void ts_set_bgcolor(struct textscreen *t, int c) {
	t->bgcolor = c;
	return;
}

void ts_put_image(struct textscreen *t, struct screen *s,
                  int sx, int sy, int w, int h, int op) {
	struct screen *ss;
	int d=0;

	if (t->attr & TS_BORDER) d = 4;
	ss = create_subscreen(t->s,t->x+d,t->y+d,t->xs-d*2,t->ys-d*2,NULL);
	push_screen(ss);

	if (w < 0) w = s->width - sx;
	if (h < 0) h = s->height - sy;
	if (w > ss->width) w = ss->width;
	if (h > ss->height) h = ss->height;

	if (t->xpos + w > ss->width) {
		t->xpos = 0;
		t->ypos += t->lh;
		t->lh = 0;
	}
	if (h > t->lh) t->lh = h;

	if (t->ypos + h > ss->height) {
		ts_scroll(ss,t->ypos + h - ss->height);
		t->ypos = ss->height - h;
	}
	if (op==0 && h<t->lh) {
		int c = pen_color;
		int hh = h;
		if (t->lh > hh) hh = t->lh;
		set_color(COLOR_WHITE);
		fill_rect(t->xpos,t->ypos,w,hh);
		set_color(c);
	}
	bitblt(NULL,t->xpos,t->ypos,s,sx,sy,w,h,op);
	t->xpos += w;
	pop_screen();
	return;
}

void write_screen_xpm(char *name, struct screen *s) {
	int x,y,c;
	FILE *fp;
	char buf[256];

	strcpy(buf,name);
	strcat(buf,".xpm");
	fp = fopen(buf,"w");

	if (s == NULL) s = current_screen;
	push_screen(s);
	fprintf(fp,"/* XPM */\n");
	fprintf(fp,"static char *%s_xpm[] = {\n",name);
	fprintf(fp,"\"%d %d %d 1\",\n",s->width,s->height,4);
	fprintf(fp,"\"  c #ffffff\",\n");
	fprintf(fp,"\". c #7f7f7f\",\n");
	fprintf(fp,"\"+ c #3f3f3f\",\n");
	fprintf(fp,"\"# c #000000\",\n");
	for (y=0; y< s->height; y++) {
		fprintf(fp,"\"");
		for (x=0; x<s->width; x++) {
			switch(get_pixel(x,y)) {
			case 0: c = ' '; break;
			case 1: c = '.'; break;
			case 2: c = '+'; break;
			case 3: c = '#'; break;
			default: c = ' '; break;
			}
			fprintf(fp,"%c",c);
		}
		fprintf(fp,"\",\n");
	}
	fprintf(fp,"};\n");
	pop_screen();
	fclose(fp);
	return;
}

void write_screen_native(char *name, struct screen *s) {
	int y,i;
	FILE *fp;
	char buf[256];

	strcpy(buf,name);
	strcat(buf,".h");
	fp = fopen(buf,"w");

	if (s == NULL) s = current_screen;
	push_screen(s);
	fprintf(fp,"static struct screen %s = {\n",name);
	fprintf(fp,"%d,%d,%d,%d,\n",0,s->wbytes,s->width,s->height);
	for (y=0; y< s->height; y++) {
		fprintf(fp,"\"");
		for (i=0;i<s->wbytes;i++) {
			fprintf(fp,"\\x%02x",s->bitmap[s->wbytes*y+i] & 0xff);
		}
		fprintf(fp,"\"\n");
	}
	fprintf(fp,"};\n");
	pop_screen();
	fclose(fp);
	return;
}
#ifdef X_EMU
int font_base_create(void) {
int    i, k;
int    code;
int    fsize[4] = {12,16,24,7};
#ifdef LOCALE
int    num_missing_charsets;
char   **missing_charsets;
char   *default_string;
#endif
unsigned char   str[2];

    fprintf(stderr, "generating fonts for mgl. wait a minute.\n");

    font_base = (char *)malloc(0x200000);
    if(!font_base){
        perror("malloc");
        return 0;
    }
    x_font_alloc = 1;
    memset(font_base, 0x00, 0x200000);

#ifdef LOCALE
    /***** locale *****/
    if(setlocale(LC_ALL, "") == NULL){
        fprintf(stderr, "setlocale failed \n");
        return 0;
    }

    /**** 8dotեȥåȤκ ****/
    fontset8 = XCreateFontSet(display,
                              FONT_8,
                              &missing_charsets,
                              &num_missing_charsets,
                              &default_string);
    if(fontset8 == NULL){
        fprintf(stderr, "8Dot FontSet can't created\n");
        return 0;
    }

    /**** 12dotեȥåȤκ ****/
    fontset12 = XCreateFontSet(display,
                               FONT_12,
                               &missing_charsets,
                               &num_missing_charsets,
                               &default_string);
    if(fontset12 == NULL){
        fprintf(stderr, "12Dot FontSet can't created\n");
        return 0;
    }

    /**** 16dotեȥåȤκ ****/
    fontset16 = XCreateFontSet(display,
                               FONT_16,
                               &missing_charsets,
                               &num_missing_charsets,
                               &default_string);
    if(fontset16 == NULL){
        fprintf(stderr, "16Dot FontSet can't created\n");
        return 0;
    }

    /**** 24dotեȥåȤκ ****/
    fontset24 = XCreateFontSet(display,
                               FONT_24,
                               &missing_charsets,
                               &num_missing_charsets,
                               &default_string);
    if(fontset24 == NULL){
        fprintf(stderr, "24Dot FontSet can't created\n");
        return 0;
    }
#else  /* non LOCALE */
    fs8a = XLoadQueryFont(display, FONT_8A);
    if(fs8a == NULL){
        fprintf(stderr, "8Dot Ascii Font can't created\n");
        return 0;
    }
    font8a = fs8a->fid;
    fs12a = XLoadQueryFont(display, FONT_12A);
    if(fs12a == NULL){
        fprintf(stderr, "12Dot Ascii Font can't created\n");
        return 0;
    }
    font12a = fs12a->fid;
    fs16a = XLoadQueryFont(display, FONT_16A);
    if(fs16a == NULL){
        fprintf(stderr, "16Dot Ascii Font can't created\n");
        return 0;
    }
    font16a = fs16a->fid;
    fs24a = XLoadQueryFont(display, FONT_24A);
    if(fs24a == NULL){
        fprintf(stderr, "24Dot Ascii Font can't created\n");
        return 0;
    }
    font24a = fs24a->fid;

#if 0
    fs8 = XLoadQueryFont(display, FONT_8);
    if(fs8 == NULL){
        fprintf(stderr, "8Dot Font can't created\n");
        return 0;
    }
    font8 = fs8->fid;
#endif
    fs12 = XLoadQueryFont(display, FONT_12);
    if(fs12 == NULL){
        fprintf(stderr, "12Dot Font can't created\n");
        return 0;
    }
    font12 = fs12->fid;
    fs16 = XLoadQueryFont(display, FONT_16);
    if(fs16 == NULL){
        fprintf(stderr, "16Dot Font can't created\n");
        return 0;
    }
    font16 = fs16->fid;
    fs24 = XLoadQueryFont(display, FONT_24);
    if(fs24 == NULL){
        fprintf(stderr, "24Dot Font can't created\n");
        return 0;
    }
    font24 = fs24->fid;

#if 0
    fs8k = XLoadQueryFont(display, FONT_8K);
    if(fs8k == NULL){
        fprintf(stderr, "8Dot JISX201 Font can't created\n");
        return 0;
    }
    font8k = fs8k->fid;
#endif
    fs12k = XLoadQueryFont(display, FONT_12K);
    if(fs12k == NULL){
        fprintf(stderr, "12Dot JISX201 Font can't created\n");
        return 0;
    }
    font12k = fs12k->fid;
    fs16k = XLoadQueryFont(display, FONT_16K);
    if(fs16k == NULL){
        fprintf(stderr, "16Dot JISX201 Font can't created\n");
        return 0;
    }
    font16k = fs16k->fid;
    fs24k = XLoadQueryFont(display, FONT_24K);
    if(fs24k == NULL){
        fprintf(stderr, "24Dot JISX201 Font can't created\n");
        return 0;
    }
    font24k = fs24k->fid;
#endif /* LOCALE */
    fg_gc = XCreateGC(display, main_win, 0, 0);
    XSetForeground(display, fg_gc, BlackPixel(display, DefaultScreen(display)));

    bg_gc = XCreateGC(display, main_win, 0, 0);
    XSetForeground(display, bg_gc, WhitePixel(display, DefaultScreen(display)));

    /* pixmap */
    pixmap = XCreatePixmap(display, DefaultRootWindow(display),
                           PIXMAP_W, PIXMAP_H,
                           DefaultDepth(display, DefaultScreen(display)));
    if(!pixmap){
        fprintf(stderr, "Pixmap can't created\n");
        return 0;
    }

    for(k = 0 ; k < 3 ; k++){
        set_font(fsize[k], 0);

        /* font_baseؤδΥ */
        /* a1a1 - a1fe */
        for(str[0] = 0xa1, str[1] = 0xa1 ; str[1] < 0xff ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2a1 - a2ae */
        for(str[0] = 0xa2, str[1] = 0xa1 ; str[1] < 0xaf ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2ba - a2c1 */
        for(str[0] = 0xa2, str[1] = 0xba ; str[1] < 0xc2 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2ca - a2d0 */
        for(str[0] = 0xa2, str[1] = 0xca ; str[1] < 0xd1 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2dc - a2ea */
        for(str[0] = 0xa2, str[1] = 0xdc ; str[1] < 0xeb ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2f2 - a2f9 */
        for(str[0] = 0xa2, str[1] = 0xf2 ; str[1] < 0xfa ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a2fe */
        str[0] = 0xa2;
        str[1] = 0xfe;
        code = (str[0] << 8) & 0xffff;
        code |= str[1] & 0xff;
        if(xset_font((char *)str, 2, code) == 0){
            return 0;
        }
        /* a3b0 - a3b9 */
        for(str[0] = 0xa3, str[1] = 0xb0 ; str[1] < 0xba ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a3c1 - a3da */
        for(str[0] = 0xa3, str[1] = 0xc1 ; str[1] < 0xdb ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a3e1 - a3fa */
        for(str[0] = 0xa3, str[1] = 0xe1 ; str[1] < 0xfb ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a4a1 - a4f3 */
        for(str[0] = 0xa4, str[1] = 0xa1 ; str[1] < 0xf4 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a5a1 - a5f6 */
        for(str[0] = 0xa5, str[1] = 0xa1 ; str[1] < 0xf7 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a6a1 - a6b8 */
        for(str[0] = 0xa6, str[1] = 0xa1 ; str[1] < 0xb9 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a6c1 - a6d8 */
        for(str[0] = 0xa6, str[1] = 0xc1 ; str[1] < 0xd9 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a7a1 - a7c1 */
        for(str[0] = 0xa7, str[1] = 0xa1 ; str[1] < 0xc2 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a7d1 - a7f1 */
        for(str[0] = 0xa7, str[1] = 0xd1 ; str[1] < 0xf2 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /* a8a1 - a8c0 */
        for(str[0] = 0xa8, str[1] = 0xa1 ; str[1] < 0xc0 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }

        /*
           b0a1 - b0fe,
           ...
           cea1 - cefe,
        */
        for(str[0] = 0xb0 ; str[0] < 0xcf ; str[0]++){
            for(str[1] = 0xa1 ; str[1] < 0xff ; str[1]++){
                code = (str[0] << 8) & 0xffff;
                code |= str[1] & 0xff;
                if(xset_font((char *)str, 2, code) == 0){
                    return 0;
                }
            }
        }
        /*
           cfa1 - cfd3,   -
        */
        for(str[0] = 0xcf, str[1] = 0xa1 ; str[1] < 0xd4 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /*
           d0a1 - d0fe,
           d1a1 - d1fe,
           ...
           f3a1 - f3fe,
        */
        for(str[0] = 0xd0 ; str[0] < 0xf4 ; str[0]++){
            for(str[1] = 0xa1 ; str[1] < 0xff ; str[1]++){
                code = (str[0] << 8) & 0xffff;
                code |= str[1] & 0xff;
                if(xset_font((char *)str, 2, code) == 0){
                    return 0;
                }
            }
        }
        /*
           f4a1 - f4a4    -
        */
        for(str[0] = 0xf4, str[1] = 0xa1 ; str[1] < 0xa5 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }
        /*
           ȾҲ̾
           8ea1 - 8edf
        */
        for(str[0] = 0x8e, str[1] = 0xa1 ; str[1] < 0xe0 ; str[1]++){
            code = (str[0] << 8) & 0xffff;
            code |= str[1] & 0xff;
            if(xset_font((char *)str, 2, code) == 0){
                return 0;
            }
        }

        /* font_baseؤasciiΥ */
        /* 20 - 7f */
        for(i = 0x20 ; i < 0x7f; i++){
            str[0] = i;
            str[1] = 0x00;
            code = str[0] & 0xff;
            if(xset_font((char *)str, 1, code) == 0){
                return 0;
            }
        }

        /* font_baseؤglyph ? Υ */
    }

    set_font(fsize[3], 0);
    /* font_baseؤ8dot asciiΥ */
    /* 20 - 7f */
    for(i = 0x20 ; i < 0x7f; i++){
        str[0] = i;
        str[1] = 0x00;
        code = str[0] & 0xff;
        if(xset_font((char *)str, 1, code) == 0){
            return 0;
        }
    }

    set_font(fsize[0], 0);

#ifdef LOCALE
    XFreeFontSet(display, fontset8);
    XFreeFontSet(display, fontset12);
    XFreeFontSet(display, fontset16);
    XFreeFontSet(display, fontset24);
#else  /* non LOCALE */
    XFreeFont(display, fs8a);
    XFreeFont(display, fs12a);
    XFreeFont(display, fs16a);
    XFreeFont(display, fs24a);
#if 0
    XFreeFont(display, fs8);
#endif
    XFreeFont(display, fs12);
    XFreeFont(display, fs16);
    XFreeFont(display, fs24);
#if 0
    XFreeFont(display, fs8k);
#endif
    XFreeFont(display, fs12k);
    XFreeFont(display, fs16k);
    XFreeFont(display, fs24k);
#endif /* LOCALE */
    XFreeGC(display, fg_gc);
    XFreeGC(display, bg_gc);
    XFreePixmap(display, pixmap);

    if(access(FONTPATH, F_OK) == -1){
        if(errno == ENOENT){
int  fd;
            fd = open(FONTPATH, O_RDWR | O_CREAT, 0644);
            if(fd == -1){
                return 1;
            }
            write(fd, font_base, 0x200000);
            close(fd);
            return 1;
        }
        return 1;
    }

    return 1;
}

int xset_font(char *str, int bytes, int code) {
#ifdef LOCALE
    XFontSet font;
#else  /* non LOCALE */
    XCharStruct ts;
    int         dir;
    int         as;
    int         de;
#endif /* LOCALE */
    XImage   *image;
    int      width;
    int      height;
    int      sx, sy;
    int      i, j, k;
    int      xcode = code;
    char     *p;
    char     *q;

    if(pen_font == &romfont[0]){
        sx = 0;
#ifdef LOCALE
        sy = 12;
        font = fontset12;
#else  /* non LOCALE */
        if(bytes == 1){
            XSetFont(display, fg_gc, font12a);
            XTextExtents(fs12a, str, bytes, &dir, &as, &de, &ts);
        }else{
            if (((code >> 8) & 0xff) == 0x8E) {
                XSetFont(display, fg_gc, font12k);
                XTextExtents16(fs12k, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            } else {
                XSetFont(display, fg_gc, font12);
                XTextExtents16(fs12, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            }
        }
        sy = as;
#endif /* LOCALE */
    }else
    if(pen_font == &romfont[1]){
        sx = 0;
#ifdef LOCALE
        sy = 16;
        font = fontset16;
#else  /* non LOCALE */
        if(bytes == 1){
            XSetFont(display, fg_gc, font16a);
            XTextExtents(fs16a, str, bytes, &dir, &as, &de, &ts);
        }else{
            if (((code >> 8) & 0xff) == 0x8E) {
                XSetFont(display, fg_gc, font16k);
                XTextExtents16(fs16k, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            } else {
                XSetFont(display, fg_gc, font16);
                XTextExtents16(fs16, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            }
        }
        sy = as;
#endif /* LOCALE */
    }else
    if(pen_font == &romfont[2]){
        sx = 0;
#ifdef LOCALE
        sy = 24;
        font = fontset24;
#else  /* non LOCALE */
        if(bytes == 1){
            XSetFont(display, fg_gc, font24a);
            XTextExtents(fs24a, str, bytes, &dir, &as, &de, &ts);
        }else{
            if (((code >> 8) & 0xff) == 0x8E) {
                XSetFont(display, fg_gc, font24k);
                XTextExtents16(fs24k, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            } else {
                XSetFont(display, fg_gc, font24);
                XTextExtents16(fs24, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            }
        }
        sy = as;
#endif /* LOCALE */
    }else
    if(pen_font == &romfont[3]){
        sx = 0;
#ifdef LOCALE
        sy = 8;
        font = fontset8;
#else  /* non LOCALE */
        if(bytes == 1){
            XSetFont(display, fg_gc, font8a);
            XTextExtents(fs8a, str, bytes, &dir, &as, &de, &ts);
        }else{
            if (((code >> 8) & 0xff) == 0x8E) {
                XSetFont(display, fg_gc, font8k);
                XTextExtents16(fs8k, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            } else {
                XSetFont(display, fg_gc, font8);
                XTextExtents16(fs8, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            }
        }
        sy = as;
#endif /* LOCALE */
    }else{
        sx = 0;
#ifdef LOCALE
        sy = 12;
        font = fontset12;
#else  /* non LOCALE */
        if(bytes == 1){
            XSetFont(display, fg_gc, font12a);
            XTextExtents(fs12a, str, bytes, &dir, &as, &de, &ts);
        }else{
            if (((code >> 8) & 0xff) == 0x8E) {
                XSetFont(display, fg_gc, font12k);
                XTextExtents16(fs12k, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            } else {
                XSetFont(display, fg_gc, font12);
                XTextExtents16(fs12, (XChar2b *)str, bytes, &dir, &as, &de, &ts);
            }
        }
        sy = as;
#endif /* LOCALE */
    }

    /* pixmap򥯥ꥢ */
    XFillRectangle(display, pixmap, bg_gc, 0, 0, PIXMAP_W, PIXMAP_H);

    /* pixmap˥եȤ */
#ifdef LOCALE
    XmbDrawImageString(display, pixmap, font,
                  fg_gc,
                  sx,
                  sy,
                  str, bytes);
#else  /* non LOCALE */
    if(bytes == 1){
        XDrawString(display, pixmap, fg_gc,
                    sx,
                    sy,
                    str, bytes);
    }else{
char dtext[2];

        if (((code >> 8) & 0xff) == 0x8E) {
            dtext[0] = 0x00;
            dtext[1] = str[1] | 0x80;
        } else {
            dtext[0] = str[0] & 0x7f;
            dtext[1] = str[1] & 0x7f;
        }
        XDrawString16(display, pixmap, fg_gc,
                      sx,
                      sy,
                     (XChar2b *)dtext, bytes);
    }
#endif /* LOCALE */

    /* pixmapեȤ */
    image = XGetImage(display, pixmap,
                      0, 0,
                      PIXMAP_W, PIXMAP_H,
                      1, XYPixmap);
    if(!image){
        fprintf(stderr, "Font Image can't get.\n");
        return 0;
    }

    q = image->data;

    /* եȥǡfont_base */
    if (code < 0x100) {
        p = font_base + pen_font->aoffset + pen_font->abytes * code;
        height = pen_font->height;
        width  = pen_font->awidth;
    } else if (((code >> 8) & 0xff) == 0x8E) {
	xcode = (code & 0xff);
	p = font_base + pen_font->aoffset + pen_font->abytes * xcode;
	height = pen_font->height;
	width = pen_font->awidth;
    } else {
        xcode = (((code >> 8) & 0x7f) - 0x21) * 96
               + ((code & 0x7f) - 0x20);
        p = font_base + pen_font->koffset + pen_font->kbytes * xcode;
        height = pen_font->height;
        width  = pen_font->kwidth;
    }

    for (i=0; i< height; i++) {
        /* ä֤ */
        for (j=0; j < width; j++,q++,p++) {
          if (XBitmapBitOrder(display)) {
            *p = ~*q;
          } else {
            for (k=0; k<8; k++) {
                if (!(*q & (1<<(7-k)))) {
                    *p = *p | (1<<k);
                }
            }
          }
        }
        q+=((PIXMAP_W / 8) - width);
    }

    XDestroyImage(image);
    return 1;
}
#endif /* X_EMU */

struct key_buf {
	int code;
	short x,y;
} key_buf[100];

int key_buf_cnt=0;

static int im_mode;

int get_key_im(int time_out) {
	struct romfont *sav_font;
	int sav_fattr;
	int sav_color;
	int ret;
	char buf[256];

retry:
	if (key_buf_cnt) {
		ret = get_key(time_out);
		if (ret == ('O'&0x1f)) im_mode = !im_mode;
		else return ret;
	}
	sav_font = pen_font;
	sav_fattr = pen_fattr;
	sav_color = pen_color;

	if (!im_mode) {
		ret = get_key(time_out);
		if (ret == ('O'&0x1f)) im_mode = !im_mode;
		else return ret;
	}
	if (im_mode && im_readp) {
		int i;
		while ((ret = (im_readp)(buf,256))==0)
			;
		if (ret>0) {
			for (i=0; i< ret; i++) {
				put_key(buf[i]&0xff);
			}
		}
	}
	set_font(sav_font->height,sav_fattr);
	set_color(sav_color);
	ret = get_key(time_out);
	if (ret == ('O'&0x1f)) im_mode = !im_mode;
	else return ret;
	goto retry;
}

void toggle_im_mode(void) {
    put_key('O'&0x1f);
    return;
}

int get_im_mode(void) {
    return im_mode;
}

int get_key(int time_out) {
	fd_set fds;
	int c;
#if 0
	int i;
#endif

	if (key_buf_cnt) {
		c = key_buf[0].code;
		if (key_buf[0].x >= 0) {
			vk_x = key_buf[0].x;
			vk_y = key_buf[0].y;
		}
		memcpy(key_buf,key_buf+1,(key_buf_cnt-1)*sizeof(key_buf[0]));
		key_buf_cnt--;
#if 0
                fprintf(stderr, "0x%04x = get_key(0, %d)\n", c, key_buf_cnt);
#endif
		return c;
	}
	FD_ZERO( &fds );
        FD_SET(0, &fds);
	if (time_out >= 0) {
		time_out *= 100;
	}
	key_select(1, &fds, time_out);
	if (key_buf_cnt) {
		c = key_buf[0].code;
		memcpy(key_buf,key_buf+1,(key_buf_cnt-1)*sizeof(key_buf[0]));
		key_buf_cnt--;
#if 0
                fprintf(stderr, "0x%04x = get_key(1)\n", c);
#endif
		return c;
	}
#if 0
        fprintf(stderr, "unknown get_key()\n");
#endif
	return (-1);
}

extern int whole_screen;

void put_key(int c) {
      if(key_buf_cnt < 100){
	key_buf[key_buf_cnt].code = c;
	key_buf[key_buf_cnt].x = -1;
	key_buf[key_buf_cnt].y = -1;
	key_buf_cnt++;
      }
      return;
}

void put_key_xy(c,x,y) {
	key_buf[key_buf_cnt].code = c;
	key_buf[key_buf_cnt].x = x;
	key_buf[key_buf_cnt].y = y;
	key_buf_cnt++;
}

#ifdef X_EMU /* X_EMU */
static long long millitime(void) {
	struct timeval n;
	gettimeofday(&n,0);
	return (n.tv_sec *1000 + n.tv_usec/1000);
}
#endif /* X_EMU */

#if !defined(X_EMU) || defined(TTY_KEY)
static int getch(void) {
	int ret;
	char buf[10];
	ret = read(0,buf,1);
	if (ret == 1) return (buf[0] & 0xff);
	return (-1);
}

#if defined (__FreeBSD__)
int key_select(int nfds, fd_set *readfds, int timeout) {
    int  ret;
    int c,r;
    struct timeval to,*top;
    fd_set fds;

    if (key_buf_cnt) {
        FD_ZERO(readfds);
        FD_SET(0, readfds);
	return 1;
    }
    if (timeout < 0) {
	top = 0;
    } else {
	top = &to;
	to.tv_sec = timeout / 1000;
	to.tv_usec = (timeout % 1000) * 1000;
    }
    ret = select(nfds, readfds, 0, 0, top);
    if (ret > 0 && FD_ISSET(0 , readfds)) {
	c = getch();
	if (c != 033) {
		put_key(c);
		return ret;
	}
retry:
	FD_ZERO(&fds);
        FD_SET(0, &fds);
	to.tv_sec = 0;
	to.tv_usec = 300000;
    	r = select(1, &fds, 0, 0, &to);
        if (r > 0 && FD_ISSET(0 , &fds)) {
		c = getch();
		if (c == 033) {
			put_key(033);
			goto retry;
		}
		if (c != '[') {
			put_key(033);
			put_key(c);
			return ret;
		}
		c = getch();
		switch (c) {
		case 'M':  c = MK_F1; break;
		case 'N':  c = MK_F2; break;
		case 'O':  c = MK_F3; break;
		case 'P':  c = MK_F4; break;
		case 'Q':  c = MK_F5; break;
		case 'R':  c = MK_F6; break;
		case 'S':  c = MK_F7; break;
		case 'T':  c = MK_F8; break;
		case 'U':  c = MK_F9; break;
		case 'V':  c = MK_F10; break;
		case 'W':  c = MK_F11; break; /* */
		case 'X':  c = MK_F12; break; /* */
		case 'A':  c = MK_UP; break;
		case 'B':  c = MK_DOWN; break;
		case 'C':  c = MK_RIGHT; break;
		case 'D':  c = MK_LEFT; break;
		case 'I':  c = MK_PAGE_UP; break;
		case 'G':  c = MK_PAGE_DOWN; break;
		case 'F':  c = MK_END; break; /* */
		case 'H':  c = MK_HOME; break; /* */
		case 'L':  c = MK_INS; break;
		case 'E':  c = MK_DEL; break;
		default:
			put_key(033);
			put_key('[');
			put_key(c);
			return ret;
		}
	}
	put_key(c);
    }
    return ret;
}
#elif defined (__linux__)
int key_select(int nfds, fd_set *readfds, int timeout) {
    int  ret;
    int c,r;
    struct timeval to,*top;
    fd_set fds;

    if (key_buf_cnt) {
        FD_ZERO(readfds);
        FD_SET(0, readfds);
	return 1;
    }
    if (timeout < 0) {
	top = 0;
    } else {
	top = &to;
	to.tv_sec = timeout / 1000;
	to.tv_usec = (timeout % 1000) * 1000;
    }
    ret = select(nfds, readfds, 0, 0, top);

    if (ret > 0 && FD_ISSET(0 , readfds)) {
	c = getch();
	if (c != 033) {
		put_key(c);
		return ret;
	}
retry:
	do{
	    FD_ZERO(&fds);
	    FD_SET(0, &fds);
	    to.tv_sec = 0;
	    to.tv_usec = 100000;       /* 0.1 sec */
	    r = select(1, &fds, NULL, NULL, &to);
	}while(! (r > 0 && FD_ISSET(0 , &fds)));

	c = getch();
	if (c == 033) {
		put_key(033);
		goto retry;
	}
	if (c != '[') {
		put_key(033);
		put_key(c);
		return ret;
	}
	c = getch();
	switch (c) {
	case 'W':  c = MK_F11; break; /* */
	case 'X':  c = MK_F12; break; /* */
	case 'A':  c = MK_UP; break;
	case 'B':  c = MK_DOWN; break;
	case 'C':  c = MK_RIGHT; break;
	case 'D':  c = MK_LEFT; break;
	case 'I':  c = MK_PAGE_UP; break;
	case 'G':  c = MK_PAGE_DOWN; break;
	case 'F':  c = MK_END; break; /* */
	case 'H':  c = MK_HOME; break; /* */
	case '3':
		c = getch();
		switch(c){
			case '~':  c = MK_DEL; break;
			default:
				put_key(033);
				put_key('[');
				put_key('3');
				break;
		}
		break;
	case '2':
		c = getch();
		switch(c){
			case '0':  c = MK_F9; break;
			case '1':  c = MK_F10; break;
			case '~':  c = MK_INS; break;
			default:
				put_key(033);
				put_key('[');
				put_key('2');
				break;
		}
		break;
	case '1':
		c = getch();
		switch(c){
			case '7':  c = MK_F6; break;
			case '8':  c = MK_F7; break;
			case '9':  c = MK_F8; break;
			default:
				put_key(033);
				put_key('[');
				put_key('1');
				break;
		}
		break;
	case '[':
		c = getch();
		switch(c){
			case 'A':  c = MK_F1; break;
			case 'B':  c = MK_F2; break;
			case 'C':  c = MK_F3; break;
			case 'D':  c = MK_F4; break;
			case 'E':  c = MK_F5; break;
			default:
				put_key(033);
				put_key('[');
				put_key('[');
				break;
		}
		break;
	default:
		put_key(033);
		put_key('[');
	}
    	put_key(c);
    }
    return ret;
}
#endif

#else
static int    _key_select(int nfds, fd_set *readfds, int timeout);

int key_select(nfds, readfds, timeout)
int    nfds;
fd_set *readfds;
int timeout;
{
	int ret;
	fd_set local;

#if 0
	refresh();
#endif
	if (timeout > 0) {
		ret = _key_select(nfds,readfds,timeout);
		return ret;
	}

	for (;;) {
		local = *readfds;
		ret = key_select(nfds,&local,200);
		if (ret != 0) break;
	}
	*readfds = local;
	return ret;
}
int _key_select(int nfds, fd_set *readfds, int timeout) {
    int  xfd = -1;
    int  ret;
    struct timeval to,*top;
    static long long t1,t2;
    fd_set sv_fds = *readfds;
    int             ret_key;
    char            string[10];
#ifdef X_ROTATE
    XSizeHints      xsh;
#endif /* X_ROTATE */
    KeySym          key;
    XComposeStatus  cs;
retry:
    *readfds = sv_fds;
    if (display) {
        xfd = ConnectionNumber(display);
	while (XCheckMaskEvent(display, GET_EVENT,&event)) {
	    int x = event.xbutton.x;
	    int y = event.xbutton.y;
       	    switch(event.type) {
       	    case Expose:
		r_rect = 2;
       		refresh();
		break;
       	    case KeyPress:
#ifdef SOLARIS
                for(ret_key = 0 ; ret_key < 10 ; ret_key++){
                    string[ret_key] = 0x00;
                }
#endif /* SOLARIS */
       		ret_key = XLookupString((XKeyEvent *)&event, string, 1, &key, &cs);
                if(!key_conv(key)){
		    if (string[0]) {
       			put_key(string[0]&0xff);
#ifdef X_CTRL_SPACE
                    }else{
                        if(key == 0x20){
                            put_key(('@' & 0x1f));
                        }
#endif /* X_CTRL_SPACE */
		    }
		}
#ifdef X_ROTATE
		break;
	    case MotionNotify:
		if (!vk_cur) {
		   	if (vk_search(vk_root,x,y)) {
				goto press;
			}
			break;
		}
		if (vk_match(vk_cur,x,y)) {
//printf("Motion: Match\n");
			if (vk_cur->down == MK_V0) {
				vk_reverse(vk_cur,0);
			} else if (vk_cur->move != MK_V0) {
				put_key_xy(vk_cur->move,x,y);
			}
		} else {
			if (vk_cur->down == MK_V0) {
				vk_reverse(vk_cur,1);
			}
		   	if (vk_search(vk_root,x,y)) {
//printf("Motion: UnMatch -- match other \n");
				goto press;
			}
//printf("Motion: UnMatch\n");
		}
		break;
	    case ButtonRelease:
		if (!vk_cur) break;
                if(event.xbutton.button != Button1) {
			break;
		}
//printf("ButtonRelease\n");
		if (vk_cur->down == MK_V0) {
			vk_reverse(vk_cur,1);
			if (vk_match(vk_cur,x,y)) {
				put_key_xy(vk_cur->up,x,y);
			}
		} else {
			put_key_xy(vk_cur->up,x,y);
		}
		vk_cur = NULL;
		break;
       	    case ButtonPress:
                if(event.xbutton.button == Button1) {
press:
		   if (vk_cur) { /* same as release */
			if (vk_cur->down == MK_V0) {
			    vk_reverse(vk_cur,1);
			    if (vk_match(vk_cur,x,y)) {
				put_key_xy(vk_cur->up,x,y);
			    }
			} else {
			    put_key_xy(vk_cur->up,x,y);
			}
			vk_cur = NULL;
		   }
		   if (vk_search(vk_root,event.xbutton.x,event.xbutton.y)) {
			if ((vkp->up == MK_V0)&&(vkp->down != MK_V0)) {
				put_key_xy(vkp->down,x,y);
			} else if ((vkp->up != MK_V0)&&(vkp->down != MK_V0)) {
				vk_cur = vkp;
				put_key_xy(vkp->down,x,y);
			} else if ((vkp->up != MK_V0)&&(vkp->down == MK_V0)) {
				vk_cur = vkp;
				vk_reverse(vk_cur,0);
			}
		   }
		}else
                if(event.xbutton.button == Button2){
                    /*  ->  ->  ->  */
                    x_rotated = (x_rotated + 1) % 4;
                }else
                if(event.xbutton.button == Button3){
                    /*  ->  ->  ->  */
                    x_rotated = (x_rotated + 3) % 4;
                }else{
                    break;
                }
                switch(x_rotated){
                case DIR_NORTH:
                case DIR_SOUTH:
                    xsh.flags  = (PBaseSize | USSize | PMinSize | PMaxSize);
                    xsh.width       = SCREEN_WIDTH;
                    xsh.height      = SCREEN_HEIGHT;
                    xsh.base_width  = SCREEN_WIDTH;
                    xsh.base_height = SCREEN_HEIGHT;
                    xsh.min_width   = SCREEN_WIDTH;
                    xsh.min_height  = SCREEN_HEIGHT;
                    xsh.max_width   = SCREEN_WIDTH;
                    xsh.max_height  = SCREEN_HEIGHT;
                    break;
                case DIR_WEST:
                case DIR_EAST:
                    xsh.flags  = (PBaseSize | USSize | PMinSize | PMaxSize);
                    xsh.width       = SCREEN_HEIGHT;
                    xsh.height      = SCREEN_WIDTH;
                    xsh.base_width  = SCREEN_HEIGHT;
                    xsh.base_height = SCREEN_WIDTH;
                    xsh.min_width   = SCREEN_HEIGHT;
                    xsh.min_height  = SCREEN_WIDTH;
                    xsh.max_width   = SCREEN_HEIGHT;
                    xsh.max_height  = SCREEN_WIDTH;
                    break;
                default:
                    x_rotated = DIR_NORTH;
                    xsh.flags  = (PBaseSize | USSize | PMinSize | PMaxSize);
                    xsh.width       = SCREEN_WIDTH;
                    xsh.height      = SCREEN_HEIGHT;
                    xsh.base_width  = SCREEN_WIDTH;
                    xsh.base_height = SCREEN_HEIGHT;
                    xsh.min_width   = SCREEN_WIDTH;
                    xsh.min_height  = SCREEN_HEIGHT;
                    xsh.max_width   = SCREEN_WIDTH;
                    xsh.max_height  = SCREEN_HEIGHT;
                    break;
                }
                XSetWMNormalHints(display, main_win, &xsh);
                XResizeWindow(display, main_win, xsh.width, xsh.height);
                r_rect = 2;
       		refresh();
		break;
#endif /* X_ROTATE */
       	    }
	}
#ifdef X_APM
        /* ⡼ɡŸޡ */
        draw_mode_power();
	/* refresh(); */
#endif /* X_APM */
    }
    if (key_buf_cnt) {
	//if (xfd > 0) FD_CLR(xfd, readfds);
	FD_ZERO(readfds);
        FD_SET(0, readfds);
	return 1;
    }
    if(display && xfd > 0) {
        /* Xfile descriptorɲ */
        if(nfds < xfd){
            nfds = xfd + 1;
        }

        /* read check */
        FD_SET(xfd, readfds);
    }
    FD_CLR(0, readfds);

    if (timeout < 0) {
	top = 0;
    } else {
	top = &to;
	to.tv_sec = timeout / 1000;
	to.tv_usec = (timeout % 1000) * 1000;
    }
    /* select()Ƥ */
    t1 = millitime();
    ret = select(nfds, readfds, 0, 0, top);
    t2 = millitime();
    if ((ret < 0)) {
        /* error or time out */
        return ret;
    }
    if (timeout > 0) {
	timeout -= (t2 - t1);
	if (timeout < 0) timeout = 0;
    }
    if (display && FD_ISSET(xfd, readfds)) {
	goto retry;
    }
    return ret;
}

int key_conv(KeySym key) {
    switch(key){
    case XK_Shift_L:
    case XK_Shift_R:
    case XK_Control_L:
    case XK_Control_R:
    case XK_Caps_Lock:
    case XK_Shift_Lock:
    case XK_Meta_L:
    case XK_Meta_R:
    case XK_Alt_L:
    case XK_Alt_R:
      return 1;
    case XK_Home:	put_key(MK_HOME); return 1;
    case XK_End:	put_key(MK_END); return 1;
    case XK_Up:		put_key(MK_UP); return 1;
    case XK_Down:	put_key(MK_DOWN); return 1;
    case XK_Right:	put_key(MK_RIGHT); return 1;
    case XK_Left:	put_key(MK_LEFT); return 1;
    case XK_Page_Up:	put_key(MK_PAGE_UP); return 1;
    case XK_Page_Down:	put_key(MK_PAGE_DOWN); return 1;
    case XK_Insert:	put_key(MK_INS); return 1;
    case XK_F1:		put_key(MK_F1); return 1;
    case XK_F2:		put_key(MK_F2); return 1;
    case XK_F3:		put_key(MK_F3); return 1;
    case XK_F4:		put_key(MK_F4); return 1;
    case XK_F5:		put_key(MK_F5); return 1;
    case XK_F6:		put_key(MK_F6); return 1;
    case XK_F7:		put_key(MK_F7); return 1;
    case XK_F8:		put_key(MK_F8); return 1;
    case XK_F9:		put_key(MK_F9); return 1;
    case XK_F10:	put_key(MK_F10); return 1;
    case XK_F11:	put_key(MK_F11); return 1;
    case XK_F12:	put_key(MK_F12); return 1;
    }
    return 0;
}
#endif

int load_font(char *fname, int width, int height) {
	struct romfont *f = NULL;
	unsigned char *p;
	int fd;
	char fpath[256];

	if (romfont[4].height == 0) {
		f = &romfont[4];
	} if (romfont[5].height == 0) {
		f = &romfont[5];
	} else {
		return 0;
	}

	strcpy(fpath,MGLDIR);
	strcat(fpath,"/");
	strcat(fpath,fname);
	fd = open(fpath,0);
	if (fd < 0) return 0;

	f->height = height;
	f->width = width;
	f->aoffset = 32;
	f->awidth = (width/2+7)/8;
	f->abytes = f->awidth * height;
	f->koffset = f->aoffset + f->abytes * 256;
	f->kwidth = (width+7)/8;
	f->kbytes = f->kwidth * height;
	f->fd = fd;

	p = (unsigned char *) mmap((caddr_t)0, f->koffset + f->kbytes*8064,
				   PROT_READ,
				   MAP_SHARED | MAP_FILE,
				   fd, 0);

	f->aoffset += p - font_base;
	f->koffset += p - font_base;
	return 1;
}

int write_screen_mgr(char *name, struct screen *s) {
  char buf[MAXPATHLEN];
  char magic_buf[MGR_HEADER_SIZE + 1];
  int fd;
  int size;

  /*
  if (s == NULL) s = current_screen;
   */
  push_screen(s);

#if 1
  strcpy(buf, name);
  strcat(buf, ".mgr");
#endif

  /* file open */
  fd = open(buf, O_RDWR | O_CREAT | O_TRUNC, 0644);
  if(fd == -1){
    perror("open");
    pop_screen();
    return(-1);
  }

  /* write header */
  sprintf(magic_buf, "#MGR%04d%04d%04d", 1, s->width, s->height);
  if(write(fd, magic_buf, MGR_HEADER_SIZE) != MGR_HEADER_SIZE){
    perror("write");
    close(fd);
    unlink(buf);
    pop_screen();
    return(-1);
  }

  size = s->wbytes * s->height;

  /* write data */
  if(write(fd, s->bitmap, size) != size){
    perror("write");
    close(fd);
    unlink(buf);
    pop_screen();
    return(-1);
  }

  /* file close */
  close(fd);

  pop_screen();

  return(0);
}

struct screen * read_screen_mgr(char *name) {
  char buf[MGR_HEADER_SIZE + 1];
  int fd;
  int ver, w, h;
  int size;
  struct screen *ret;

  /* file open */
  fd = open(name, O_RDONLY);
  if(fd == -1){
    perror("open");
    return(NULL);
  }

  /* read header */
  if(read(fd, buf, MGR_HEADER_SIZE) != MGR_HEADER_SIZE){
    perror("read");
    close(fd);
    return(NULL);
  }
  buf[MGR_HEADER_SIZE] = '\0';
  if(sscanf(buf, "#MGR%04d%04d%04d", &ver, &w, &h) != 3){
    close(fd);
    return(NULL);
  }

  ret = create_memscreen(w, h, NULL);
  if(!ret){
    close(fd);
    return(NULL);
  }

  size = (w + 3) / 4 * h;

  /* read data */
  if(read(fd, ret->bitmap, size) != size){
    perror("read");
    close(fd);
    free_screen(ret);
    return(NULL);
  }

  /* file close */
  close(fd);

  return(ret);
}

struct virtual_key *
create_virtual_key(x,y,xs,ys,keycode) {
	return create_virtual_key3(x,y,xs,ys,MK_V0,MK_V0,keycode);
}

struct virtual_key *
create_virtual_key3(x,y,xs,ys,down,move,up) {
	struct virtual_key *vk;

	if (vk_free) {
		vk = vk_free;
		vk_free = vk->link;
	} else {
		vk = (struct virtual_key *)malloc(sizeof(struct virtual_key));
	}
	if (!vk) return vk;
	vk->up = up;
	vk->down = down;
	vk->move = move;
	vk->x = x;
	vk->y = y;
	vk->xs = xs;
	vk->ys = ys;
	vk->link = vk->parent = vk->child = NULL;
	return vk;
}

void vk_attach(parent,vk) struct virtual_key *parent,*vk; {
	if (!parent) parent = vk_root;
	vk->parent = parent;
	vk->link = parent->child;
 	parent->child = vk;
}

void vk_detach(vk,clean_mode) struct virtual_key *vk; {
	struct virtual_key *p;
	p = vk->parent;
	if (p) {
		if (p->child ==vk) {
			p->child = vk->link;
			vk->parent = NULL;
			vk->link = NULL;
		} else {
			p = p->child;
			while (p->link) {
			   if (p->link == vk) {
				p->link = vk->link;
				vk->parent = NULL;
				vk->link = NULL;
				break;
			    }
			    p = p->link;
			}
		}
	}
	if (clean_mode) {
		while ((struct virtual_key volatile *)(vk->child)) {
			vk_detach(vk->child,clean_mode);
		}
		vk->link = vk_free;
		vk_free = vk;
	}
}

static int vk_search(p,x,y) struct virtual_key *p; {
	struct virtual_key *v;
	v = p->child;
	while (v) {
		if (vk_search(v,x - p->x,y - p->y)) return 1;
		v = v->link;
	}
	if ((p->down == MK_V0) && (p->up == MK_V0)) return 0;
	if ( (p->x <= x)&&(x <p->x+p->xs)&&(p->y <= y)&&(y < p->y+p->ys)) {
		vkp = p;
		return 1;
	}
	return 0;
}

int vk_match(p,x,y) struct virtual_key *p; {
	struct virtual_key *pp;
	struct virtual_key *v;

	pp = p->parent;
	while (pp) {
		x -= pp->x;
		y -= pp->y;
		pp = pp->parent;
	}
	v = p->child;
	while (v) {
		if (vk_search(v,x - p->x,y - p->y)) return 0;
		v = v->link;
	}
	if ( (p->x <= x)&&(x <p->x+p->xs)&&(p->y <= y)&&(y < p->y+p->ys)) {
		return 1;
	}
	return 0;
}

#ifdef X_EMU
static int vk_rev = 0;

static void vk_reverse(p,mode) struct virtual_key *p; {
	struct virtual_key *pp;
	int x,y;
	int sv;

	if (vk_rev && !mode) return;
	if (!vk_rev && mode) return;
	vk_rev = !vk_rev;

	x = p->x;
	y = p->y;
	pp = p->parent;
	while (pp) {
		x += pp->x;
		y += pp->y;
		pp = pp->parent;
	}

	push_screen(bgmode?svscreen:&physical_screen);
	sv = pen_color;
	pen_color = COLOR_REVERSE;

	fill_rect(x,y,p->xs,p->ys);

	pen_color = sv;
	pop_screen();
refresh();
}

static void vk_init() {
	vk_root = create_virtual_key(0,0,SCREEN_WIDTH,SCREEN_HEIGHT,MK_V0);
	vk_x = vk_y = 0;
}
#endif
