/*
 * MGL -- MobileGear Graphic Library -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *      
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI AND YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ifndef _MGL_H /* _MGL_H */
#define _MGL_H

#include <sys/types.h>

#if defined (__FreeBSD__)
#  include <machine/endian.h>
#elif defined (__linux__)
#  ifndef _NETINET_IN_H
#    include <asm/byteorder.h>
#  endif
#endif

struct screen {
	int type; /* 1: malloced struct screen 2: malloced bitmap */
	int wbytes;
	int width;
	int height;
	char *bitmap;
};

#define COLOR_WHITE	0
#define COLOR_LIGHTGRAY	1
#define COLOR_DARKGRAY	2
#define COLOR_BLACK	3
#define COLOR_REVERSE	4


#define BLT_TILING	0x100
#define BLT_MASKING	0x200
#define BLT_MASKCOL	0x003

#define FA_NORMAL	0
#define FA_ITALIC	1
#define FA_BOLD		2

extern int mgl_screen_width;
extern int mgl_screen_height;
#define SCREEN_WIDTH	mgl_screen_width
#define SCREEN_HEIGHT	mgl_screen_height

struct textscreen {
	struct screen *s,*save;
	int attr,x,y,xs,ys;
	int xpos,ypos,lh;
	int bgcolor;
};

#define TS_SAVE		1
#define TS_BORDER	2
#define TS_BLINE	4

/* MGL KEYCODE */
#define MK_F1		0x101
#define MK_F2		0x102
#define MK_F3		0x103
#define MK_F4		0x104
#define MK_F5		0x105
#define MK_F6		0x106
#define MK_F7		0x107
#define MK_F8		0x108
#define MK_F9		0x109
#define MK_F10		0x10a
#define MK_F11		0x10b
#define MK_F12		0x10c
#define MK_HOME		0x111
#define MK_END		0x112
#define MK_UP		0x113
#define MK_DOWN		0x114
#define MK_RIGHT	0x115
#define MK_LEFT		0x116
#define MK_PAGE_UP	0x117
#define MK_PAGE_DOWN	0x118
#define MK_INS		0x119
#define MK_DEL		0x11a

/* direction */
#define DIR_NORTH	0
#define DIR_WEST	1
#define DIR_SOUTH	2
#define DIR_EAST	3

/* MGR format */
#define MGR_HEADER_SIZE 16

/* open function */
int    open_graph(void);
void   close_graph(void);
void   set_color(int col);
void   clear_screen(void);

void   draw_pixel(int x, int y);
void   draw_pixel_wc(int x, int y, int col);
int    draw_line(int x1, int y1, int x2, int y2);
void   draw_rect(int x, int y, int xs, int ys);
void   fill_rect(int x, int y, int xs, int ys);
void   set_font(int size, int attr);
void   draw_string(int x, int y, char *str);
void   draw_font(int x, int y, int code);
void   draw_font_clipping(int x, int y, int code,
                          int x1, int y1, int x2, int y2);
void   draw_font_rotated(int x, int y, int code, int direction);
void   bitblt(struct screen *dst, int dx, int dy,
              struct screen *src, int sx, int sy, int xsize, int ysize, int op);
struct screen *create_memscreen(int xs, int ys, char *bitmap);
struct screen *create_subscreen(struct screen *org, int x, int y,
                                int xs, int ys, struct screen *b);
void   free_screen(struct screen *s);
void   push_screen(struct screen *s);
void   pop_screen(void);
int    get_pixel(int x, int y);
struct textscreen *create_textscreen(struct screen *s, int x, int y,
                                     int xs, int ys, int attr);
void   free_textscreen(struct textscreen *t);
void   ts_clear(struct textscreen *t);
void   ts_put_string(struct textscreen *t, char *str, int op);
void   ts_goto(struct textscreen *t, int x, int y);
void   ts_set_bgcolor(struct textscreen *t, int c);
void   ts_put_image(struct textscreen *t, struct screen *s,
                    int sx, int sy, int w, int h, int op);
void   write_screen_xpm(char *name, struct screen *s);
void   write_screen_native(char *name, struct screen *s);
int    get_font_width(void);
int    get_font_height(void);
int    get_key_im(int timeout);
int    get_key(int timeout);
int    key_select(int nfds, fd_set *readfds, int timeout);
int    load_font(char *fname, int width, int height);
int    write_screen_mgr(char *, struct screen *);
struct screen *read_screen_mgr(char *);
void   refresh(void);

#if 0
/* internal function */
int (*im_readp)(char *ibuf, int ibufsize);
#  ifdef IM_STATIC
extern int im_read(char *ibuf, int ibufsize);
#  endif
static void signal_exit(int n);
static void screen_switch(void);
void im_load(char *name);
static void inline _draw_pixel(int x, int y);
static void inline _draw_pixel_wc(int x, int y, int col);
int get_font_width(void);
int get_font_height(void);
void toggle_im_mode(void);
int get_im_mode(void);
void put_key(int c);
#  ifndef X_EMU
static void switch_request(int sig);
static int getch(void);
#  endif

#  ifdef X_EMU
void draw_mode_power(void);
void draw_ac_line(void);
void draw_battery(int level);
int  get_apm(int *level);
#    ifdef X_ROTATE
void north_refresh(void);
void west_refresh(void);
void south_refresh(void);
void east_refresh(void);
#    endif
static void inline rec_point(int x, int y);
int font_base_create(void);
int xset_font(char *str, int bytes, int code);
static long long millitime(void);
int key_conv(KeySym key);
#  endif
#endif

/* vertual key */
struct virtual_key {
	int up,down,move;
	int x,y,xs,ys;
	struct virtual_key *parent;
	struct virtual_key *child;
	struct virtual_key *link;
};

extern int vk_x,vk_y;

struct virtual_key *create_virtual_key();
struct virtual_key *create_virtual_key3();
void vk_attach();
void vk_detach();

/* vk Key Code 0x201 -- 0x2ff */

#define MK_V0	0x200
#define MK_V1	0x201
#define MK_V2	0x202
#define MK_V3	0x203
#define MK_V4	0x204
#define MK_V5	0x205
#define MK_V6	0x206
#define MK_V7	0x207
#define MK_V8	0x208
#define MK_V9	0x209
#define MK_V10	0x20a
#define MK_V11	0x20b
#define MK_V12	0x20c
#define MK_V13	0x20d
#define MK_V14	0x20e
#define MK_V15	0x20f

#endif  /* _MGL_H */
