/*
 * MGTERM -- Terminal Emulator for MobileGear -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include "mgl.h"

static char *keys[] = {
"","","","","",
"","","","","",
"+-","","","","",
"","","","","",
};
static int keyvals[] = {
'(','7','8','9','/',
')','4','5','6','*',
',','1','2','3','-',
'c','0','.','=','+',
};


#if 0
main() {
	open_graph();
	mcalc(-2);
	for (;;) {
		mcalc(get_key(10));
	}
}
#endif

static long long value[10];
static int dot[10];
static int op[10];
static int minus;
static int stk,ven,den,fixed;

#define XOFF	480
#define YOFF	8
#define XSIZE	160
#define YSIZE	224
#define IXOFF	(XOFF+4)
#define IYOFF	(YOFF+4)
#define IXSIZE	(XSIZE-8)
#define IYSIZE	(YSIZE-8)

void mcalc(int c);
void clear_input(void);
void fix_input(void);
void fix_dot(int d, int s);
void do_calc(void);

void mcalc(c) {
	static struct textscreen *ts;
	char buf[512];
	int x,y;
	int i,j;
	int cc = 0;
	static int ri=0,rj=0;
#if 0
	double v;
#endif

	if (c == -1) return;
	if (ts == NULL) {
		ts = create_textscreen(NULL,XOFF,YOFF,160,224,TS_BORDER|TS_BLINE);
		ts_set_bgcolor(ts,COLOR_LIGHTGRAY);
	}

	if (c == -3) {
		set_color(COLOR_BLACK);
		draw_rect(XOFF,YOFF,160-1,224-1);
		return;
	} else if (c == -4) {
		set_color(ts->bgcolor);
		draw_rect(XOFF,YOFF,160-1,224-1);
		return;
	}
	if (c == -2) {
		ts_clear(ts);
		stk = 0;
		clear_input();
	}

	if (c == -2) {
		set_font(16,FA_BOLD);
		for (i=0; i<4; i++) {
			y = YOFF+4+92+i*28;
			for (j=0; j<5; j++) {
				x = XOFF+4+8+j*28;
				set_color(COLOR_DARKGRAY);
				fill_rect(x,y,26,26);
				set_color(COLOR_LIGHTGRAY);
				draw_pixel(x,y);
				draw_pixel(x,y+25);
				draw_pixel(x+25,y);
				draw_pixel(x+25,y+25);
				set_color(COLOR_WHITE);
				ts_goto(ts,8+j*28+1+4,92+i*28+1+4);
				ts_put_string(ts,keys[i*5+j],1);
				set_color(COLOR_BLACK);
			}
		}
	} 
	i = ri;
	j = rj;
	if (c == MK_LEFT && rj>0) { cc = 1; rj--;} 
	else if (c == MK_RIGHT && rj<4) { cc = 1; rj++;} 
	else if (c == MK_UP && ri>0) { cc = 1; ri--; }
	else if (c == MK_DOWN && ri<3) { cc = 1; ri++;} 
	if (cc && c != -2) {
		x = XOFF+4+8+j*28+1;
		y = YOFF+4+92+i*28+1;
		set_color(COLOR_REVERSE);
		draw_rect(x,y,24,24);
	}
	if (cc || c == -2) {
		x = XOFF+4+8+rj*28+1;
		y = YOFF+4+92+ri*28+1;
		set_color(COLOR_REVERSE);
		draw_rect(x,y,24,24);
	}
	if (cc) {
		refresh();
		return;
	}
	if (c == ' ') c = keyvals[ri*5+rj];

	switch(c) {
	case '.':
		if (fixed) clear_input();
		den = 1;
		ven = 1;
		break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		ven = 1;
		if (fixed) clear_input();
		if (den) dot[stk]++;
		value[stk] *= 10;
		value[stk] += c - '0';
		break;
	case '=':
		if (stk != 0 && op[stk-1]) do_calc();
		break;
	case '-':
		if (stk != 0 && op[stk-1]) do_calc();
		if (!ven) {minus = !minus; break;}
		fix_input();
		op[stk] = c;
		stk++;
		clear_input();
		break;
	case ',':
		minus = !minus;
		break;
	case '(':
	case ')':
		break;
	case '+':
	case '*':
	case '/':
		if (stk != 0 && op[stk-1]) do_calc();
		fix_input();
		op[stk] = c;
		stk++;
		clear_input();
	}
	set_font(24,0);
	set_color(COLOR_BLACK);
	set_color(COLOR_WHITE);
	fill_rect(XOFF+4+8,YOFF+4+4+4+4,12*10+12,24+16);
	set_color(COLOR_BLACK);
	draw_rect(XOFF+4+8,YOFF+4+4+4+4,12*10+12,24+16);
	set_font(12,0);
	ts_goto(ts,8+6,4);
	ts_set_bgcolor(ts,COLOR_WHITE);
	sprintf(buf,"%c %d %d",minus?'-':'_',stk,dot[stk]);
#if 0
	ts_put_string(ts,buf,0);
#endif
	set_font(24,FA_BOLD);
	ts_goto(ts,8+6,4+4+12);
	sprintf(buf,"%10d",(int)(value[stk]*(minus?-1:1)));
	ts_put_string(ts,buf,0);
	ts_set_bgcolor(ts,COLOR_LIGHTGRAY);
	draw_pixel(XOFF+4+8+6+(10-dot[stk])*12-1,YOFF+4+4+4+12+23);
	draw_pixel(XOFF+4+8+6+(10-dot[stk])*12,YOFF+4+4+4+12+23);
	draw_pixel(XOFF+4+8+6+(10-dot[stk])*12-1,YOFF+4+4+4+12+24);
	draw_pixel(XOFF+4+8+6+(10-dot[stk])*12,YOFF+4+4+4+12+24);
refresh();
}

void clear_input() {
	value[stk] = 0;
	op[stk] = 0;
	dot[stk] = 0;
	den = 0;
	ven = 0;
	minus = 0;
	fixed = 0;
}

void fix_input() {
	if (minus) value[stk] *= -1;
	minus = 0;
	fixed = 1;
}

void fix_dot(d,s) {
	while (d > dot[s]) {
		value[s] *= 10;
		dot[s] ++;
	}
	while (d < dot[s]) {
		value[s] /= 10;
		dot[s] --;
	}
}

void do_calc() {
	int maxdot;

	if (stk == 0) return;
	fix_input();
	maxdot = dot[stk];
	if (maxdot > 9) maxdot = 9;

#if 0
	if (dot[stk-1] > maxdot) dot[stk-1];
#endif
	if (op[stk-1] == '*') {
		value[stk-1] = value[stk-1] * value[stk];
		dot[stk-1] = dot[stk] + dot[stk-1];
		op[stk] = 0;
		stk--;
		fix_dot(maxdot,stk);
	} else if (op[stk-1] == '/') {
		while (dot[stk-1] < 9 && value[stk-1] < 10000000000LL) {
			dot[stk-1]++;
			value[stk-1] *= 10;
		}
		value[stk-1] = value[stk-1] / value[stk];
		dot[stk-1] = dot[stk-1] - dot[stk];
		while (dot[stk-1] != 0 && (dot[stk-1] % 10 == 0) ) {
			dot[stk-1]--;
			value[stk-1] /= 10;
		}
		op[stk-1] = 0;
		stk--;
		fix_dot(maxdot,stk);
	} else if (op[stk-1] == '+') {
		fix_dot(maxdot,stk);
		fix_dot(maxdot,stk-1);
printf("%d + %d",(int)(value[stk-1]),(int)(value[stk]));
		value[stk-1] = value[stk-1] + value[stk];
printf("= %d\n",(int)(value[stk-1]));
		op[stk-1] = 0;
		stk--;
	} else if (op[stk-1] == '-') {
		fix_dot(maxdot,stk);
		fix_dot(maxdot,stk-1);
		value[stk-1] = value[stk-1] - value[stk];
		op[stk-1] = 0;
		stk--;
	}
}
