/*
 * MGTERM -- Terminal Emulator for MobileGear -
 * Copyright (C) 1998, 1999
 *      Koji Suzuki (suz@at.sakura.ne.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY KOJI SUZUKI ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "mgl.h"

#define MAXHEIGHT	12
#define WIDTH	10
#define HEIGHT	10

static int rest;
static char map[WIDTH][HEIGHT];
static char mark[WIDTH][HEIGHT];
static int markcnt=0;
static int curx,cury;
static int point=0;
static int auto_mode=0;

#define XOFF	480
#define YOFF	8
#define XSIZE	160
#define YSIZE	224
#define IXOFF	(XOFF+4)
#define IYOFF	(YOFF+4)
#define IXSIZE	(XSIZE-8)
#define IYSIZE	(YSIZE-8)

static void help(void);
static void auto_play(void);
static void flush_undo(int e);

static void help() {
	static char *help_msg = "\
᤬ɤǤ\n\
\n\
륭\n\
     \n\
      h j k l\n\
\n\
u     ɥ\n\
SPACE ä\n\
R     ȥ饤\n\
N     ƥ\n\
a     auto ⡼\n";

	struct textscreen *ts;
	ts = create_textscreen(0,XOFF,YOFF,160,224,TS_BORDER|TS_BLINE|TS_SAVE);
	ts_clear(ts);
	set_color(COLOR_BLACK);
	set_font(12,0);
	ts_put_string(ts,help_msg,0);
	refresh();
	get_key(50);
	free_textscreen(ts);
	refresh();
}

static void show_point() {
	char buf[100];
	int xpos,ypos;

	xpos = XOFF + 40;
	ypos = YOFF + 12;
	set_color(COLOR_LIGHTGRAY);
	fill_rect(xpos-2,ypos-2,100,20);
	set_color(COLOR_BLACK);
	set_font(16,FA_BOLD);

	sprintf(buf,"%12d",point);
	draw_string(xpos,ypos,buf);
}

static void show_cursor() {
	int xpos,ypos;

	if (mark[curx][cury]) return;
	xpos = XOFF + 10+ curx*14;
	ypos = YOFF + 200 - (cury*14);
	set_color(COLOR_REVERSE);
	fill_rect(xpos,ypos,14-1,14-1);
}

static void hide_cursor() {
	int xpos,ypos;

	if (mark[curx][cury]) return;
	xpos = XOFF + 10+ curx*14;
	ypos = YOFF + 200 - (cury*14);
	set_color(COLOR_REVERSE);
	fill_rect(xpos,ypos,14-1,14-1);
}

static void draw_icon(x,y,sel) {

	static char *pat = "";
	int xpos,ypos;
	int c,code;
	xpos = XOFF + 10+ x*14;
	ypos = YOFF + 200 - (y*14);
	set_color(sel?COLOR_WHITE:COLOR_LIGHTGRAY);
	fill_rect(xpos,ypos,14,14);
	c = map[x][y];
	set_color(COLOR_BLACK);
	set_font(12,0);
	if (c > 0) {
		code = ((pat[(c-1)*2] << 8) & 0xff00) | (pat[(c-1)*2+1] & 0xff);
		draw_font(xpos+1,ypos+1,code);
	}
#if 0
	if (sel) {
		set_color(COLOR_REVERSE);
		fill_rect(xpos,ypos,14,14);
	}
#endif
}

static void mark_same(x,y,lvl,dir) {
	int i,j;
	char c = map[x][y];
	if (lvl == 0) {
		for (i=0; i< WIDTH; i++) for (j=0; j<HEIGHT; j++) {
			mark[i][j] = 0;
		}
		markcnt = 0;
		if (c == 0) return;
		dir = 0;
	}
	if (x > 0 && !mark[x-1][y] && map[x-1][y] == c) {
		mark[x-1][y] = 1;
		markcnt++;
		if (!(dir & 1))
		mark_same(x-1,y,lvl+1,dir | 4);
	}
	if (y > 0 && !mark[x][y-1] && map[x][y-1] == c) {
		mark[x][y-1] = 1;
		markcnt++;
		if (!(dir & 2))
		mark_same(x,y-1,lvl+1,dir | 8);
	}
	if (x < WIDTH-1 && !mark[x+1][y] && map[x+1][y] == c) {
		mark[x+1][y] = 1;
		markcnt++;
		if (!(dir & 4))
		mark_same(x+1,y,lvl+1,dir | 1);
	}
	if (y < HEIGHT-1 && !mark[x][y+1] && map[x][y+1] == c) {
		mark[x][y+1] = 1;
		markcnt++;
		if (!(dir & 8))
		mark_same(x,y+1,lvl+1,dir | 2);
	}
	if (lvl == 0) {
		if (markcnt > 0) {
		}
	}
}

static void show_all(force) {
	int i,j;
	if (!force && rest == 0) {
		long t;
		time(&t);
		srandom(t);
		for (i=0; i<WIDTH; i++) {
			for (j=0; j<HEIGHT; j++) {
				map[i][j] = (random()&0x7fffffff)%5 + 1;
			}
		}
		rest = WIDTH*HEIGHT;
		point = 0;
	}
	set_color(COLOR_LIGHTGRAY);
	fill_rect(XOFF+1,YOFF+1,160-2,224-2);
	set_color(COLOR_DARKGRAY);
	draw_rect(XOFF+4,YOFF+4,160-8,224-8);
	mark_same(curx,cury,0,0);
	for (i=0; i<WIDTH; i++) {
		for (j=0; j<HEIGHT; j++) {
			draw_icon(i,j,0);
		}
	}
	show_point();
}

static void show_mark() {
	int i,j;

	mark_same(curx,cury,0,0);
	for (i=0; i<WIDTH; i++) {
		for (j=0; j<HEIGHT; j++) {
			if (mark[i][j]) {
				draw_icon(i,j,mark[i][j]);
			}
		}
	}
}

static void hide_mark() {
	int i,j;

	for (i=0; i<WIDTH; i++) {
		for (j=0; j<HEIGHT; j++) {
			if (mark[i][j]) {
				draw_icon(i,j,!mark[i][j]);
			}
		}
	}
}

struct undo_buf {
	struct undo_buf *next;
	int rest;
	int point;
	int curx,cury;
	unsigned int map[WIDTH];
} *undo_top;

static void push_undo() {
	extern char *malloc();
	struct undo_buf *buf;
	int i,j;
	unsigned int x;

	buf = (struct undo_buf *)malloc(sizeof(*buf));
	for (i=0; i<WIDTH; i++) {
		x = 0;
		for (j=0; j<HEIGHT;j++) {
			x *= 6;
			x += map[i][j];
		}
		buf->map[i] = x;
	}
	buf->rest = rest;
	buf->point = point;
	buf->curx = curx;
	buf->cury = cury;
	buf->next = undo_top;
	undo_top = buf;
}

static int pop_undo(e) {
	struct undo_buf *buf;
	int i,j;
	unsigned int x;

	buf = undo_top;
	if (!buf) return -1;
	undo_top = buf->next;

	if (e) {
		for (i=0; i<WIDTH; i++) {
			x = buf->map[i];
			for (j=HEIGHT-1; j>=0 ;j--) {
				map[i][j] = x % 6;
				x /= 6;
			}
		}
		rest = buf->rest;
		point = buf->point;
		curx = buf->curx;
		cury = buf->cury;
	}
	free(buf);
	return 0;
}

static void flush_undo(e) {
	while (pop_undo(e) == 0)
		;
}


static void delete_mark() {
	int first,last;
	int i,j,k,c,m;
	int ii;

	push_undo();
	for (i=0; i<WIDTH; i++) {
		first = -1;
		last = -1;
		for (j=0; j<HEIGHT; j++) {
			m = mark[i][j];
			if (m) {
				last = j;
				if (first == -1) first = j;
			}
		}
		if (first != -1) {
			for (j=first; j<HEIGHT; j++) {
				k = last + 1 + j - first;
				if (k >= HEIGHT) c = 0;
				else c = map[i][k];

				map[i][j] = c;
				draw_icon(i,j,0);
			}
		}
		c = 0;
		for (j=0; j<HEIGHT; j++) {
			if (map[i][j] != 0) c++;
		}
		if (first != -1 && c == 0) {
			for (ii=i; ii<WIDTH-1; ii++) {
				for (j=0; j<HEIGHT; j++) {
					map[ii][j] = map[ii+1][j];
					mark[ii][j] = mark[ii+1][j];
					draw_icon(ii,j,mark[ii][j]);
				}
			}
			for (j=0; j<HEIGHT; j++) {
				map[WIDTH-1][j] = 0;
				mark[WIDTH-1][j] = 0;
				draw_icon(WIDTH-1,j,0);
			}
			if (i != HEIGHT-1) i--;
		}
	}
	rest -= markcnt;
	point += (markcnt -1)*(markcnt -1);
	markcnt = 0;
}


static void auto_play() {
	int x,y;
	int retry_cnt = 0;
	x = y = -1;

	for (;;) {
retry:
		retry_cnt++;
		if (retry_cnt > 100) break;
		if (markcnt) hide_mark();
		cury++;
		if (cury > HEIGHT) {
			cury=0;
			curx++;
			if (curx > WIDTH) {
				pop_undo(1);
				show_all(1);
				x = curx;
				y = cury;
				goto retry;
			}
		}
		show_mark();
refresh();
		if (markcnt) {
			if (x >= 0 && y >= 0 && mark[x][y]) goto retry;
			delete_mark();
			curx = cury = 0;
			show_mark();
refresh();
		} else {
			goto retry;
		}
#if 0
			rest = 0;
			point = 0;
			flush_undo(0);
			show_all(0);
			show_mark();
refresh();
#endif
		break;
	}
}

void msame(c) int c; {
	switch (c) {
	case MK_RIGHT:
	case 'l':
		if (curx < WIDTH-1) {
			hide_cursor();
			if (markcnt) hide_mark();
			curx++;
			show_mark();
			show_cursor();
		}
		break;
	case MK_LEFT:
	case 'h':
		if (curx > 0) {
			hide_cursor();
			if (markcnt) hide_mark();
			curx--;
			show_mark();
			show_cursor();
		}
		break;
	case MK_UP:
	case 'k':
		if (cury < HEIGHT-1) {
			hide_cursor();
			if (markcnt) hide_mark();
			cury++;
			show_mark();
			show_cursor();
		}
		break;
	case MK_DOWN:
	case 'j':
		if (cury >0) {
			hide_cursor();
			if (markcnt) hide_mark();
			cury--;
			show_mark();
			show_cursor();
		}
		break;
		
	case ' ':
		if (markcnt) {
			hide_cursor();
			delete_mark();
			show_mark();
			show_cursor();
			show_point();
		}
		break;
	case 'u':
		pop_undo(1);
		show_all(1);
		show_mark();
		show_cursor();
		break;
	case 'a':
		auto_mode = !auto_mode;
		break;
	case '?':
		help();
		break;
	case 'R':
		rest = 0;
		point = 0;
		flush_undo(1);
		show_all(1);
		show_mark();
		show_cursor();
		break;
	case 'N':
		rest = 0;
		point = 0;
		flush_undo(0);
		show_all(0);
		show_mark();
		show_cursor();
		break;
	case (-2):
		show_all(0);
		show_mark();
		show_cursor();
		break;
	case (-3):
		set_color(COLOR_BLACK);
		draw_rect(XOFF,YOFF,160-1,224-1);
		break;
	case (-4):
		set_color(COLOR_LIGHTGRAY);
		draw_rect(XOFF,YOFF,160-1,224-1);
		break;
	case (-1):
		if (auto_mode) {
			hide_cursor();
			auto_play();
			show_cursor();
		}
		break;
	}
}
