/*
 * MIL -- MobileGear Image Loader -
 * Copyright (C) 1998, 1999
 *      Yukihiko Sano (yukihiko@yk.rim.or.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY YUKIHIKO SANO ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#ifdef USE_JPEG
#  include "jpeglib.h"
#endif /* USE_JPEG */
#ifdef USE_GIF
#  include "gif_lib.h"
#endif /* USE_GIF */
#include "mgl.h"

#define  MIL_VER     "0.3"
#define  DEBUG       0

#define  CTRL(c)     ((c) & 037)

#define  G_FWD       CTRL('n')
#define  G_FWD2      ' '
#define  G_BWD       CTRL('p')
#define  G_BWD2      0x7f

#define  G_UP        'k'
#define  G_DOWN      'j'
#define  G_RIGHT     'l'
#define  G_LEFT      'h'

#define  SCREEN_UP     CTRL('k')
#define  SCREEN_DOWN   CTRL('j')
#define  SCREEN_RIGHT  CTRL('l')
#define  SCREEN_LEFT   CTRL('h')

#define  HALF_UP       'K'
#define  HALF_DOWN     'J'
#define  HALF_RIGHT    'L'
#define  HALF_LEFT     'H'

#define  G_HOME      MK_HOME
#define  G_END       'G'
#define  G_END2      MK_END

int  check_mgr(char *f);
int  read_file(char *f, int *w, int *h, char **image);
char *cont_jpeg(FILE *fd, int *w, int *h);
void data_cnv(char *dst, char *src, int w, int h);
void image_copy(struct screen *st, int w, int h,
                int *sx, int *sy, int *dw, int *dh);

#ifdef MINI
#define  MINI_SX     (480 + 3)
#define  MINI_SY     (8 + 2)
#define  MINI_WD     (160 - 6)
#define  MINI_HI     (224 - 5)
#define  MINI_KEY_SX 480
#define  MINI_KEY_SY 8
#define  MINI_KEY_WD (160 - 1)
#define  MINI_KEY_HI (224 - 1)

#define  MIL_FIFO    "MIL_PIPE"
#define  MIL_FILE    "/tmp/.mil_fifo"

#define  MIL_MSG1    "MobileGear Image Loader"
#define  MIL_MSG2    "(version %s)"
#define  MIL_MSG3    "%s support"

#define  SUP_MGR     "MGR"
#define  SUP_JPEG    "JPEG"
#define  SUP_GIF     "GIF"

#define  MAX_SUP     (3 + 1)

void mini_mil(int c);
void mil_message(void);

static init = 1;
static i_exist = 0;
static key_in = 0;
static int fd = -1;
static int sx, sy;
static int dx, dy;
static int dw, dh;
static int w, h;
static struct screen *st;
static char *image = NULL;
static char fifo[MAXPATHLEN];
static char img_file[MAXPATHLEN];
static char *sup_list[MAX_SUP + 1];

void mil_message(void)
{
    int i;
    int fnt = 12;
    int mx, my;
    int rx, ry, rw, rh;
    char buff[32];

    /* Title */
    set_color(COLOR_BLACK);
    set_font(fnt, FA_ITALIC | FA_BOLD);
    mx = MINI_SX + (MINI_WD - strlen(MIL_MSG1) * fnt / 2) / 2;
    my = MINI_SY + 5;
    draw_string(mx, my, MIL_MSG1);

    rx = MINI_SX + (mx - MINI_SX) / 2;
    ry = MINI_SY + 2;
    rw = MINI_WD - (mx - MINI_SX);

    sprintf(buff, MIL_MSG2, MIL_VER);
    mx = MINI_SX + (MINI_WD - strlen(buff) * fnt / 2) / 2;
    my = MINI_SY + 5 + fnt + 5;
    draw_string(mx, my, buff);

    rh = my + fnt + 2 - ry;

    set_color(COLOR_DARKGRAY);
    draw_rect(rx, ry, rw, rh);

    /* support format list */
    set_color(COLOR_BLACK);
    set_font(fnt, FA_NORMAL);
    strcpy(buff, "supported format");
    mx = MINI_SX + (MINI_WD - strlen(buff) * fnt / 2) / 2;
    my = MINI_SY + 5 + (fnt + 5) * 2 + 5;
    draw_string(mx, my, buff);

    my = MINI_SY + 5 + (fnt + 5) * 3 + 5;
    for(i = 0 ; sup_list[i] ; i++){
#if 0
        sprintf(buff, MIL_MSG3, sup_list[i]);
        mx = MINI_SX + (MINI_WD - strlen(buff) * fnt / 2) / 2;
        mx = MINI_SX + fnt / 2 * 2;
#else
        sprintf(buff, "%s", sup_list[i]);
        mx = MINI_SX + (MINI_WD - strlen(buff) * fnt / 2) / 2;
#endif
        draw_string(mx, my, buff);
        my = MINI_SY + 5 + (fnt + 5) * 3 + 5 + (fnt + 3) * (i + 1);
    }

    return;
}

void mini_mil(int c)
{
    fd_set fds;
    int i;
    int ret;
    char buff[MAXPATHLEN];

    if(init){
        if(getenv(MIL_FIFO)){
            strcpy(fifo, (const char *)getenv(MIL_FIFO));
        }else{
            strcpy(fifo, MIL_FILE);
        }
#if DEBUG
        printf("fifo [%s]\n", fifo);
#endif

#ifdef __FreeBSD__
        if(mkfifo(fifo, 0666) < 0){
            if(errno != EEXIST){
                perror("mknod");
                return;
            }
        }
#else
        if(mknod(fifo, S_IFIFO | 0666, 0) < 0){
            if(errno != EEXIST){
                perror("mknod");
                return;
            }
        }
#endif
        fd = open(fifo, O_RDONLY | O_NDELAY);
        if(fd == -1){
            perror("open");
            return;
        }

        for(i = 0 ; i < MAX_SUP ; i++){
            sup_list[i] = NULL;
        }

        i = 0;
        /* Check support format */
        sup_list[i] = SUP_MGR;
        i++;

#ifdef USE_JPEG
        sup_list[i] = SUP_JPEG;
        i++;
#endif /* USE_JPEG */

#ifdef USE_GIF
        sup_list[i] = SUP_GIF;
        i++;
#endif /* USE_GIF */

        init = 0;
    }

    FD_ZERO(&fds);
    FD_SET(fd,&fds);
#if 0
    ret = key_select(fd + 1, &fds, 500);
#else
{
    struct timeval tv;

    tv.tv_sec  = 0;
    tv.tv_usec = 500 * 1000;

    ret = select(fd + 1, &fds, NULL, NULL, &tv);
}
#endif
    if(ret < 0){
        return;
    }
    if(FD_ISSET(fd, &fds)){
        /* File Name Read */
        memset(buff, 0x00, MAXPATHLEN);
        if(read(fd, buff, MAXPATHLEN - 1) > 0){
            if(buff[strlen(buff) - 1] == '\n'){
                buff[strlen(buff) - 1] = 0x00;
            }
            strcpy(img_file, buff);
#if DEBUG
            printf("img_file [%s]\n", img_file);
#endif
            if(i_exist){
                /* Screen Clear */
                set_color(COLOR_WHITE);
                fill_rect(MINI_SX, MINI_SY, MINI_WD, MINI_HI);

                free_screen(st);
                if(image) free(image);
                image = NULL;
                i_exist = 0;
            }

            st = read_screen_mgr(img_file);
            if(st){
                w = st->width;
                h = st->height;
            }else{
#if defined (USE_JPEG) || defined (USE_GIF)
                ret = read_file(img_file, &w, &h, &image);
                if(ret != 0){
                    mil_message();
                    return;
                }

                st = create_memscreen(w, h, NULL);
                if(!st){
                    mil_message();
                    free(image);
                    image = NULL;
                    return;
                }
                data_cnv(st->bitmap, image, w, h);
#else
                mil_message();
                if(image) free(image);
                image = NULL;
                return;
#endif
            }

            dx = 0;
            dy = 0;

            image_copy(st, w, h, &sx, &sy, &dw, &dh);
            refresh();
            i_exist = 1;
        }else{
        }
    }else{
    }

    switch(c){
    case MK_UP:
    case G_UP:
        if(i_exist){
            if(dy > 0){
                dy--;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case MK_DOWN:
    case G_DOWN:
        if(i_exist){
            if(h - dy - MINI_HI > 0){
                dy++;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case MK_RIGHT:
    case G_RIGHT:
        if(i_exist){
            if(w - dx - MINI_WD > 0){
                dx++;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case MK_LEFT:
    case G_LEFT:
        if(i_exist){
            if(dx > 0){
                dx--;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case SCREEN_UP: /* Ctrl-k */
        if(i_exist){
            if(dy > 0){
                if(dy - MINI_HI < 0){
                    dy = 0;
                }else{
                    dy-=MINI_HI;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case SCREEN_DOWN: /* Ctrl-j */
        if(i_exist){
            if(h - dy - MINI_HI > 0){
                if(h - (dy + MINI_HI) > dh){
                    dy+=MINI_HI;
                }else{
                    dy = h - MINI_HI;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case SCREEN_RIGHT: /* Ctrl-l */
        if(i_exist){
            if(w - dx - MINI_WD > 0){
                if(w - (dx + MINI_WD) > dw){
                    dx+=MINI_WD;
                }else{
                    dx = w - MINI_WD;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case SCREEN_LEFT: /* Ctrl-h */
        if(i_exist){
            if(dx > 0){
                if(dx - MINI_WD < 0){
                    dx = 0;
                }else{
                    dx-=MINI_WD;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case HALF_UP: /* Shift-k */
        if(i_exist){
            if(dy > 0){
                if(dy - (MINI_HI / 2) < 0){
                    dy = 0;
                }else{
                    dy-=(MINI_HI / 2);
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case HALF_DOWN: /* Shift-j */
        if(i_exist){
            if(h - dy - MINI_HI > 0){
                if(h - (dy + (MINI_HI / 2)) > dh){
                    dy+=(MINI_HI / 2);
                }else{
                    dy = h - MINI_HI;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case HALF_RIGHT: /* Shift-l */
        if(i_exist){
            if(w - dx - MINI_WD > 0){
                if(w - (dx + (MINI_WD / 2)) > dw){
                    dx+=(MINI_WD / 2);
                }else{
                    dx = w - MINI_WD;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case HALF_LEFT: /* Shift-h */
        if(i_exist){
            if(dx > 0){
                if(dx - (MINI_WD / 2) < 0){
                    dx = 0;
                }else{
                    dx-=(MINI_WD / 2);
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case G_HOME:
        if(i_exist){
            if(dy != 0){
                dy = 0;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case G_END:
    case G_END2:
        if(i_exist){
            if(h > MINI_HI){
                dy = h - MINI_HI;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            refresh();
        }
        break;
    case (-1):
        break;
    case (-2):
        if(i_exist){
            set_color(COLOR_WHITE);
            fill_rect(MINI_SX, MINI_SY, MINI_WD, MINI_HI);
            bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
        }else{
            set_color(COLOR_WHITE);
            fill_rect(MINI_SX, MINI_SY, MINI_WD, MINI_HI);

            /* No Image */
            mil_message();
        }
        refresh();
        return;
    case (-3):
        key_in = 1;
        set_color(COLOR_BLACK);
        draw_rect(MINI_KEY_SX, MINI_KEY_SY, MINI_KEY_WD, MINI_KEY_HI);
        refresh();
        return;
    case (-4):
        key_in = 0;
        set_color(COLOR_LIGHTGRAY);
        draw_rect(MINI_KEY_SX, MINI_KEY_SY, MINI_KEY_WD, MINI_KEY_HI);
        refresh();
        return;
    }

    return;
}
#else  /* non MINI */

int main(int argc, char **argv)
{
    int ret = 0;
    int sx, sy;
    int dx, dy;
    int dw, dh;
    int w = 0, h = 0;
    int i;
    int loop = 1;
    int file_num = 1;
    struct screen *st = NULL;
    struct screen *new_st;
    char *image = NULL;
#if defined (USE_JPEG) || defined (USE_GIF)
    int new_w, new_h;
    char *new_image = NULL;
#endif

    if(argc == 1){
        fprintf(stderr, "Usage: %s file...\n", argv[0]);
        exit(1);
    }else{
        for(i = 1 ; i < argc ; i++){
            ret = check_mgr(argv[i]);
            if(!ret){
                file_num = i;
                break;
            }else{
#if defined (USE_JPEG) || defined (USE_GIF)
                ret = read_file(argv[i], &w, &h, &image);
                if(ret){
                    continue;
                }
                file_num = i;
                break;
#else
            fprintf(stderr, "%s: unsupported format\n", argv[i]);
            continue;
#endif
            }
        }
    }
    if(ret){
        exit(1);
    }

    /* Init mgl */
    if(open_graph() != 1){
        fprintf(stderr, "Cannot initialize mgl\n");
        exit(1);
    }
    set_color(COLOR_WHITE);
    clear_screen();

    st = read_screen_mgr(argv[file_num]);
    if(st){
        w = st->width;
        h = st->height;
    }else{
#if defined (USE_JPEG) || defined (USE_GIF)
        st = create_memscreen(w, h, NULL);
        if(!st){
            free(image);
            image = NULL;
            exit(1);
        }

        data_cnv(st->bitmap, image, w, h);
#else
        exit(1);
#endif
    }

    dx = 0;
    dy = 0;
    image_copy(st, w, h, &sx, &sy, &dw, &dh);

    refresh();

    while (loop) {
        ret = get_key(10);
        if(ret == -1){
            continue;
        }
        switch(ret){
        case MK_UP:
        case G_UP:
            if(dy > 0){
                dy--;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case MK_DOWN:
        case G_DOWN:
            if(h - dy - SCREEN_HEIGHT > 0){
                dy++;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case MK_RIGHT:
        case G_RIGHT:
            if(w - dx - SCREEN_WIDTH > 0){
                dx++;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case MK_LEFT:
        case G_LEFT:
            if(dx > 0){
                dx--;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case SCREEN_UP: /* Ctrl-k */
            if(dy > 0){
                if(dy - SCREEN_HEIGHT < 0){
                    dy = 0;
                }else{
                    dy-=SCREEN_HEIGHT;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case SCREEN_DOWN: /* Ctrl-j */
            if(h - dy - SCREEN_HEIGHT > 0){
                if(h - (dy + SCREEN_HEIGHT) > dh){
                    dy+=SCREEN_HEIGHT;
                }else{
                    dy = h - SCREEN_HEIGHT;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case SCREEN_RIGHT: /* Ctrl-l */
            if(w - dx - SCREEN_WIDTH > 0){
                if(w - (dx + SCREEN_WIDTH) > dw){
                    dx+=SCREEN_WIDTH;
                }else{
                    dx = w - SCREEN_WIDTH;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case SCREEN_LEFT: /* Ctrl-h */
            if(dx > 0){
                if(dx - SCREEN_WIDTH < 0){
                    dx = 0;
                }else{
                    dx-=SCREEN_WIDTH;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case HALF_UP: /* Shift-k */
            if(dy > 0){
                if(dy - (SCREEN_HEIGHT / 2) < 0){
                    dy = 0;
                }else{
                    dy-=(SCREEN_HEIGHT / 2);
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case HALF_DOWN: /* Shift-j */
            if(h - dy - SCREEN_HEIGHT > 0){
                if(h - (dy + (SCREEN_HEIGHT / 2)) > dh){
                    dy+=(SCREEN_HEIGHT / 2);
                }else{
                    dy = h - SCREEN_HEIGHT;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case HALF_RIGHT: /* Shift-l */
            if(w - dx - SCREEN_WIDTH > 0){
                if(w - (dx + (SCREEN_WIDTH / 2)) > dw){
                    dx+=(SCREEN_WIDTH / 2);
                }else{
                    dx = w - SCREEN_WIDTH;
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case HALF_LEFT: /* Shift-h */
            if(dx > 0){
                if(dx - (SCREEN_WIDTH / 2) < 0){
                    dx = 0;
                }else{
                    dx-=(SCREEN_WIDTH / 2);
                }
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case G_HOME:
            if(dy != 0){
                dy = 0;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case G_END:
        case G_END2:
            if(h > SCREEN_HEIGHT){
                dy = h - SCREEN_HEIGHT;
                bitblt(NULL, sx, sy, st, dx, dy, dw, dh, 0);
            }
            break;
        case G_FWD:  /* Next */
        case G_FWD2: /* Next */
            if(file_num + 1 < argc){
                file_num++;
                new_st = read_screen_mgr(argv[file_num]);
                if(new_st){
                    free_screen(st);
                    st = new_st;
                    w = st->width;
                    h = st->height;
                }else{
#if defined (USE_JPEG) || defined (USE_GIF)
                    ret = read_file(argv[file_num], &new_w, &new_h, &new_image);
                    if(ret != 0){
                        continue;
                    }
#if 0
fprintf(stderr, "FWD [%d]\n", file_num);
#endif
                    new_st = create_memscreen(new_w, new_h, NULL);
                    if(!new_st){
                        free(new_image);
                    }

                    free_screen(st);
                    if(image) free(image);
                    image = new_image;
                    st = new_st;
                    w = new_w;
                    h = new_h;

                    data_cnv(st->bitmap, image, w, h);
#else
                    fprintf(stderr, "%s: unsupported format\n", argv[file_num]);
                    continue;
#endif
                }

                set_color(COLOR_WHITE);
                clear_screen();

                dx = 0;
                dy = 0;
                image_copy(st, w, h, &sx, &sy, &dw, &dh);
            }
            break;
        case G_BWD:  /* Pre */
        case G_BWD2: /* Pre */
            if(file_num - 1 >= 1){
                file_num--;
                new_st = read_screen_mgr(argv[file_num]);
                if(new_st){
                    free_screen(st);
                    st = new_st;
                    w = st->width;
                    h = st->height;
                }else{
#if defined (USE_JPEG) || defined (USE_GIF)
                    ret = read_file(argv[file_num], &new_w, &new_h, &new_image);
                    if(ret != 0){
                        continue;
                    }
#if 0
fprintf(stderr, "BWD [%d]\n", file_num);
#endif
                    new_st = create_memscreen(new_w, new_h, NULL);
                    if(!new_st){
                        free(new_image);
                    }

                    free_screen(st);
                    if(image) free(image);
                    image = new_image;
                    st = new_st;
                    w = new_w;
                    h = new_h;

                    data_cnv(st->bitmap, image, w, h);
#else
                    fprintf(stderr, "%s: unsupported format\n", argv[file_num]);
                    continue;
#endif
                }

                set_color(COLOR_WHITE);
                clear_screen();

                dx = 0;
                dy = 0;
                image_copy(st, w, h, &sx, &sy, &dw, &dh);
            }
            break;
        case 'q':
            loop = 0;
            break;
        }
#if DEBUG
    printf("[0%o][0x%x]\n", ret, ret);
    printf("sx [%d]\n",sx);
    printf("sy [%d]\n",sy);
    printf("dx [%d]\n",dx);
    printf("dy [%d]\n",dy);
    printf("dw [%d]\n",dw);
    printf("dh [%d]\n",dh);
#endif
        refresh();
    }
    free_screen(st);
    if(image) free(image);

    exit(0);
}
#endif /* MINI */

int check_mgr(char *f)
{
    int fd;
    char buf[MGR_HEADER_SIZE + 1];

    /* file open */
    fd = open(f, O_RDONLY);
    if(fd == -1){
        return(-1);
    }

    /* read header */
    if(read(fd, buf, MGR_HEADER_SIZE) != MGR_HEADER_SIZE){
        close(fd);
        return(1);
    }
    close(fd);

    buf[MGR_HEADER_SIZE] = '\0';

    if(strncmp(buf, "#MGR", 4)){
        return(1);
    }

    return(0);
}

void data_cnv(char *dst, char *src, int w, int h)
{
    int i, j, k;
    unsigned char c;

    for(i = 0 ; i < h ; i++){
        for(j = 0 ; j < w / 4 ; j++){
            for(k = 0 ; k < 4 ; k++){
                c = (unsigned char)(*src);
                switch(c){
                case 0:
                    c = 3;
                    break;
                case 1:
                    c = 2;
                    break;
                case 2:
                    c = 1;
                    break;
                case 3:
                    c = 0;
                    break;
                default:
                    c = 0;
                    break;
                }
                *dst = *dst | c<<(k * 2);
                src++;
            }
            dst++;
        }
        for(k = 0 ; k < w % 4 ; k++){
            c = (unsigned char)(*src);
            switch(c){
            case 0:
                c = 3;
                break;
            case 1:
                c = 2;
                break;
            case 2:
                c = 1;
                break;
            case 3:
                c = 0;
                break;
            default:
                c = 0;
                break;
            }
            *dst = *dst | c<<(k * 2);
            src++;
        }
        if(w % 4){
            dst++;
        }
    }
}

#ifdef USE_GIF
static int intcompare(i, j)
int *i;
int *j;
{
    if (*i > *j)
        return (1);
    if (*i < *j)
        return (-1);
    return (0);
}

extern int _GifError;

int read_gif_file(char *f, int *w, int *h, char **image)
{
    /* The way Interlaced image should. */
    static int InterlacedOffset[] = { 0, 4, 2, 1 };
    /* be read - offsets and jumps... */
    static int InterlacedJumps[] = { 8, 8, 4, 2 };
    static int ColorMapSize = 0;
    static int BackGround = 0;
    static ColorMapObject *ColorMap;

    GifFileType *GifFile;
    GifRowType  *ScreenBuffer;
    GifRowType  GifRow;
    GifRecordType RecordType;
    GifByteType *Extension;
    static GifColorType *ColorMapEntry;
    char *p;
    int Size, Row, Col, Width, Height, ExtCode;
    int i, j;
    int value;
    int hist[256][2];

    int li = 0;
    int cl = 0;
    int ent[256];
    int th[3][2];

    /* Gif File Check */
    GifFile = DGifOpenFileName(f);
    if(!GifFile){
/*
        if(_GifError != D_GIF_ERR_NOT_GIF_FILE){
            PrintGifError();
        }
*/
        return -1;
    }

    ScreenBuffer = (GifRowType *)malloc(GifFile->SHeight *
                                        sizeof(GifRowType *));
    if(!ScreenBuffer){
        perror("malloc");
        if(DGifCloseFile(GifFile) == GIF_ERROR){
            PrintGifError();
        }
        return -1;
    }

    /* Size in bytes one row.*/
    Size = GifFile->SWidth * sizeof(GifPixelType);

    /* First row. */
    ScreenBuffer[0] = (GifRowType)malloc(Size);
    if(!ScreenBuffer[0]){
        perror("malloc");
        if(DGifCloseFile(GifFile) == GIF_ERROR){
            PrintGifError();
        }
        free(ScreenBuffer);
        return -1;
    }

    /* Set its color to BackGround. */
    for(i = 0; i < GifFile->SWidth ; i++){
        ScreenBuffer[0][i] = GifFile->SBackGroundColor;
    }

    for(i = 1; i < GifFile->SHeight ; i++){
        /* Allocate the other rows,
           and set their color to background too: */
        ScreenBuffer[i] = (GifRowType)malloc(Size);
        if(!ScreenBuffer[i]){
            perror("malloc");
            for(j = 1 ; j < (i - 1) ; j++){
                free(ScreenBuffer[i]);
            }
            free(ScreenBuffer[0]);
            free(ScreenBuffer);
            if(DGifCloseFile(GifFile) == GIF_ERROR){
                PrintGifError();
            }
            return -1;
        }
        memcpy(ScreenBuffer[i], ScreenBuffer[0], Size);
    }

    /* Scan the content of the GIF file and load the image(s) in: */
    do{
        if(DGifGetRecordType(GifFile, &RecordType) == GIF_ERROR){
            PrintGifError();
            for(i = 1 ; i < GifFile->SHeight ; i++){
                free(ScreenBuffer[i]);
            }
            free(ScreenBuffer[0]);
            free(ScreenBuffer);
            if(DGifCloseFile(GifFile) == GIF_ERROR){
                PrintGifError();
            }
            return -1;
        }
        switch(RecordType){
        case IMAGE_DESC_RECORD_TYPE:
            if(DGifGetImageDesc(GifFile) == GIF_ERROR){
                PrintGifError();
                for(i = 1 ; i < GifFile->SHeight ; i++){
                    free(ScreenBuffer[i]);
                }
                free(ScreenBuffer[0]);
                free(ScreenBuffer);
                if(DGifCloseFile(GifFile) == GIF_ERROR){
                    PrintGifError();
                }
                return -1;
            }
            /* Image Position relative to Screen.  */
            Row = GifFile->Image.Top;
            Col = GifFile->Image.Left;
            Width = GifFile->Image.Width;
            Height = GifFile->Image.Height;

            if(GifFile->Image.Left + GifFile->Image.Width > GifFile->SWidth ||
               GifFile->Image.Top + GifFile->Image.Height > GifFile->SHeight){
                fprintf(stderr,
                       "Image is not confined to screen dimension, aborted.\n");
                for(i = 1 ; i < GifFile->SHeight ; i++){
                    free(ScreenBuffer[i]);
                }
                free(ScreenBuffer[0]);
                free(ScreenBuffer);
                if(DGifCloseFile(GifFile) == GIF_ERROR){
                    PrintGifError();
                }
                return -1;
            }
            if(GifFile->Image.Interlace){
                /* Need to perform 4 passes on the images: */
                for(i = 0; i < 4; i++){
                    for(j = Row + InterlacedOffset[i];
                        j < Row + Height; j += InterlacedJumps[i]){
                        if(DGifGetLine(GifFile, &ScreenBuffer[j][Col],
                                       Width) == GIF_ERROR){
                            PrintGifError();
                            for(i = 1 ; i < GifFile->SHeight ; i++){
                                free(ScreenBuffer[i]);
                            }
                            free(ScreenBuffer[0]);
                            free(ScreenBuffer);
                            if(DGifCloseFile(GifFile) == GIF_ERROR){
                                PrintGifError();
                            }
                            return -1;
                        }
                    }
                }
            }
            else{
                for(i = 0; i < Height; i++){
                    if(DGifGetLine(GifFile, &ScreenBuffer[Row++][Col],
                                   Width) == GIF_ERROR){
                        PrintGifError();
                        for(i = 1 ; i < GifFile->SHeight ; i++){
                            free(ScreenBuffer[i]);
                        }
                        free(ScreenBuffer[0]);
                        free(ScreenBuffer);
                        if(DGifCloseFile(GifFile) == GIF_ERROR){
                            PrintGifError();
                        }
                        return -1;
                    }
                }
            }
            break;
        case EXTENSION_RECORD_TYPE:
            /* Skip any extension blocks in file: */
            if(DGifGetExtension(GifFile, &ExtCode, &Extension) == GIF_ERROR){
                PrintGifError();
                for(i = 1 ; i < GifFile->SHeight ; i++){
                    free(ScreenBuffer[i]);
                }
                free(ScreenBuffer[0]);
                free(ScreenBuffer);
                if(DGifCloseFile(GifFile) == GIF_ERROR){
                    PrintGifError();
                }
                return -1;
            }
            while(Extension != NULL){
                if(DGifGetExtensionNext(GifFile, &Extension) == GIF_ERROR){
                    PrintGifError();
                    for(i = 1 ; i < GifFile->SHeight ; i++){
                        free(ScreenBuffer[i]);
                    }
                    free(ScreenBuffer[0]);
                    free(ScreenBuffer);
                    if(DGifCloseFile(GifFile) == GIF_ERROR){
                        PrintGifError();
                    }
                    return -1;
                }
            }
            break;
        case TERMINATE_RECORD_TYPE:
            break;
        default:    /* Should be traps by DGifGetRecordType. */
            break;
        }
    }
    while(RecordType != TERMINATE_RECORD_TYPE);

    /* Lets dump it - set the global variables required and do it: */
    BackGround = GifFile->SBackGroundColor;
    ColorMap = (GifFile->Image.ColorMap ? GifFile->Image.ColorMap
                : GifFile->SColorMap);
    ColorMapSize = ColorMap->ColorCount;

    *w = GifFile->Image.Width;
    *h = GifFile->Image.Height;

    *image = (char *)malloc(*w * *h);
    if(!*image){
        perror("malloc");
        for(i = 1 ; i < GifFile->SHeight ; i++){
            free(ScreenBuffer[i]);
        }
        free(ScreenBuffer[0]);
        free(ScreenBuffer);
        return -1;
    }
    for(i = 0 ; i < 256 ; i++){
        hist[i][0] = 0;
        hist[i][1] = i;
    }
    p = *image;
    /***** To Gray ****/
    for(i = 0 ; i < GifFile->SHeight ; i++){
        GifRow = ScreenBuffer[i];
        for(j = 0 ; j < GifFile->SWidth ; j++) {
            ColorMapEntry = &ColorMap->Colors[GifRow[j]];
            value =  77 * (int)ColorMapEntry->Red +
                    150 * (int)ColorMapEntry->Green +
                     29 * (int)ColorMapEntry->Blue;
            value = (value + 128) >> 8;
            if(value < 0){
                value = 0;
            }
            if(value > 255){
                value = 255;
            }
            (hist[value][0])++;
            *p = (char)value;
            p++;
        }
    }

#if 1
    for(i = 1 ; i < GifFile->SHeight ; i++){
        free(ScreenBuffer[i]);
    }
    free(ScreenBuffer[0]);
    free(ScreenBuffer);
#endif

    if(DGifCloseFile(GifFile) == GIF_ERROR){
        PrintGifError();
        free(*image);
        return -1;
    }

    th[0][0] = 0; th[1][0] = 0; th[2][0] = 0;
    th[0][1] = 0; th[1][1] = 0; th[2][1] = 0;
    for(i = 0 ; i < 256 ; i++){
        if(hist[i][0] != 0){
            li = li + hist[i][0];
            cl++;
        }
        ent[i] = 0;
    }
    li = li / cl / 3;

    for(i = 0, cl = 0 ; i < 256 ; i++){
        if(hist[i][0] - li <= 0){
            continue;
        }
        ent[cl] = hist[i][0];
        cl++;
    }

    qsort((void *)ent, cl, sizeof(int), intcompare);

    li = cl / 4;
    for(i = 0 ; i < 3 ; i++){
        th[i][0] = ent[li * i];
    }
    for(i = 0 ; i < 256 ; i++){
        if(hist[i][0] == th[0][0]){
            th[0][1] = i;
        }
        if(hist[i][0] == th[1][0]){
            th[1][1] = i;
        }
        if(hist[i][0] == th[2][0]){
            th[2][1] = i;
        }
    }

    if(th[2][1] > th[1][1]){
        i = th[1][1];
        th[1][1] = th[2][1];
        th[2][1] = i;
    }
    if(th[1][1] > th[0][1]){
        i = th[0][1];
        th[0][1] = th[1][1];
        th[1][1] = i;
    }
    if(th[2][1] > th[1][1]){
        i = th[1][1];
        th[1][1] = th[2][1];
        th[2][1] = i;
    }

    if(th[0][1] == 0){
        th[0][1] = 0xc0;
    }
    if(th[1][1] == 0){
        th[1][1] = 0x80;
    }
    if(th[2][1] == 0){
        th[2][1] = 0x40;
    }

    if(th[0][1] == 255){
        th[0][1] = th[1][1] + (th[0][1] - th[1][1]) / 2;
    }

    p = *image;
    for(i = 0 ; i < (*w * *h) ; i++){
            value = (int)((unsigned char)(*p));
            if(value < th[2][1]){
                value = 0;
            }else
            if(th[2][1] <= value && value < th[1][1]){
                value = 1;
            }else
            if(th[1][1] <= value && value < th[0][1]){
                value = 2;
            }else
            if(th[0][1] <= value && value <= 0xff){
                value = 3;
            }else{
                value = 0;
            }
            *p = (char)value;
            p++;
    }

    /***** dither ****/

    return 0;
}
#endif /* USE_GIF */

#ifdef USE_JPEG
int read_jpeg_file(char *f, int *w, int *h, char **image)
{
    FILE *fd;

    /* File Open */
    fd = fopen(f, "rb");
    if(fd == (FILE *)NULL){
        fprintf(stderr, "open %s failed.\n", f);
        return -1;
    }

    /* File Read */
    *image = cont_jpeg(fd, w, h);
    if(!*image){
        fclose(fd);
        return -1;
    }

    fclose(fd);

    return 0;
}
#endif /* USE_JPEG */

/* image  : 8bit Gray */
/* bitmap : 2bit Gray */
int read_file(char *f, int *w, int *h, char **image)
{
    int ret = -1;

#ifdef USE_GIF
    ret = read_gif_file(f, w, h, image);
    if(!ret){
        return(ret);
    }
#endif /* USE_GIF */

#if defined (USE_JPEG)
    ret = read_jpeg_file(f, w, h, image);
#endif /* USE_JPEG */

#if !defined (USE_JPEG) && defined (USE_GIF)
    if(_GifError == D_GIF_ERR_NOT_GIF_FILE){
        PrintGifError();
    }
#endif /* !USE_JPEG && USE_GIF */

    return(ret);
}

void image_copy(struct screen *st, int w, int h,
                int *sx, int *sy, int *dw, int *dh)
{
#ifdef MINI
    *sx = 0 + MINI_SX;
    *sy = 0 + MINI_SY;
    *dw = w;
    *dh = h;

    if(w < MINI_WD){
        *sx = (MINI_WD - w) / 2 + MINI_SX;
    }
    if(h < MINI_HI){
        *sy = (MINI_HI - h) / 2 + MINI_SY;
    }
    if(w > MINI_WD){
        *dw  = MINI_WD;
    }
    if(h > MINI_HI){
        *dh  = MINI_HI;
    }
#else  /* non MINI */
    *sx = 0;
    *sy = 0;
    *dw = w;
    *dh = h;

    if(w < SCREEN_WIDTH){
        *sx = (SCREEN_WIDTH - w) / 2;
    }
    if(h < SCREEN_HEIGHT){
        *sy = (SCREEN_HEIGHT - h) / 2;
    }
    if(w > SCREEN_WIDTH){
        *dw  = SCREEN_WIDTH;
    }
    if(h > SCREEN_HEIGHT){
        *dh  = SCREEN_HEIGHT;
    }
#endif  /* MINI */
#if DEBUG
    printf("sx [%d]\n",*sx);
    printf("sy [%d]\n",*sy);
    printf("dw [%d]\n",*dw);
    printf("dh [%d]\n",*dh);
#endif
    bitblt(NULL, *sx, *sy, st, 0, 0, *dw, *dh, 0);
    return;
}

#ifdef USE_JPEG
#include <setjmp.h>

struct return_error_mgr{
  struct jpeg_error_mgr pub;
  jmp_buf setjmp_buffer;
};

typedef struct return_error_mgr * return_error_ptr;

void error_return (j_common_ptr cinfo)
{
  /* cinfo->err really points to a return_error_mgr struct, so coerce pointer */
  return_error_ptr returnerr = (return_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(returnerr->setjmp_buffer, 1);
}

char *cont_jpeg(FILE *fd, int *w, int *h)
{
    JSAMPROW buffer[1];
    int row_stride;
#if 0
    int ret;
#endif
    struct jpeg_decompress_struct cinfo;
#if 0
    struct jpeg_error_mgr jerr;
#else
    struct return_error_mgr jerr;
#endif
    char *jpeg_image;
#if 0
    char *p;
#endif
    int  pl;

    cinfo.err = jpeg_std_error(&jerr.pub);
#if 1
    cinfo.err->error_exit = error_return;

    if(setjmp(jerr.setjmp_buffer)){
       jpeg_destroy_decompress(&cinfo);
       return NULL;
    }
#endif

    jpeg_create_decompress(&cinfo);

    jpeg_stdio_src(&cinfo, fd);

    /* header read */
    jpeg_read_header(&cinfo, TRUE);

#if DEBUG
    /* Width and height of image */
    printf("image_width     [%d]\n",cinfo.image_width);
    printf("image_height    [%d]\n",cinfo.image_height);
    /* Number of color components */
    printf("num_components  [%d]\n",cinfo.num_components);
    /* Colorspace of image */
    printf("jpeg_color_space[%d]\n",cinfo.jpeg_color_space);
#endif

    cinfo.out_color_space = JCS_GRAYSCALE;

    /* JDITHER_NONE     no dithering: fast, very low quality */
    /* JDITHER_ORDERED  ordered dither: moderate speed and quality */
    /* JDITHER_FS       Floyd-Steinberg dither: slow, high quality */
    cinfo.dither_mode = JDITHER_NONE;
    cinfo.dither_mode = JDITHER_ORDERED;
    cinfo.dither_mode = JDITHER_FS;
    cinfo.quantize_colors = TRUE;
    cinfo.desired_number_of_colors = 4;

    jpeg_calc_output_dimensions(&cinfo);

#if DEBUG
    /* Actual dimensions of output image. */
    printf("output_width        [%d]\n",cinfo.output_width);
    printf("output_height       [%d]\n",cinfo.output_height);
    /* Number of color components in out_color_space. */
    printf("out_color_components[%d]\n",cinfo.out_color_components);
    /* Number of color components returned. */
    printf("output_components   [%d]\n",cinfo.output_components);
    /* Recommended height of scanline buffer. */
    printf("rec_outbuf_height   [%d]\n",cinfo.rec_outbuf_height);
#endif

    row_stride = cinfo.output_width * cinfo.output_components;

    *w  = (int)cinfo.output_width;
    *h  = (int)cinfo.output_height;
    pl  = (int)cinfo.output_components;

    jpeg_image = (char *)malloc(*w * *h * pl);
    if(!jpeg_image){
        perror("malloc");
        /* End */
        jpeg_finish_decompress(&cinfo);
        jpeg_destroy_decompress(&cinfo);

        return NULL;
    }

    /* Start decompressor */
    jpeg_start_decompress(&cinfo);

    while(cinfo.output_scanline < cinfo.output_height){
       buffer[0] = (JSAMPROW) &jpeg_image[cinfo.output_scanline * *w * pl];
       jpeg_read_scanlines(&cinfo, buffer, (JDIMENSION)1);
    }

    /* End */
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);

    return jpeg_image;
}
#endif /* USE_JPEG */
